/*
 * $Id$
 */

/*
 * $Id$
 */
#ifndef HIPE_X86_ASM_H
#define HIPE_X86_ASM_H

/*
 * Tunables.
 */

#define X86_LEAF_WORDS	24
#define LEAF_WORDS	24

/*
 * Reserved registers.
 */
#define P	%ebp

#define X86_HP_IN_ESI	1
#if X86_HP_IN_ESI
#define SAVE_HP		movl %esi, P_HP(P)
#define RESTORE_HP	movl P_HP(P), %esi
#else
#define SAVE_HP		/*empty*/
#define RESTORE_HP	/*empty*/
#endif

#define NSP		%esp
#define SAVE_CSP	movl %esp, P_CSP(P)
#define RESTORE_CSP	movl P_CSP(P), %esp

#define X86_SIMULATE_NSP	0

/*
 * Context switching macros.
 */
#define SWITCH_C_TO_ERLANG_QUICK	\
	SAVE_CSP; \
	movl P_NSP(P), NSP

#define SWITCH_ERLANG_TO_C_QUICK	\
	movl NSP, P_NSP(P); \
	RESTORE_CSP

#define SAVE_CACHED_STATE	\
	SAVE_HP

#define RESTORE_CACHED_STATE	\
	RESTORE_HP

#define SWITCH_C_TO_ERLANG	\
	RESTORE_CACHED_STATE;	\
	SWITCH_C_TO_ERLANG_QUICK

#define SWITCH_ERLANG_TO_C	\
	SAVE_CACHED_STATE;	\
	SWITCH_ERLANG_TO_C_QUICK

/*
 * Argument (parameter) registers.
 */
#define X86_NR_ARG_REGS	3
#define NR_ARG_REGS	3

#define ARG0	%eax
#define ARG1	%edx
#define ARG2	%ecx

/*
 * TEMP_RV:
 *	Used in nbif_stack_trap_ra to preserve the return value.
 *	Must be a C callee-save register.
 *	Must be otherwise unused in the return path.
 */
#define TEMP_RV	%ebx

/*
 * TEMP_NSP:
 *	Used in BIF wrappers to permit copying stacked parameter from
 *	the native stack to the C stack.
 *	Set up by NBIF_COPY_NSP(arity) and used by NBIF_ARG(arity,argno).
 *	TEMP_NSP may alias the last BIF argument register.
 *	NBIF_COPY_NSP and NBIF_ARG currently fail if ARITY > NR_ARG_REGS!
 */
#define TEMP_NSP	%edi


#define LOAD_ARG_REGS	movl P_ARG0(P), ARG0 ; movl P_ARG1(P), ARG1 ; movl P_ARG2(P), ARG2 ; 

#define STORE_ARG_REGS	movl ARG0, P_ARG0(P) ; movl ARG1, P_ARG1(P) ; movl ARG2, P_ARG2(P) ; 

#define NSP_CALL(FUN)	call FUN
#define NSP_RETN(NPOP)	ret $NPOP
#define NSP_RET0	ret

/* #define NBIF_COPY_NSP_0	 */
/* #define NBIF_COPY_NSP_1	 */
/* #define NBIF_COPY_NSP_2	 */
/* #define NBIF_COPY_NSP_3	 */
/* #define NBIF_COPY_NSP_5	movl	%esp, TEMP_NSP */

/* #define NBIF_ARG_1_0	ARG0 */
/* #define NBIF_ARG_2_0	ARG0 */
/* #define NBIF_ARG_2_1	ARG1 */
/* #define NBIF_ARG_3_0	ARG0 */
/* #define NBIF_ARG_3_1	ARG1 */
/* #define NBIF_ARG_3_2	ARG2 */

/* #define NBIF_RET_0	NSP_RET0 */
/* #define NBIF_RET_1	NSP_RET0 */
/* #define NBIF_RET_2	NSP_RET0 */
/* #define NBIF_RET_3	NSP_RET0 */
/* #define NBIF_RET_5	NSP_RETN(8) */

/* #define NBIF_SAVE_RESCHED_ARGS_1 movl ARG0, P_ARG0(P);  */
/* #define NBIF_SAVE_RESCHED_ARGS_2 movl ARG0, P_ARG0(P); movl ARG1, P_ARG1(P);  */

#define NBIF_SAVE_CALLER_SAVE	pushl ARG2 ; pushl ARG1 ; pushl ARG0 ; 
#define NBIF_RESTORE_CALLER_SAVE_1	movl 4(%esp), ARG0 ; movl 8(%esp), ARG1 ; movl 12(%esp), ARG2 ; addl $16, %esp

#endif /* HIPE_X86_ASM_H */

#include "hipe_literals.h"

/*
 * These m4 macros expand to assembly code which
 * is further processed using the C pre-processor:
 * - Expansion of symbolic names for registers and PCB fields.
 * - Conditional assembly. Some BIFs need specialised code.
 *   Instead of special-casing them in all generated BIF lists,
 *   we use #ifndef wrappers to allow hand-written code to
 *   override that generated by the standard m4 macros.
 *   This is used for:
 *   - demonitor/1, exit/2, group_leader/2, link/1, monitor/2,
 *     port_command/2, send/2, unlink/1: can fail with RESCHEDULE
 * - Appropriate BIF exception test for debug and non-debug mode.
 *
 * XXX: TODO:
 * - Can a BIF with arity 0 fail? beam_emu doesn't think so.
 */

#if THE_NON_VALUE == 0
#define TEST_GOT_EXN	testl	%eax,%eax
#else
#define TEST_GOT_EXN	cmpl	$THE_NON_VALUE,%eax
#endif

/*
 * standard_bif_interface_0(nbif_name, cbif_name)
 * standard_bif_interface_1(nbif_name, cbif_name)
 * standard_bif_interface_2(nbif_name, cbif_name)
 * standard_bif_interface_3(nbif_name, cbif_name)
 *
 * Generate native interface for a BIF with 0-3 parameters and
 * standard failure mode (may fail, but not with RESCHEDULE).
 */








/*
 * expensive_bif_interface_1(nbif_name, cbif_name)
 * expensive_bif_interface_2(nbif_name, cbif_name)
 *
 * Generate native interface for a BIF with 1-2 parameters and
 * an expensive failure mode (may fail with RESCHEDULE).
 */




/*
 * nofail_primop_interface_0(nbif_name, cbif_name)
 * nofail_primop_interface_1(nbif_name, cbif_name)
 * nofail_primop_interface_2(nbif_name, cbif_name)
 * nofail_primop_interface_3(nbif_name, cbif_name)
 *
 * Generate native interface for a primop with implicit P
 * parameter, 0-3 ordinary parameters and no failure mode.
 * Also used for guard BIFs.
 */








/*
 * nocons_nofail_primop_interface_0(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_1(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_2(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_3(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_5(nbif_name, cbif_name)
 *
 * Generate native interface for a primop with implicit P
 * parameter, 0-3 or 5 ordinary parameters, and no failure mode.
 * The primop cannot CONS or gc.
 */










/*
 * noproc_primop_interface_0(nbif_name, cbif_name)
 * noproc_primop_interface_1(nbif_name, cbif_name)
 * noproc_primop_interface_2(nbif_name, cbif_name)
 * noproc_primop_interface_3(nbif_name, cbif_name)
 * noproc_primop_interface_5(nbif_name, cbif_name)
 *
 * Generate native interface for a primop with no implicit P
 * parameter, 0-3 or 5 ordinary parameters, and no failure mode.
 * The primop cannot CONS or gc.
 */










/*
 * x86-specific primops.
 */

#ifndef HAVE_nbif_handle_fp_exception
#define HAVE_nbif_handle_fp_exception
	.section ".text"
	.align	4
	.global	nbif_handle_fp_exception
nbif_handle_fp_exception:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	call	erts_restore_fpu

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_handle_fp_exception,.-nbif_handle_fp_exception
	.type	nbif_handle_fp_exception,@function
#endif

/*
 * Implement gc_nofail_primop_interface_1 as nofail_primop_interface_1.
 */


/* $Id$
 *
 * List all non architecture-specific BIFs and primops, and
 * classify each as belonging to one of the classes below.
 * This list is included in hipe_${ARCH}_bifs.m4, which is
 * responsible for translating these classifications to the
 * best possible native code wrappers.
 *
 * XXX: We should have a more detailed BIF classification
 * with a number of orthogonal properties (e.g., UPDATES_HP,
 * NEEDS_NSP, CAN_FAIL, CAN_GC, etc), from which we should
 * generate appropriate interfaces.
 */

/****************************************************************
 *			BIF CLASS DESCRIPTIONS			*
 ****************************************************************/

/*
 * standard_bif_interface_0(nbif_name, cbif_name)
 * standard_bif_interface_1(nbif_name, cbif_name)
 * standard_bif_interface_2(nbif_name, cbif_name)
 * standard_bif_interface_3(nbif_name, cbif_name)
 *
 * A BIF with implicit P parameter, 0-3 ordinary parameters,
 * which may fail but not with RESCHEDULE.
 * Native P state (RA/NSP/HP/etc) may be accessed and updated.
 * [XXX: most will CONS but not access P->hipe]
 * GC may not occur (NSP/HP limits unchanged).
 * [XXX: Some actually do GC. SPARC could be broken.]
 */

/*
 * expensive_bif_interface_1(nbif_name, cbif_name)
 * expensive_bif_interface_2(nbif_name, cbif_name)
 *
 * A BIF which may fail with RESCHEDULE, otherwise
 * identical to standard_bif_interface_N.
 */

/*
 * nofail_primop_interface_0(nbif_name, cbif_name)
 * nofail_primop_interface_1(nbif_name, cbif_name)
 * nofail_primop_interface_2(nbif_name, cbif_name)
 * nofail_primop_interface_3(nbif_name, cbif_name)
 *
 * A primop or guard BIF with no failure mode, otherwise
 * identical to standard_bif_interface_N.
 */

/*
 * gc_nofail_primop_interface_1(nbif_name, cbif_name)
 *
 * A primop with implicit P parameter, 1 ordinary parameter,
 * and no failure mode.
 * Native P state (RA/NSP/HP/etc) may be accessed and updated.
 * GC may occur (NSP/HP limits changed).
 */

/*
 * nocons_nofail_primop_interface_0(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_1(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_2(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_3(nbif_name, cbif_name)
 * nocons_nofail_primop_interface_5(nbif_name, cbif_name)
 *
 * A primop with implicit P parameter, 0-3 or 5 ordinary parameters,
 * and no failure mode.
 * Native P state will not be accessed or updated.
 * GC may not occur (NSP/HP limits unchanged).
 */

/*
 * noproc_primop_interface_0(nbif_name, cbif_name)
 * noproc_primop_interface_1(nbif_name, cbif_name)
 * noproc_primop_interface_2(nbif_name, cbif_name)
 * noproc_primop_interface_3(nbif_name, cbif_name)
 * noproc_primop_interface_5(nbif_name, cbif_name)
 *
 * A primop with no P parameter, 0-3 or 5 ordinary parameters,
 * and no failure mode.
 * Native P state will not be accessed or updated.
 * GC may not occur (NSP/HP limits unchanged).
 */

/****************************************************************
 *			BIF CLASSIFICATION			*
 ****************************************************************/

/*
 * BIFs with expensive failure modes.
 */

#ifndef HAVE_nbif_demonitor_1
#define HAVE_nbif_demonitor_1
	.section ".text"
	.align	4
	.global	nbif_demonitor_1
nbif_demonitor_1:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	demonitor_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_demonitor_1, %edx	/* resumption address */
	jmp	nbif_1_hairy_exception
	.size	nbif_demonitor_1,.-nbif_demonitor_1
	.type	nbif_demonitor_1,@function
#endif

#ifndef HAVE_nbif_exit_2
#define HAVE_nbif_exit_2
	.section ".text"
	.align	4
	.global	nbif_exit_2
nbif_exit_2:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); movl ARG1, P_ARG1(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	exit_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_exit_2, %edx	/* resumption address */
	jmp	nbif_2_hairy_exception
	.size	nbif_exit_2,.-nbif_exit_2
	.type	nbif_exit_2,@function
#endif

#ifndef HAVE_nbif_group_leader_2
#define HAVE_nbif_group_leader_2
	.section ".text"
	.align	4
	.global	nbif_group_leader_2
nbif_group_leader_2:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); movl ARG1, P_ARG1(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	group_leader_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_group_leader_2, %edx	/* resumption address */
	jmp	nbif_2_hairy_exception
	.size	nbif_group_leader_2,.-nbif_group_leader_2
	.type	nbif_group_leader_2,@function
#endif

#ifndef HAVE_nbif_link_1
#define HAVE_nbif_link_1
	.section ".text"
	.align	4
	.global	nbif_link_1
nbif_link_1:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	link_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_link_1, %edx	/* resumption address */
	jmp	nbif_1_hairy_exception
	.size	nbif_link_1,.-nbif_link_1
	.type	nbif_link_1,@function
#endif

#ifndef HAVE_nbif_monitor_2
#define HAVE_nbif_monitor_2
	.section ".text"
	.align	4
	.global	nbif_monitor_2
nbif_monitor_2:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); movl ARG1, P_ARG1(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	monitor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_monitor_2, %edx	/* resumption address */
	jmp	nbif_2_hairy_exception
	.size	nbif_monitor_2,.-nbif_monitor_2
	.type	nbif_monitor_2,@function
#endif

#ifndef HAVE_nbif_port_command_2
#define HAVE_nbif_port_command_2
	.section ".text"
	.align	4
	.global	nbif_port_command_2
nbif_port_command_2:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); movl ARG1, P_ARG1(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_command_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_port_command_2, %edx	/* resumption address */
	jmp	nbif_2_hairy_exception
	.size	nbif_port_command_2,.-nbif_port_command_2
	.type	nbif_port_command_2,@function
#endif

#ifndef HAVE_nbif_send_2
#define HAVE_nbif_send_2
	.section ".text"
	.align	4
	.global	nbif_send_2
nbif_send_2:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); movl ARG1, P_ARG1(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	send_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_send_2, %edx	/* resumption address */
	jmp	nbif_2_hairy_exception
	.size	nbif_send_2,.-nbif_send_2
	.type	nbif_send_2,@function
#endif

#ifndef HAVE_nbif_unlink_1
#define HAVE_nbif_unlink_1
	.section ".text"
	.align	4
	.global	nbif_unlink_1
nbif_unlink_1:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	unlink_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_unlink_1, %edx	/* resumption address */
	jmp	nbif_1_hairy_exception
	.size	nbif_unlink_1,.-nbif_unlink_1
	.type	nbif_unlink_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_test_reschedule_1
#define HAVE_nbif_hipe_bifs_test_reschedule_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_test_reschedule_1
nbif_hipe_bifs_test_reschedule_1:
	/* copy native stack pointer */
	

	/* save actual parameters in case we must reschedule */
	movl ARG0, P_ARG0(P); 

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_test_reschedule_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	1f
	NSP_RET0
1:
	movl	$nbif_hipe_bifs_test_reschedule_1, %edx	/* resumption address */
	jmp	nbif_1_hairy_exception
	.size	nbif_hipe_bifs_test_reschedule_1,.-nbif_hipe_bifs_test_reschedule_1
	.type	nbif_hipe_bifs_test_reschedule_1,@function
#endif

/*
 * BIFs that may trigger a native stack walk with p->hipe.narity != 0.
 * Relevant when NR_ARG_REGS < the arity of the BIF.
 */

#ifndef HAVE_nbif_check_process_code_2
#define HAVE_nbif_check_process_code_2
	.section ".text"
	.align	4
	.global	nbif_check_process_code_2
nbif_check_process_code_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_check_process_code_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_check_process_code_2,.-nbif_check_process_code_2
	.type	nbif_check_process_code_2,@function
#endif

#ifndef HAVE_nbif_garbage_collect_1
#define HAVE_nbif_garbage_collect_1
	.section ".text"
	.align	4
	.global	nbif_garbage_collect_1
nbif_garbage_collect_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_garbage_collect_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_garbage_collect_1,.-nbif_garbage_collect_1
	.type	nbif_garbage_collect_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_nstack_1
#define HAVE_nbif_hipe_bifs_show_nstack_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_nstack_1
nbif_hipe_bifs_show_nstack_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_show_nstack_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_nstack_1,.-nbif_hipe_bifs_show_nstack_1
	.type	nbif_hipe_bifs_show_nstack_1,@function
#endif

/*
 * Arithmetic operators called indirectly by the HiPE compiler.
 */

#ifndef HAVE_nbif_add_2
#define HAVE_nbif_add_2
	.section ".text"
	.align	4
	.global	nbif_add_2
nbif_add_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_mixed_plus
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_add_2,.-nbif_add_2
	.type	nbif_add_2,@function
#endif

#ifndef HAVE_nbif_sub_2
#define HAVE_nbif_sub_2
	.section ".text"
	.align	4
	.global	nbif_sub_2
nbif_sub_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_mixed_minus
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_sub_2,.-nbif_sub_2
	.type	nbif_sub_2,@function
#endif

#ifndef HAVE_nbif_mul_2
#define HAVE_nbif_mul_2
	.section ".text"
	.align	4
	.global	nbif_mul_2
nbif_mul_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_mixed_times
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_mul_2,.-nbif_mul_2
	.type	nbif_mul_2,@function
#endif

#ifndef HAVE_nbif_div_2
#define HAVE_nbif_div_2
	.section ".text"
	.align	4
	.global	nbif_div_2
nbif_div_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_mixed_div
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_div_2,.-nbif_div_2
	.type	nbif_div_2,@function
#endif

#ifndef HAVE_nbif_intdiv_2
#define HAVE_nbif_intdiv_2
	.section ".text"
	.align	4
	.global	nbif_intdiv_2
nbif_intdiv_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	intdiv_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_intdiv_2,.-nbif_intdiv_2
	.type	nbif_intdiv_2,@function
#endif

#ifndef HAVE_nbif_rem_2
#define HAVE_nbif_rem_2
	.section ".text"
	.align	4
	.global	nbif_rem_2
nbif_rem_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	rem_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_rem_2,.-nbif_rem_2
	.type	nbif_rem_2,@function
#endif

#ifndef HAVE_nbif_bsl_2
#define HAVE_nbif_bsl_2
	.section ".text"
	.align	4
	.global	nbif_bsl_2
nbif_bsl_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bsl_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bsl_2,.-nbif_bsl_2
	.type	nbif_bsl_2,@function
#endif

#ifndef HAVE_nbif_bsr_2
#define HAVE_nbif_bsr_2
	.section ".text"
	.align	4
	.global	nbif_bsr_2
nbif_bsr_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bsr_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bsr_2,.-nbif_bsr_2
	.type	nbif_bsr_2,@function
#endif

#ifndef HAVE_nbif_band_2
#define HAVE_nbif_band_2
	.section ".text"
	.align	4
	.global	nbif_band_2
nbif_band_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	band_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_band_2,.-nbif_band_2
	.type	nbif_band_2,@function
#endif

#ifndef HAVE_nbif_bor_2
#define HAVE_nbif_bor_2
	.section ".text"
	.align	4
	.global	nbif_bor_2
nbif_bor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bor_2,.-nbif_bor_2
	.type	nbif_bor_2,@function
#endif

#ifndef HAVE_nbif_bxor_2
#define HAVE_nbif_bxor_2
	.section ".text"
	.align	4
	.global	nbif_bxor_2
nbif_bxor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bxor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bxor_2,.-nbif_bxor_2
	.type	nbif_bxor_2,@function
#endif

#ifndef HAVE_nbif_bnot_1
#define HAVE_nbif_bnot_1
	.section ".text"
	.align	4
	.global	nbif_bnot_1
nbif_bnot_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	bnot_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_bnot_1,.-nbif_bnot_1
	.type	nbif_bnot_1,@function
#endif

/*
 * Miscellaneous primops.
 */

#ifndef HAVE_nbif_gc_1
#define HAVE_nbif_gc_1
	.section ".text"
	.align	4
	.global	nbif_gc_1
nbif_gc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_gc
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* return */
	NSP_RET0
	.size	nbif_gc_1,.-nbif_gc_1
	.type	nbif_gc_1,@function
#endif

#ifndef HAVE_nbif_set_timeout
#define HAVE_nbif_set_timeout
	.section ".text"
	.align	4
	.global	nbif_set_timeout
nbif_set_timeout:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_set_timeout
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_set_timeout,.-nbif_set_timeout
	.type	nbif_set_timeout,@function
#endif

#ifndef HAVE_nbif_conv_big_to_float
#define HAVE_nbif_conv_big_to_float
	.section ".text"
	.align	4
	.global	nbif_conv_big_to_float
nbif_conv_big_to_float:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_conv_big_to_float
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_conv_big_to_float,.-nbif_conv_big_to_float
	.type	nbif_conv_big_to_float,@function
#endif

#ifndef HAVE_nbif_rethrow
#define HAVE_nbif_rethrow
	.section ".text"
	.align	4
	.global	nbif_rethrow
nbif_rethrow:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_rethrow
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_rethrow,.-nbif_rethrow
	.type	nbif_rethrow,@function
#endif

#ifndef HAVE_nbif_find_na_or_make_stub
#define HAVE_nbif_find_na_or_make_stub
	.section ".text"
	.align	4
	.global	nbif_find_na_or_make_stub
nbif_find_na_or_make_stub:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_find_na_or_make_stub
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_find_na_or_make_stub,.-nbif_find_na_or_make_stub
	.type	nbif_find_na_or_make_stub,@function
#endif

/*
 * Mbox primops with implicit P parameter.
 */

#ifndef HAVE_nbif_select_msg
#define HAVE_nbif_select_msg
	.section ".text"
	.align	4
	.global	nbif_select_msg
nbif_select_msg:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	P
	call	hipe_select_msg
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_select_msg,.-nbif_select_msg
	.type	nbif_select_msg,@function
#endif

/*
 * Primops without any P parameter.
 * These cannot CONS or gc.
 */

#ifndef HAVE_nbif_cmp_2
#define HAVE_nbif_cmp_2
	.section ".text"
	.align	4
	.global	nbif_cmp_2
nbif_cmp_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	call	cmp
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_cmp_2,.-nbif_cmp_2
	.type	nbif_cmp_2,@function
#endif

#ifndef HAVE_nbif_eq_2
#define HAVE_nbif_eq_2
	.section ".text"
	.align	4
	.global	nbif_eq_2
nbif_eq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	call	eq
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_eq_2,.-nbif_eq_2
	.type	nbif_eq_2,@function
#endif

/*
 * Bit-syntax primops with implicit P parameter.
 * XXX: only get_integer conses on the ordinary heap
 */

#ifndef HAVE_nbif_bs_get_integer
#define HAVE_nbif_bs_get_integer
	.section ".text"
	.align	4
	.global	nbif_bs_get_integer
nbif_bs_get_integer:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_bs_get_integer
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* return */
	NSP_RET0
	.size	nbif_bs_get_integer,.-nbif_bs_get_integer
	.type	nbif_bs_get_integer,@function
#endif

#ifndef HAVE_nbif_bs_get_binary
#define HAVE_nbif_bs_get_binary
	.section ".text"
	.align	4
	.global	nbif_bs_get_binary
nbif_bs_get_binary:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_bs_get_binary
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* return */
	NSP_RET0
	.size	nbif_bs_get_binary,.-nbif_bs_get_binary
	.type	nbif_bs_get_binary,@function
#endif

#ifndef HAVE_nbif_bs_get_float
#define HAVE_nbif_bs_get_float
	.section ".text"
	.align	4
	.global	nbif_bs_get_float
nbif_bs_get_float:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_bs_get_float
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* return */
	NSP_RET0
	.size	nbif_bs_get_float,.-nbif_bs_get_float
	.type	nbif_bs_get_float,@function
#endif

#ifndef HAVE_nbif_bs_get_binary_all
#define HAVE_nbif_bs_get_binary_all
	.section ".text"
	.align	4
	.global	nbif_bs_get_binary_all
nbif_bs_get_binary_all:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	erts_bs_get_binary_all
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* return */
	NSP_RET0
	.size	nbif_bs_get_binary_all,.-nbif_bs_get_binary_all
	.type	nbif_bs_get_binary_all,@function
#endif

#ifndef HAVE_nbif_bs_final
#define HAVE_nbif_bs_final
	.section ".text"
	.align	4
	.global	nbif_bs_final
nbif_bs_final:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	erts_bs_final
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* return */
	NSP_RET0
	.size	nbif_bs_final,.-nbif_bs_final
	.type	nbif_bs_final,@function
#endif

/*
 * Bit-syntax primops without any P parameter.
 * These cannot CONS or gc.
 */

#ifndef HAVE_nbif_bs_allocate
#define HAVE_nbif_bs_allocate
	.section ".text"
	.align	4
	.global	nbif_bs_allocate
nbif_bs_allocate:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG0
	call	hipe_bs_allocate
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_allocate,.-nbif_bs_allocate
	.type	nbif_bs_allocate,@function
#endif

/*
 * Bit-syntax primops. The ERTS_SMP runtime system requires P,
 * hence the use of nocons_nofail_primop_interface_N().
 * When ERTS_SMP is disabled, noproc_primop_interface_N()
 * should be used instead.
 */


#ifndef HAVE_nbif_bs_start_match
#define HAVE_nbif_bs_start_match
	.section ".text"
	.align	4
	.global	nbif_bs_start_match
nbif_bs_start_match:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG0
	call	erts_bs_start_match
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_start_match,.-nbif_bs_start_match
	.type	nbif_bs_start_match,@function
#endif

#ifndef HAVE_nbif_bs_skip_bits
#define HAVE_nbif_bs_skip_bits
	.section ".text"
	.align	4
	.global	nbif_bs_skip_bits
nbif_bs_skip_bits:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG0
	call	erts_bs_skip_bits
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_skip_bits,.-nbif_bs_skip_bits
	.type	nbif_bs_skip_bits,@function
#endif

#ifndef HAVE_nbif_bs_skip_bits_all
#define HAVE_nbif_bs_skip_bits_all
	.section ".text"
	.align	4
	.global	nbif_bs_skip_bits_all
nbif_bs_skip_bits_all:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	call	erts_bs_skip_bits_all

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_skip_bits_all,.-nbif_bs_skip_bits_all
	.type	nbif_bs_skip_bits_all,@function
#endif

#ifndef HAVE_nbif_bs_test_tail
#define HAVE_nbif_bs_test_tail
	.section ".text"
	.align	4
	.global	nbif_bs_test_tail
nbif_bs_test_tail:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG0
	call	erts_bs_test_tail
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_test_tail,.-nbif_bs_test_tail
	.type	nbif_bs_test_tail,@function
#endif

#ifndef HAVE_nbif_bs_save
#define HAVE_nbif_bs_save
	.section ".text"
	.align	4
	.global	nbif_bs_save
nbif_bs_save:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG0
	call	erts_bs_save
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_save,.-nbif_bs_save
	.type	nbif_bs_save,@function
#endif

#ifndef HAVE_nbif_bs_restore
#define HAVE_nbif_bs_restore
	.section ".text"
	.align	4
	.global	nbif_bs_restore
nbif_bs_restore:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG0
	call	erts_bs_restore
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_restore,.-nbif_bs_restore
	.type	nbif_bs_restore,@function
#endif

#ifndef HAVE_nbif_bs_init
#define HAVE_nbif_bs_init
	.section ".text"
	.align	4
	.global	nbif_bs_init
nbif_bs_init:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	call	erts_bs_init

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_init,.-nbif_bs_init
	.type	nbif_bs_init,@function
#endif

#ifndef HAVE_nbif_bs_put_integer
#define HAVE_nbif_bs_put_integer
	.section ".text"
	.align	4
	.global	nbif_bs_put_integer
nbif_bs_put_integer:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	call	erts_bs_put_integer
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_put_integer,.-nbif_bs_put_integer
	.type	nbif_bs_put_integer,@function
#endif

#ifndef HAVE_nbif_bs_put_binary
#define HAVE_nbif_bs_put_binary
	.section ".text"
	.align	4
	.global	nbif_bs_put_binary
nbif_bs_put_binary:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	call	erts_bs_put_binary
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_put_binary,.-nbif_bs_put_binary
	.type	nbif_bs_put_binary,@function
#endif

#ifndef HAVE_nbif_bs_put_binary_all
#define HAVE_nbif_bs_put_binary_all
	.section ".text"
	.align	4
	.global	nbif_bs_put_binary_all
nbif_bs_put_binary_all:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG0
	call	erts_bs_put_binary_all
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_put_binary_all,.-nbif_bs_put_binary_all
	.type	nbif_bs_put_binary_all,@function
#endif

#ifndef HAVE_nbif_bs_put_float
#define HAVE_nbif_bs_put_float
	.section ".text"
	.align	4
	.global	nbif_bs_put_float
nbif_bs_put_float:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	call	erts_bs_put_float
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_put_float,.-nbif_bs_put_float
	.type	nbif_bs_put_float,@function
#endif

#ifndef HAVE_nbif_bs_put_string
#define HAVE_nbif_bs_put_string
	.section ".text"
	.align	4
	.global	nbif_bs_put_string
nbif_bs_put_string:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	call	erts_bs_put_string
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RET0
	.size	nbif_bs_put_string,.-nbif_bs_put_string
	.type	nbif_bs_put_string,@function
#endif

#ifndef HAVE_nbif_bs_put_big_integer
#define HAVE_nbif_bs_put_big_integer
	.section ".text"
	.align	4
	.global	nbif_bs_put_big_integer
nbif_bs_put_big_integer:
	/* copy native stack pointer */
	movl	%esp, TEMP_NSP

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	4(TEMP_NSP)
	pushl	8(TEMP_NSP)
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	call	hipe_bs_put_big_integer
	addl	$20, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RETN(8)
	.size	nbif_bs_put_big_integer,.-nbif_bs_put_big_integer
	.type	nbif_bs_put_big_integer,@function
#endif

#ifndef HAVE_nbif_bs_put_small_float
#define HAVE_nbif_bs_put_small_float
	.section ".text"
	.align	4
	.global	nbif_bs_put_small_float
nbif_bs_put_small_float:
	/* copy native stack pointer */
	movl	%esp, TEMP_NSP

	/* switch to C stack */
	SWITCH_ERLANG_TO_C_QUICK

	/* make the call on the C stack */
	pushl	4(TEMP_NSP)
	pushl	8(TEMP_NSP)
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	call	hipe_bs_put_small_float
	addl	$20, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG_QUICK

	/* return */
	NSP_RETN(8)
	.size	nbif_bs_put_small_float,.-nbif_bs_put_small_float
	.type	nbif_bs_put_small_float,@function
#endif

/*
 * SMP-specific stuff
 */

/*
 * Standard BIFs.
 * BIF_LIST(ModuleAtom,FunctionAtom,Arity,CFun,Index)
 */

/*
 *  Warning: Do not edit this file.  It was automatically
 *  generated by 'make_tables' on Tue Feb 24 03:53:18 2009.
 */


#ifndef HAVE_nbif_abs_1
#define HAVE_nbif_abs_1
	.section ".text"
	.align	4
	.global	nbif_abs_1
nbif_abs_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	abs_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_abs_1,.-nbif_abs_1
	.type	nbif_abs_1,@function
#endif

#ifndef HAVE_nbif_ebif_abs_1
#define HAVE_nbif_ebif_abs_1
	.section ".text"
	.align	4
	.global	nbif_ebif_abs_1
nbif_ebif_abs_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_abs_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_abs_1,.-nbif_ebif_abs_1
	.type	nbif_ebif_abs_1,@function
#endif

#ifndef HAVE_nbif_apply_3
#define HAVE_nbif_apply_3
	.section ".text"
	.align	4
	.global	nbif_apply_3
nbif_apply_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	apply_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_apply_3,.-nbif_apply_3
	.type	nbif_apply_3,@function
#endif

#ifndef HAVE_nbif_ebif_apply_3
#define HAVE_nbif_ebif_apply_3
	.section ".text"
	.align	4
	.global	nbif_ebif_apply_3
nbif_ebif_apply_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_apply_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_apply_3,.-nbif_ebif_apply_3
	.type	nbif_ebif_apply_3,@function
#endif

#ifndef HAVE_nbif_atom_to_list_1
#define HAVE_nbif_atom_to_list_1
	.section ".text"
	.align	4
	.global	nbif_atom_to_list_1
nbif_atom_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	atom_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_atom_to_list_1,.-nbif_atom_to_list_1
	.type	nbif_atom_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_atom_to_string_1
#define HAVE_nbif_ebif_atom_to_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_atom_to_string_1
nbif_ebif_atom_to_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_atom_to_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_atom_to_string_1,.-nbif_ebif_atom_to_string_1
	.type	nbif_ebif_atom_to_string_1,@function
#endif

#ifndef HAVE_nbif_binary_to_list_1
#define HAVE_nbif_binary_to_list_1
	.section ".text"
	.align	4
	.global	nbif_binary_to_list_1
nbif_binary_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	binary_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_binary_to_list_1,.-nbif_binary_to_list_1
	.type	nbif_binary_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_binary_to_list_1
#define HAVE_nbif_ebif_binary_to_list_1
	.section ".text"
	.align	4
	.global	nbif_ebif_binary_to_list_1
nbif_ebif_binary_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_binary_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_binary_to_list_1,.-nbif_ebif_binary_to_list_1
	.type	nbif_ebif_binary_to_list_1,@function
#endif

#ifndef HAVE_nbif_binary_to_list_3
#define HAVE_nbif_binary_to_list_3
	.section ".text"
	.align	4
	.global	nbif_binary_to_list_3
nbif_binary_to_list_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	binary_to_list_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_binary_to_list_3,.-nbif_binary_to_list_3
	.type	nbif_binary_to_list_3,@function
#endif

#ifndef HAVE_nbif_ebif_binary_to_list_3
#define HAVE_nbif_ebif_binary_to_list_3
	.section ".text"
	.align	4
	.global	nbif_ebif_binary_to_list_3
nbif_ebif_binary_to_list_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_binary_to_list_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_binary_to_list_3,.-nbif_ebif_binary_to_list_3
	.type	nbif_ebif_binary_to_list_3,@function
#endif

#ifndef HAVE_nbif_binary_to_term_1
#define HAVE_nbif_binary_to_term_1
	.section ".text"
	.align	4
	.global	nbif_binary_to_term_1
nbif_binary_to_term_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	binary_to_term_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_binary_to_term_1,.-nbif_binary_to_term_1
	.type	nbif_binary_to_term_1,@function
#endif

#ifndef HAVE_nbif_ebif_binary_to_term_1
#define HAVE_nbif_ebif_binary_to_term_1
	.section ".text"
	.align	4
	.global	nbif_ebif_binary_to_term_1
nbif_ebif_binary_to_term_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_binary_to_term_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_binary_to_term_1,.-nbif_ebif_binary_to_term_1
	.type	nbif_ebif_binary_to_term_1,@function
#endif

#ifndef HAVE_nbif_check_process_code_2
#define HAVE_nbif_check_process_code_2
	.section ".text"
	.align	4
	.global	nbif_check_process_code_2
nbif_check_process_code_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	check_process_code_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_check_process_code_2,.-nbif_check_process_code_2
	.type	nbif_check_process_code_2,@function
#endif

#ifndef HAVE_nbif_ebif_check_process_code_2
#define HAVE_nbif_ebif_check_process_code_2
	.section ".text"
	.align	4
	.global	nbif_ebif_check_process_code_2
nbif_ebif_check_process_code_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_check_process_code_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_check_process_code_2,.-nbif_ebif_check_process_code_2
	.type	nbif_ebif_check_process_code_2,@function
#endif

#ifndef HAVE_nbif_date_0
#define HAVE_nbif_date_0
	.section ".text"
	.align	4
	.global	nbif_date_0
nbif_date_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	date_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_date_0,.-nbif_date_0
	.type	nbif_date_0,@function
#endif

#ifndef HAVE_nbif_ebif_date_0
#define HAVE_nbif_ebif_date_0
	.section ".text"
	.align	4
	.global	nbif_ebif_date_0
nbif_ebif_date_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_date_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_date_0,.-nbif_ebif_date_0
	.type	nbif_ebif_date_0,@function
#endif

#ifndef HAVE_nbif_delete_module_1
#define HAVE_nbif_delete_module_1
	.section ".text"
	.align	4
	.global	nbif_delete_module_1
nbif_delete_module_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	delete_module_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_delete_module_1,.-nbif_delete_module_1
	.type	nbif_delete_module_1,@function
#endif

#ifndef HAVE_nbif_ebif_delete_module_1
#define HAVE_nbif_ebif_delete_module_1
	.section ".text"
	.align	4
	.global	nbif_ebif_delete_module_1
nbif_ebif_delete_module_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_delete_module_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_delete_module_1,.-nbif_ebif_delete_module_1
	.type	nbif_ebif_delete_module_1,@function
#endif

#ifndef HAVE_nbif_display_1
#define HAVE_nbif_display_1
	.section ".text"
	.align	4
	.global	nbif_display_1
nbif_display_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	display_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_display_1,.-nbif_display_1
	.type	nbif_display_1,@function
#endif

#ifndef HAVE_nbif_ebif_display_1
#define HAVE_nbif_ebif_display_1
	.section ".text"
	.align	4
	.global	nbif_ebif_display_1
nbif_ebif_display_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_display_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_display_1,.-nbif_ebif_display_1
	.type	nbif_ebif_display_1,@function
#endif

#ifndef HAVE_nbif_display_string_1
#define HAVE_nbif_display_string_1
	.section ".text"
	.align	4
	.global	nbif_display_string_1
nbif_display_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	display_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_display_string_1,.-nbif_display_string_1
	.type	nbif_display_string_1,@function
#endif

#ifndef HAVE_nbif_ebif_display_string_1
#define HAVE_nbif_ebif_display_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_display_string_1
nbif_ebif_display_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_display_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_display_string_1,.-nbif_ebif_display_string_1
	.type	nbif_ebif_display_string_1,@function
#endif

#ifndef HAVE_nbif_display_nl_0
#define HAVE_nbif_display_nl_0
	.section ".text"
	.align	4
	.global	nbif_display_nl_0
nbif_display_nl_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	display_nl_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_display_nl_0,.-nbif_display_nl_0
	.type	nbif_display_nl_0,@function
#endif

#ifndef HAVE_nbif_ebif_display_nl_0
#define HAVE_nbif_ebif_display_nl_0
	.section ".text"
	.align	4
	.global	nbif_ebif_display_nl_0
nbif_ebif_display_nl_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_display_nl_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_display_nl_0,.-nbif_ebif_display_nl_0
	.type	nbif_ebif_display_nl_0,@function
#endif

#ifndef HAVE_nbif_element_2
#define HAVE_nbif_element_2
	.section ".text"
	.align	4
	.global	nbif_element_2
nbif_element_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	element_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_element_2,.-nbif_element_2
	.type	nbif_element_2,@function
#endif

#ifndef HAVE_nbif_ebif_element_2
#define HAVE_nbif_ebif_element_2
	.section ".text"
	.align	4
	.global	nbif_ebif_element_2
nbif_ebif_element_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_element_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_element_2,.-nbif_ebif_element_2
	.type	nbif_ebif_element_2,@function
#endif

#ifndef HAVE_nbif_erase_0
#define HAVE_nbif_erase_0
	.section ".text"
	.align	4
	.global	nbif_erase_0
nbif_erase_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	erase_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_erase_0,.-nbif_erase_0
	.type	nbif_erase_0,@function
#endif

#ifndef HAVE_nbif_ebif_erase_0
#define HAVE_nbif_ebif_erase_0
	.section ".text"
	.align	4
	.global	nbif_ebif_erase_0
nbif_ebif_erase_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_erase_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_erase_0,.-nbif_ebif_erase_0
	.type	nbif_ebif_erase_0,@function
#endif

#ifndef HAVE_nbif_erase_1
#define HAVE_nbif_erase_1
	.section ".text"
	.align	4
	.global	nbif_erase_1
nbif_erase_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	erase_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_erase_1,.-nbif_erase_1
	.type	nbif_erase_1,@function
#endif

#ifndef HAVE_nbif_ebif_erase_1
#define HAVE_nbif_ebif_erase_1
	.section ".text"
	.align	4
	.global	nbif_ebif_erase_1
nbif_ebif_erase_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_erase_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_erase_1,.-nbif_ebif_erase_1
	.type	nbif_ebif_erase_1,@function
#endif

#ifndef HAVE_nbif_exit_1
#define HAVE_nbif_exit_1
	.section ".text"
	.align	4
	.global	nbif_exit_1
nbif_exit_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	exit_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_exit_1,.-nbif_exit_1
	.type	nbif_exit_1,@function
#endif

#ifndef HAVE_nbif_ebif_exit_1
#define HAVE_nbif_ebif_exit_1
	.section ".text"
	.align	4
	.global	nbif_ebif_exit_1
nbif_ebif_exit_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_exit_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_exit_1,.-nbif_ebif_exit_1
	.type	nbif_ebif_exit_1,@function
#endif

#ifndef HAVE_nbif_exit_2
#define HAVE_nbif_exit_2
	.section ".text"
	.align	4
	.global	nbif_exit_2
nbif_exit_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	exit_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_exit_2,.-nbif_exit_2
	.type	nbif_exit_2,@function
#endif

#ifndef HAVE_nbif_ebif_signal_2
#define HAVE_nbif_ebif_signal_2
	.section ".text"
	.align	4
	.global	nbif_ebif_signal_2
nbif_ebif_signal_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_signal_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_signal_2,.-nbif_ebif_signal_2
	.type	nbif_ebif_signal_2,@function
#endif

#ifndef HAVE_nbif_external_size_1
#define HAVE_nbif_external_size_1
	.section ".text"
	.align	4
	.global	nbif_external_size_1
nbif_external_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	external_size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_external_size_1,.-nbif_external_size_1
	.type	nbif_external_size_1,@function
#endif

#ifndef HAVE_nbif_ebif_external_size_1
#define HAVE_nbif_ebif_external_size_1
	.section ".text"
	.align	4
	.global	nbif_ebif_external_size_1
nbif_ebif_external_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_external_size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_external_size_1,.-nbif_ebif_external_size_1
	.type	nbif_ebif_external_size_1,@function
#endif

#ifndef HAVE_nbif_float_1
#define HAVE_nbif_float_1
	.section ".text"
	.align	4
	.global	nbif_float_1
nbif_float_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	float_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_float_1,.-nbif_float_1
	.type	nbif_float_1,@function
#endif

#ifndef HAVE_nbif_ebif_to_float_1
#define HAVE_nbif_ebif_to_float_1
	.section ".text"
	.align	4
	.global	nbif_ebif_to_float_1
nbif_ebif_to_float_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_to_float_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_to_float_1,.-nbif_ebif_to_float_1
	.type	nbif_ebif_to_float_1,@function
#endif

#ifndef HAVE_nbif_float_to_list_1
#define HAVE_nbif_float_to_list_1
	.section ".text"
	.align	4
	.global	nbif_float_to_list_1
nbif_float_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	float_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_float_to_list_1,.-nbif_float_to_list_1
	.type	nbif_float_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_float_to_string_1
#define HAVE_nbif_ebif_float_to_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_float_to_string_1
nbif_ebif_float_to_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_float_to_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_float_to_string_1,.-nbif_ebif_float_to_string_1
	.type	nbif_ebif_float_to_string_1,@function
#endif

#ifndef HAVE_nbif_fun_info_2
#define HAVE_nbif_fun_info_2
	.section ".text"
	.align	4
	.global	nbif_fun_info_2
nbif_fun_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	fun_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_fun_info_2,.-nbif_fun_info_2
	.type	nbif_fun_info_2,@function
#endif

#ifndef HAVE_nbif_ebif_fun_info_2
#define HAVE_nbif_ebif_fun_info_2
	.section ".text"
	.align	4
	.global	nbif_ebif_fun_info_2
nbif_ebif_fun_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_fun_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_fun_info_2,.-nbif_ebif_fun_info_2
	.type	nbif_ebif_fun_info_2,@function
#endif

#ifndef HAVE_nbif_garbage_collect_0
#define HAVE_nbif_garbage_collect_0
	.section ".text"
	.align	4
	.global	nbif_garbage_collect_0
nbif_garbage_collect_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	garbage_collect_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_garbage_collect_0,.-nbif_garbage_collect_0
	.type	nbif_garbage_collect_0,@function
#endif

#ifndef HAVE_nbif_ebif_garbage_collect_0
#define HAVE_nbif_ebif_garbage_collect_0
	.section ".text"
	.align	4
	.global	nbif_ebif_garbage_collect_0
nbif_ebif_garbage_collect_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_garbage_collect_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_garbage_collect_0,.-nbif_ebif_garbage_collect_0
	.type	nbif_ebif_garbage_collect_0,@function
#endif

#ifndef HAVE_nbif_garbage_collect_1
#define HAVE_nbif_garbage_collect_1
	.section ".text"
	.align	4
	.global	nbif_garbage_collect_1
nbif_garbage_collect_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	garbage_collect_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_garbage_collect_1,.-nbif_garbage_collect_1
	.type	nbif_garbage_collect_1,@function
#endif

#ifndef HAVE_nbif_ebif_garbage_collect_1
#define HAVE_nbif_ebif_garbage_collect_1
	.section ".text"
	.align	4
	.global	nbif_ebif_garbage_collect_1
nbif_ebif_garbage_collect_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_garbage_collect_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_garbage_collect_1,.-nbif_ebif_garbage_collect_1
	.type	nbif_ebif_garbage_collect_1,@function
#endif

#ifndef HAVE_nbif_garbage_collect_message_area_0
#define HAVE_nbif_garbage_collect_message_area_0
	.section ".text"
	.align	4
	.global	nbif_garbage_collect_message_area_0
nbif_garbage_collect_message_area_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	garbage_collect_message_area_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_garbage_collect_message_area_0,.-nbif_garbage_collect_message_area_0
	.type	nbif_garbage_collect_message_area_0,@function
#endif

#ifndef HAVE_nbif_ebif_garbage_collect_message_area_0
#define HAVE_nbif_ebif_garbage_collect_message_area_0
	.section ".text"
	.align	4
	.global	nbif_ebif_garbage_collect_message_area_0
nbif_ebif_garbage_collect_message_area_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_garbage_collect_message_area_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_garbage_collect_message_area_0,.-nbif_ebif_garbage_collect_message_area_0
	.type	nbif_ebif_garbage_collect_message_area_0,@function
#endif

#ifndef HAVE_nbif_get_0
#define HAVE_nbif_get_0
	.section ".text"
	.align	4
	.global	nbif_get_0
nbif_get_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	get_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_get_0,.-nbif_get_0
	.type	nbif_get_0,@function
#endif

#ifndef HAVE_nbif_ebif_get_0
#define HAVE_nbif_ebif_get_0
	.section ".text"
	.align	4
	.global	nbif_ebif_get_0
nbif_ebif_get_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_get_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_get_0,.-nbif_ebif_get_0
	.type	nbif_ebif_get_0,@function
#endif

#ifndef HAVE_nbif_get_1
#define HAVE_nbif_get_1
	.section ".text"
	.align	4
	.global	nbif_get_1
nbif_get_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	get_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_get_1,.-nbif_get_1
	.type	nbif_get_1,@function
#endif

#ifndef HAVE_nbif_ebif_get_1
#define HAVE_nbif_ebif_get_1
	.section ".text"
	.align	4
	.global	nbif_ebif_get_1
nbif_ebif_get_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_get_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_get_1,.-nbif_ebif_get_1
	.type	nbif_ebif_get_1,@function
#endif

#ifndef HAVE_nbif_get_keys_1
#define HAVE_nbif_get_keys_1
	.section ".text"
	.align	4
	.global	nbif_get_keys_1
nbif_get_keys_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	get_keys_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_get_keys_1,.-nbif_get_keys_1
	.type	nbif_get_keys_1,@function
#endif

#ifndef HAVE_nbif_ebif_get_keys_1
#define HAVE_nbif_ebif_get_keys_1
	.section ".text"
	.align	4
	.global	nbif_ebif_get_keys_1
nbif_ebif_get_keys_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_get_keys_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_get_keys_1,.-nbif_ebif_get_keys_1
	.type	nbif_ebif_get_keys_1,@function
#endif

#ifndef HAVE_nbif_group_leader_0
#define HAVE_nbif_group_leader_0
	.section ".text"
	.align	4
	.global	nbif_group_leader_0
nbif_group_leader_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	group_leader_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_group_leader_0,.-nbif_group_leader_0
	.type	nbif_group_leader_0,@function
#endif

#ifndef HAVE_nbif_ebif_group_leader_0
#define HAVE_nbif_ebif_group_leader_0
	.section ".text"
	.align	4
	.global	nbif_ebif_group_leader_0
nbif_ebif_group_leader_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_group_leader_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_group_leader_0,.-nbif_ebif_group_leader_0
	.type	nbif_ebif_group_leader_0,@function
#endif

#ifndef HAVE_nbif_group_leader_2
#define HAVE_nbif_group_leader_2
	.section ".text"
	.align	4
	.global	nbif_group_leader_2
nbif_group_leader_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	group_leader_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_group_leader_2,.-nbif_group_leader_2
	.type	nbif_group_leader_2,@function
#endif

#ifndef HAVE_nbif_ebif_group_leader_2
#define HAVE_nbif_ebif_group_leader_2
	.section ".text"
	.align	4
	.global	nbif_ebif_group_leader_2
nbif_ebif_group_leader_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_group_leader_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_group_leader_2,.-nbif_ebif_group_leader_2
	.type	nbif_ebif_group_leader_2,@function
#endif

#ifndef HAVE_nbif_halt_0
#define HAVE_nbif_halt_0
	.section ".text"
	.align	4
	.global	nbif_halt_0
nbif_halt_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	halt_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_halt_0,.-nbif_halt_0
	.type	nbif_halt_0,@function
#endif

#ifndef HAVE_nbif_ebif_halt_0
#define HAVE_nbif_ebif_halt_0
	.section ".text"
	.align	4
	.global	nbif_ebif_halt_0
nbif_ebif_halt_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_halt_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_halt_0,.-nbif_ebif_halt_0
	.type	nbif_ebif_halt_0,@function
#endif

#ifndef HAVE_nbif_halt_1
#define HAVE_nbif_halt_1
	.section ".text"
	.align	4
	.global	nbif_halt_1
nbif_halt_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	halt_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_halt_1,.-nbif_halt_1
	.type	nbif_halt_1,@function
#endif

#ifndef HAVE_nbif_ebif_halt_1
#define HAVE_nbif_ebif_halt_1
	.section ".text"
	.align	4
	.global	nbif_ebif_halt_1
nbif_ebif_halt_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_halt_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_halt_1,.-nbif_ebif_halt_1
	.type	nbif_ebif_halt_1,@function
#endif

#ifndef HAVE_nbif_phash_2
#define HAVE_nbif_phash_2
	.section ".text"
	.align	4
	.global	nbif_phash_2
nbif_phash_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	phash_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_phash_2,.-nbif_phash_2
	.type	nbif_phash_2,@function
#endif

#ifndef HAVE_nbif_phash2_1
#define HAVE_nbif_phash2_1
	.section ".text"
	.align	4
	.global	nbif_phash2_1
nbif_phash2_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	phash2_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_phash2_1,.-nbif_phash2_1
	.type	nbif_phash2_1,@function
#endif

#ifndef HAVE_nbif_phash2_2
#define HAVE_nbif_phash2_2
	.section ".text"
	.align	4
	.global	nbif_phash2_2
nbif_phash2_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	phash2_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_phash2_2,.-nbif_phash2_2
	.type	nbif_phash2_2,@function
#endif

#ifndef HAVE_nbif_ebif_phash2_1
#define HAVE_nbif_ebif_phash2_1
	.section ".text"
	.align	4
	.global	nbif_ebif_phash2_1
nbif_ebif_phash2_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_phash2_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_phash2_1,.-nbif_ebif_phash2_1
	.type	nbif_ebif_phash2_1,@function
#endif

#ifndef HAVE_nbif_ebif_phash2_2
#define HAVE_nbif_ebif_phash2_2
	.section ".text"
	.align	4
	.global	nbif_ebif_phash2_2
nbif_ebif_phash2_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_phash2_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_phash2_2,.-nbif_ebif_phash2_2
	.type	nbif_ebif_phash2_2,@function
#endif

#ifndef HAVE_nbif_hd_1
#define HAVE_nbif_hd_1
	.section ".text"
	.align	4
	.global	nbif_hd_1
nbif_hd_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hd_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hd_1,.-nbif_hd_1
	.type	nbif_hd_1,@function
#endif

#ifndef HAVE_nbif_ebif_hd_1
#define HAVE_nbif_ebif_hd_1
	.section ".text"
	.align	4
	.global	nbif_ebif_hd_1
nbif_ebif_hd_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_hd_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_hd_1,.-nbif_ebif_hd_1
	.type	nbif_ebif_hd_1,@function
#endif

#ifndef HAVE_nbif_integer_to_list_1
#define HAVE_nbif_integer_to_list_1
	.section ".text"
	.align	4
	.global	nbif_integer_to_list_1
nbif_integer_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	integer_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_integer_to_list_1,.-nbif_integer_to_list_1
	.type	nbif_integer_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_integer_to_string_1
#define HAVE_nbif_ebif_integer_to_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_integer_to_string_1
nbif_ebif_integer_to_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_integer_to_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_integer_to_string_1,.-nbif_ebif_integer_to_string_1
	.type	nbif_ebif_integer_to_string_1,@function
#endif

#ifndef HAVE_nbif_is_alive_0
#define HAVE_nbif_is_alive_0
	.section ".text"
	.align	4
	.global	nbif_is_alive_0
nbif_is_alive_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	is_alive_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_is_alive_0,.-nbif_is_alive_0
	.type	nbif_is_alive_0,@function
#endif

#ifndef HAVE_nbif_ebif_is_alive_0
#define HAVE_nbif_ebif_is_alive_0
	.section ".text"
	.align	4
	.global	nbif_ebif_is_alive_0
nbif_ebif_is_alive_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_is_alive_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_alive_0,.-nbif_ebif_is_alive_0
	.type	nbif_ebif_is_alive_0,@function
#endif

#ifndef HAVE_nbif_length_1
#define HAVE_nbif_length_1
	.section ".text"
	.align	4
	.global	nbif_length_1
nbif_length_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	length_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_length_1,.-nbif_length_1
	.type	nbif_length_1,@function
#endif

#ifndef HAVE_nbif_ebif_length_1
#define HAVE_nbif_ebif_length_1
	.section ".text"
	.align	4
	.global	nbif_ebif_length_1
nbif_ebif_length_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_length_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_length_1,.-nbif_ebif_length_1
	.type	nbif_ebif_length_1,@function
#endif

#ifndef HAVE_nbif_link_1
#define HAVE_nbif_link_1
	.section ".text"
	.align	4
	.global	nbif_link_1
nbif_link_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	link_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_link_1,.-nbif_link_1
	.type	nbif_link_1,@function
#endif

#ifndef HAVE_nbif_ebif_link_1
#define HAVE_nbif_ebif_link_1
	.section ".text"
	.align	4
	.global	nbif_ebif_link_1
nbif_ebif_link_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_link_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_link_1,.-nbif_ebif_link_1
	.type	nbif_ebif_link_1,@function
#endif

#ifndef HAVE_nbif_list_to_atom_1
#define HAVE_nbif_list_to_atom_1
	.section ".text"
	.align	4
	.global	nbif_list_to_atom_1
nbif_list_to_atom_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	list_to_atom_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_list_to_atom_1,.-nbif_list_to_atom_1
	.type	nbif_list_to_atom_1,@function
#endif

#ifndef HAVE_nbif_ebif_string_to_atom_1
#define HAVE_nbif_ebif_string_to_atom_1
	.section ".text"
	.align	4
	.global	nbif_ebif_string_to_atom_1
nbif_ebif_string_to_atom_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_string_to_atom_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_string_to_atom_1,.-nbif_ebif_string_to_atom_1
	.type	nbif_ebif_string_to_atom_1,@function
#endif

#ifndef HAVE_nbif_list_to_binary_1
#define HAVE_nbif_list_to_binary_1
	.section ".text"
	.align	4
	.global	nbif_list_to_binary_1
nbif_list_to_binary_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	list_to_binary_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_list_to_binary_1,.-nbif_list_to_binary_1
	.type	nbif_list_to_binary_1,@function
#endif

#ifndef HAVE_nbif_ebif_list_to_binary_1
#define HAVE_nbif_ebif_list_to_binary_1
	.section ".text"
	.align	4
	.global	nbif_ebif_list_to_binary_1
nbif_ebif_list_to_binary_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_list_to_binary_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_list_to_binary_1,.-nbif_ebif_list_to_binary_1
	.type	nbif_ebif_list_to_binary_1,@function
#endif

#ifndef HAVE_nbif_list_to_float_1
#define HAVE_nbif_list_to_float_1
	.section ".text"
	.align	4
	.global	nbif_list_to_float_1
nbif_list_to_float_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	list_to_float_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_list_to_float_1,.-nbif_list_to_float_1
	.type	nbif_list_to_float_1,@function
#endif

#ifndef HAVE_nbif_ebif_string_to_float_1
#define HAVE_nbif_ebif_string_to_float_1
	.section ".text"
	.align	4
	.global	nbif_ebif_string_to_float_1
nbif_ebif_string_to_float_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_string_to_float_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_string_to_float_1,.-nbif_ebif_string_to_float_1
	.type	nbif_ebif_string_to_float_1,@function
#endif

#ifndef HAVE_nbif_list_to_integer_1
#define HAVE_nbif_list_to_integer_1
	.section ".text"
	.align	4
	.global	nbif_list_to_integer_1
nbif_list_to_integer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	list_to_integer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_list_to_integer_1,.-nbif_list_to_integer_1
	.type	nbif_list_to_integer_1,@function
#endif

#ifndef HAVE_nbif_ebif_string_to_integer_1
#define HAVE_nbif_ebif_string_to_integer_1
	.section ".text"
	.align	4
	.global	nbif_ebif_string_to_integer_1
nbif_ebif_string_to_integer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_string_to_integer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_string_to_integer_1,.-nbif_ebif_string_to_integer_1
	.type	nbif_ebif_string_to_integer_1,@function
#endif

#ifndef HAVE_nbif_list_to_pid_1
#define HAVE_nbif_list_to_pid_1
	.section ".text"
	.align	4
	.global	nbif_list_to_pid_1
nbif_list_to_pid_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	list_to_pid_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_list_to_pid_1,.-nbif_list_to_pid_1
	.type	nbif_list_to_pid_1,@function
#endif

#ifndef HAVE_nbif_ebif_string_to_pid_1
#define HAVE_nbif_ebif_string_to_pid_1
	.section ".text"
	.align	4
	.global	nbif_ebif_string_to_pid_1
nbif_ebif_string_to_pid_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_string_to_pid_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_string_to_pid_1,.-nbif_ebif_string_to_pid_1
	.type	nbif_ebif_string_to_pid_1,@function
#endif

#ifndef HAVE_nbif_list_to_tuple_1
#define HAVE_nbif_list_to_tuple_1
	.section ".text"
	.align	4
	.global	nbif_list_to_tuple_1
nbif_list_to_tuple_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	list_to_tuple_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_list_to_tuple_1,.-nbif_list_to_tuple_1
	.type	nbif_list_to_tuple_1,@function
#endif

#ifndef HAVE_nbif_ebif_list_to_tuple_1
#define HAVE_nbif_ebif_list_to_tuple_1
	.section ".text"
	.align	4
	.global	nbif_ebif_list_to_tuple_1
nbif_ebif_list_to_tuple_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_list_to_tuple_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_list_to_tuple_1,.-nbif_ebif_list_to_tuple_1
	.type	nbif_ebif_list_to_tuple_1,@function
#endif

#ifndef HAVE_nbif_load_module_2
#define HAVE_nbif_load_module_2
	.section ".text"
	.align	4
	.global	nbif_load_module_2
nbif_load_module_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	load_module_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_load_module_2,.-nbif_load_module_2
	.type	nbif_load_module_2,@function
#endif

#ifndef HAVE_nbif_ebif_load_module_2
#define HAVE_nbif_ebif_load_module_2
	.section ".text"
	.align	4
	.global	nbif_ebif_load_module_2
nbif_ebif_load_module_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_load_module_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_load_module_2,.-nbif_ebif_load_module_2
	.type	nbif_ebif_load_module_2,@function
#endif

#ifndef HAVE_nbif_loaded_0
#define HAVE_nbif_loaded_0
	.section ".text"
	.align	4
	.global	nbif_loaded_0
nbif_loaded_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	loaded_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_loaded_0,.-nbif_loaded_0
	.type	nbif_loaded_0,@function
#endif

#ifndef HAVE_nbif_ebif_loaded_0
#define HAVE_nbif_ebif_loaded_0
	.section ".text"
	.align	4
	.global	nbif_ebif_loaded_0
nbif_ebif_loaded_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_loaded_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_loaded_0,.-nbif_ebif_loaded_0
	.type	nbif_ebif_loaded_0,@function
#endif

#ifndef HAVE_nbif_localtime_0
#define HAVE_nbif_localtime_0
	.section ".text"
	.align	4
	.global	nbif_localtime_0
nbif_localtime_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	localtime_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_localtime_0,.-nbif_localtime_0
	.type	nbif_localtime_0,@function
#endif

#ifndef HAVE_nbif_ebif_localtime_0
#define HAVE_nbif_ebif_localtime_0
	.section ".text"
	.align	4
	.global	nbif_ebif_localtime_0
nbif_ebif_localtime_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_localtime_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_localtime_0,.-nbif_ebif_localtime_0
	.type	nbif_ebif_localtime_0,@function
#endif

#ifndef HAVE_nbif_localtime_to_universaltime_2
#define HAVE_nbif_localtime_to_universaltime_2
	.section ".text"
	.align	4
	.global	nbif_localtime_to_universaltime_2
nbif_localtime_to_universaltime_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	localtime_to_universaltime_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_localtime_to_universaltime_2,.-nbif_localtime_to_universaltime_2
	.type	nbif_localtime_to_universaltime_2,@function
#endif

#ifndef HAVE_nbif_ebif_localtime_to_universaltime_2
#define HAVE_nbif_ebif_localtime_to_universaltime_2
	.section ".text"
	.align	4
	.global	nbif_ebif_localtime_to_universaltime_2
nbif_ebif_localtime_to_universaltime_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_localtime_to_universaltime_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_localtime_to_universaltime_2,.-nbif_ebif_localtime_to_universaltime_2
	.type	nbif_ebif_localtime_to_universaltime_2,@function
#endif

#ifndef HAVE_nbif_make_ref_0
#define HAVE_nbif_make_ref_0
	.section ".text"
	.align	4
	.global	nbif_make_ref_0
nbif_make_ref_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	make_ref_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_make_ref_0,.-nbif_make_ref_0
	.type	nbif_make_ref_0,@function
#endif

#ifndef HAVE_nbif_ebif_make_ref_0
#define HAVE_nbif_ebif_make_ref_0
	.section ".text"
	.align	4
	.global	nbif_ebif_make_ref_0
nbif_ebif_make_ref_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_make_ref_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_make_ref_0,.-nbif_ebif_make_ref_0
	.type	nbif_ebif_make_ref_0,@function
#endif

#ifndef HAVE_nbif_md5_1
#define HAVE_nbif_md5_1
	.section ".text"
	.align	4
	.global	nbif_md5_1
nbif_md5_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	md5_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_md5_1,.-nbif_md5_1
	.type	nbif_md5_1,@function
#endif

#ifndef HAVE_nbif_ebif_md5_1
#define HAVE_nbif_ebif_md5_1
	.section ".text"
	.align	4
	.global	nbif_ebif_md5_1
nbif_ebif_md5_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_md5_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_md5_1,.-nbif_ebif_md5_1
	.type	nbif_ebif_md5_1,@function
#endif

#ifndef HAVE_nbif_md5_init_0
#define HAVE_nbif_md5_init_0
	.section ".text"
	.align	4
	.global	nbif_md5_init_0
nbif_md5_init_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	md5_init_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_md5_init_0,.-nbif_md5_init_0
	.type	nbif_md5_init_0,@function
#endif

#ifndef HAVE_nbif_ebif_md5_init_0
#define HAVE_nbif_ebif_md5_init_0
	.section ".text"
	.align	4
	.global	nbif_ebif_md5_init_0
nbif_ebif_md5_init_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_md5_init_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_md5_init_0,.-nbif_ebif_md5_init_0
	.type	nbif_ebif_md5_init_0,@function
#endif

#ifndef HAVE_nbif_md5_update_2
#define HAVE_nbif_md5_update_2
	.section ".text"
	.align	4
	.global	nbif_md5_update_2
nbif_md5_update_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	md5_update_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_md5_update_2,.-nbif_md5_update_2
	.type	nbif_md5_update_2,@function
#endif

#ifndef HAVE_nbif_ebif_md5_update_2
#define HAVE_nbif_ebif_md5_update_2
	.section ".text"
	.align	4
	.global	nbif_ebif_md5_update_2
nbif_ebif_md5_update_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_md5_update_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_md5_update_2,.-nbif_ebif_md5_update_2
	.type	nbif_ebif_md5_update_2,@function
#endif

#ifndef HAVE_nbif_md5_final_1
#define HAVE_nbif_md5_final_1
	.section ".text"
	.align	4
	.global	nbif_md5_final_1
nbif_md5_final_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	md5_final_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_md5_final_1,.-nbif_md5_final_1
	.type	nbif_md5_final_1,@function
#endif

#ifndef HAVE_nbif_ebif_md5_final_1
#define HAVE_nbif_ebif_md5_final_1
	.section ".text"
	.align	4
	.global	nbif_ebif_md5_final_1
nbif_ebif_md5_final_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_md5_final_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_md5_final_1,.-nbif_ebif_md5_final_1
	.type	nbif_ebif_md5_final_1,@function
#endif

#ifndef HAVE_nbif_module_loaded_1
#define HAVE_nbif_module_loaded_1
	.section ".text"
	.align	4
	.global	nbif_module_loaded_1
nbif_module_loaded_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	module_loaded_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_module_loaded_1,.-nbif_module_loaded_1
	.type	nbif_module_loaded_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_loaded_1
#define HAVE_nbif_ebif_is_loaded_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_loaded_1
nbif_ebif_is_loaded_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_loaded_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_loaded_1,.-nbif_ebif_is_loaded_1
	.type	nbif_ebif_is_loaded_1,@function
#endif

#ifndef HAVE_nbif_function_exported_3
#define HAVE_nbif_function_exported_3
	.section ".text"
	.align	4
	.global	nbif_function_exported_3
nbif_function_exported_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	function_exported_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_function_exported_3,.-nbif_function_exported_3
	.type	nbif_function_exported_3,@function
#endif

#ifndef HAVE_nbif_ebif_is_loaded_3
#define HAVE_nbif_ebif_is_loaded_3
	.section ".text"
	.align	4
	.global	nbif_ebif_is_loaded_3
nbif_ebif_is_loaded_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_is_loaded_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_loaded_3,.-nbif_ebif_is_loaded_3
	.type	nbif_ebif_is_loaded_3,@function
#endif

#ifndef HAVE_nbif_monitor_node_2
#define HAVE_nbif_monitor_node_2
	.section ".text"
	.align	4
	.global	nbif_monitor_node_2
nbif_monitor_node_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	monitor_node_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_monitor_node_2,.-nbif_monitor_node_2
	.type	nbif_monitor_node_2,@function
#endif

#ifndef HAVE_nbif_ebif_monitor_node_2
#define HAVE_nbif_ebif_monitor_node_2
	.section ".text"
	.align	4
	.global	nbif_ebif_monitor_node_2
nbif_ebif_monitor_node_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_monitor_node_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_monitor_node_2,.-nbif_ebif_monitor_node_2
	.type	nbif_ebif_monitor_node_2,@function
#endif

#ifndef HAVE_nbif_node_1
#define HAVE_nbif_node_1
	.section ".text"
	.align	4
	.global	nbif_node_1
nbif_node_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	node_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_node_1,.-nbif_node_1
	.type	nbif_node_1,@function
#endif

#ifndef HAVE_nbif_ebif_node_1
#define HAVE_nbif_ebif_node_1
	.section ".text"
	.align	4
	.global	nbif_ebif_node_1
nbif_ebif_node_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_node_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_node_1,.-nbif_ebif_node_1
	.type	nbif_ebif_node_1,@function
#endif

#ifndef HAVE_nbif_node_0
#define HAVE_nbif_node_0
	.section ".text"
	.align	4
	.global	nbif_node_0
nbif_node_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	node_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_node_0,.-nbif_node_0
	.type	nbif_node_0,@function
#endif

#ifndef HAVE_nbif_ebif_node_0
#define HAVE_nbif_ebif_node_0
	.section ".text"
	.align	4
	.global	nbif_ebif_node_0
nbif_ebif_node_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_node_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_node_0,.-nbif_ebif_node_0
	.type	nbif_ebif_node_0,@function
#endif

#ifndef HAVE_nbif_nodes_1
#define HAVE_nbif_nodes_1
	.section ".text"
	.align	4
	.global	nbif_nodes_1
nbif_nodes_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	nodes_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_nodes_1,.-nbif_nodes_1
	.type	nbif_nodes_1,@function
#endif

#ifndef HAVE_nbif_ebif_nodes_1
#define HAVE_nbif_ebif_nodes_1
	.section ".text"
	.align	4
	.global	nbif_ebif_nodes_1
nbif_ebif_nodes_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_nodes_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_nodes_1,.-nbif_ebif_nodes_1
	.type	nbif_ebif_nodes_1,@function
#endif

#ifndef HAVE_nbif_now_0
#define HAVE_nbif_now_0
	.section ".text"
	.align	4
	.global	nbif_now_0
nbif_now_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	now_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_now_0,.-nbif_now_0
	.type	nbif_now_0,@function
#endif

#ifndef HAVE_nbif_ebif_now_0
#define HAVE_nbif_ebif_now_0
	.section ".text"
	.align	4
	.global	nbif_ebif_now_0
nbif_ebif_now_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_now_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_now_0,.-nbif_ebif_now_0
	.type	nbif_ebif_now_0,@function
#endif

#ifndef HAVE_nbif_open_port_2
#define HAVE_nbif_open_port_2
	.section ".text"
	.align	4
	.global	nbif_open_port_2
nbif_open_port_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	open_port_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_open_port_2,.-nbif_open_port_2
	.type	nbif_open_port_2,@function
#endif

#ifndef HAVE_nbif_ebif_open_port_2
#define HAVE_nbif_ebif_open_port_2
	.section ".text"
	.align	4
	.global	nbif_ebif_open_port_2
nbif_ebif_open_port_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_open_port_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_open_port_2,.-nbif_ebif_open_port_2
	.type	nbif_ebif_open_port_2,@function
#endif

#ifndef HAVE_nbif_pid_to_list_1
#define HAVE_nbif_pid_to_list_1
	.section ".text"
	.align	4
	.global	nbif_pid_to_list_1
nbif_pid_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	pid_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_pid_to_list_1,.-nbif_pid_to_list_1
	.type	nbif_pid_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_pid_to_string_1
#define HAVE_nbif_ebif_pid_to_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_pid_to_string_1
nbif_ebif_pid_to_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_pid_to_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_pid_to_string_1,.-nbif_ebif_pid_to_string_1
	.type	nbif_ebif_pid_to_string_1,@function
#endif

#ifndef HAVE_nbif_port_info_1
#define HAVE_nbif_port_info_1
	.section ".text"
	.align	4
	.global	nbif_port_info_1
nbif_port_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	port_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_port_info_1,.-nbif_port_info_1
	.type	nbif_port_info_1,@function
#endif

#ifndef HAVE_nbif_ebif_port_info_1
#define HAVE_nbif_ebif_port_info_1
	.section ".text"
	.align	4
	.global	nbif_ebif_port_info_1
nbif_ebif_port_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_port_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_info_1,.-nbif_ebif_port_info_1
	.type	nbif_ebif_port_info_1,@function
#endif

#ifndef HAVE_nbif_port_info_2
#define HAVE_nbif_port_info_2
	.section ".text"
	.align	4
	.global	nbif_port_info_2
nbif_port_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_port_info_2,.-nbif_port_info_2
	.type	nbif_port_info_2,@function
#endif

#ifndef HAVE_nbif_ebif_port_info_2
#define HAVE_nbif_ebif_port_info_2
	.section ".text"
	.align	4
	.global	nbif_ebif_port_info_2
nbif_ebif_port_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_port_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_info_2,.-nbif_ebif_port_info_2
	.type	nbif_ebif_port_info_2,@function
#endif

#ifndef HAVE_nbif_ports_0
#define HAVE_nbif_ports_0
	.section ".text"
	.align	4
	.global	nbif_ports_0
nbif_ports_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ports_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ports_0,.-nbif_ports_0
	.type	nbif_ports_0,@function
#endif

#ifndef HAVE_nbif_ebif_ports_0
#define HAVE_nbif_ebif_ports_0
	.section ".text"
	.align	4
	.global	nbif_ebif_ports_0
nbif_ebif_ports_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_ports_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_ports_0,.-nbif_ebif_ports_0
	.type	nbif_ebif_ports_0,@function
#endif

#ifndef HAVE_nbif_pre_loaded_0
#define HAVE_nbif_pre_loaded_0
	.section ".text"
	.align	4
	.global	nbif_pre_loaded_0
nbif_pre_loaded_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	pre_loaded_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_pre_loaded_0,.-nbif_pre_loaded_0
	.type	nbif_pre_loaded_0,@function
#endif

#ifndef HAVE_nbif_ebif_pre_loaded_0
#define HAVE_nbif_ebif_pre_loaded_0
	.section ".text"
	.align	4
	.global	nbif_ebif_pre_loaded_0
nbif_ebif_pre_loaded_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_pre_loaded_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_pre_loaded_0,.-nbif_ebif_pre_loaded_0
	.type	nbif_ebif_pre_loaded_0,@function
#endif

#ifndef HAVE_nbif_process_flag_2
#define HAVE_nbif_process_flag_2
	.section ".text"
	.align	4
	.global	nbif_process_flag_2
nbif_process_flag_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	process_flag_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_process_flag_2,.-nbif_process_flag_2
	.type	nbif_process_flag_2,@function
#endif

#ifndef HAVE_nbif_ebif_process_flag_2
#define HAVE_nbif_ebif_process_flag_2
	.section ".text"
	.align	4
	.global	nbif_ebif_process_flag_2
nbif_ebif_process_flag_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_process_flag_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_process_flag_2,.-nbif_ebif_process_flag_2
	.type	nbif_ebif_process_flag_2,@function
#endif

#ifndef HAVE_nbif_process_flag_3
#define HAVE_nbif_process_flag_3
	.section ".text"
	.align	4
	.global	nbif_process_flag_3
nbif_process_flag_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	process_flag_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_process_flag_3,.-nbif_process_flag_3
	.type	nbif_process_flag_3,@function
#endif

#ifndef HAVE_nbif_ebif_process_flag_3
#define HAVE_nbif_ebif_process_flag_3
	.section ".text"
	.align	4
	.global	nbif_ebif_process_flag_3
nbif_ebif_process_flag_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_process_flag_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_process_flag_3,.-nbif_ebif_process_flag_3
	.type	nbif_ebif_process_flag_3,@function
#endif

#ifndef HAVE_nbif_process_info_1
#define HAVE_nbif_process_info_1
	.section ".text"
	.align	4
	.global	nbif_process_info_1
nbif_process_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	process_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_process_info_1,.-nbif_process_info_1
	.type	nbif_process_info_1,@function
#endif

#ifndef HAVE_nbif_ebif_process_info_1
#define HAVE_nbif_ebif_process_info_1
	.section ".text"
	.align	4
	.global	nbif_ebif_process_info_1
nbif_ebif_process_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_process_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_process_info_1,.-nbif_ebif_process_info_1
	.type	nbif_ebif_process_info_1,@function
#endif

#ifndef HAVE_nbif_process_info_2
#define HAVE_nbif_process_info_2
	.section ".text"
	.align	4
	.global	nbif_process_info_2
nbif_process_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	process_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_process_info_2,.-nbif_process_info_2
	.type	nbif_process_info_2,@function
#endif

#ifndef HAVE_nbif_ebif_process_info_2
#define HAVE_nbif_ebif_process_info_2
	.section ".text"
	.align	4
	.global	nbif_ebif_process_info_2
nbif_ebif_process_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_process_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_process_info_2,.-nbif_ebif_process_info_2
	.type	nbif_ebif_process_info_2,@function
#endif

#ifndef HAVE_nbif_processes_0
#define HAVE_nbif_processes_0
	.section ".text"
	.align	4
	.global	nbif_processes_0
nbif_processes_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	processes_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_processes_0,.-nbif_processes_0
	.type	nbif_processes_0,@function
#endif

#ifndef HAVE_nbif_ebif_processes_0
#define HAVE_nbif_ebif_processes_0
	.section ".text"
	.align	4
	.global	nbif_ebif_processes_0
nbif_ebif_processes_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_processes_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_processes_0,.-nbif_ebif_processes_0
	.type	nbif_ebif_processes_0,@function
#endif

#ifndef HAVE_nbif_purge_module_1
#define HAVE_nbif_purge_module_1
	.section ".text"
	.align	4
	.global	nbif_purge_module_1
nbif_purge_module_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	purge_module_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_purge_module_1,.-nbif_purge_module_1
	.type	nbif_purge_module_1,@function
#endif

#ifndef HAVE_nbif_ebif_purge_module_1
#define HAVE_nbif_ebif_purge_module_1
	.section ".text"
	.align	4
	.global	nbif_ebif_purge_module_1
nbif_ebif_purge_module_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_purge_module_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_purge_module_1,.-nbif_ebif_purge_module_1
	.type	nbif_ebif_purge_module_1,@function
#endif

#ifndef HAVE_nbif_put_2
#define HAVE_nbif_put_2
	.section ".text"
	.align	4
	.global	nbif_put_2
nbif_put_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	put_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_put_2,.-nbif_put_2
	.type	nbif_put_2,@function
#endif

#ifndef HAVE_nbif_ebif_put_2
#define HAVE_nbif_ebif_put_2
	.section ".text"
	.align	4
	.global	nbif_ebif_put_2
nbif_ebif_put_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_put_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_put_2,.-nbif_ebif_put_2
	.type	nbif_ebif_put_2,@function
#endif

#ifndef HAVE_nbif_register_2
#define HAVE_nbif_register_2
	.section ".text"
	.align	4
	.global	nbif_register_2
nbif_register_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	register_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_register_2,.-nbif_register_2
	.type	nbif_register_2,@function
#endif

#ifndef HAVE_nbif_ebif_register_2
#define HAVE_nbif_ebif_register_2
	.section ".text"
	.align	4
	.global	nbif_ebif_register_2
nbif_ebif_register_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_register_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_register_2,.-nbif_ebif_register_2
	.type	nbif_ebif_register_2,@function
#endif

#ifndef HAVE_nbif_registered_0
#define HAVE_nbif_registered_0
	.section ".text"
	.align	4
	.global	nbif_registered_0
nbif_registered_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	registered_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_registered_0,.-nbif_registered_0
	.type	nbif_registered_0,@function
#endif

#ifndef HAVE_nbif_ebif_registered_0
#define HAVE_nbif_ebif_registered_0
	.section ".text"
	.align	4
	.global	nbif_ebif_registered_0
nbif_ebif_registered_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_registered_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_registered_0,.-nbif_ebif_registered_0
	.type	nbif_ebif_registered_0,@function
#endif

#ifndef HAVE_nbif_round_1
#define HAVE_nbif_round_1
	.section ".text"
	.align	4
	.global	nbif_round_1
nbif_round_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	round_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_round_1,.-nbif_round_1
	.type	nbif_round_1,@function
#endif

#ifndef HAVE_nbif_ebif_round_1
#define HAVE_nbif_ebif_round_1
	.section ".text"
	.align	4
	.global	nbif_ebif_round_1
nbif_ebif_round_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_round_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_round_1,.-nbif_ebif_round_1
	.type	nbif_ebif_round_1,@function
#endif

#ifndef HAVE_nbif_self_0
#define HAVE_nbif_self_0
	.section ".text"
	.align	4
	.global	nbif_self_0
nbif_self_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	self_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_self_0,.-nbif_self_0
	.type	nbif_self_0,@function
#endif

#ifndef HAVE_nbif_ebif_self_0
#define HAVE_nbif_ebif_self_0
	.section ".text"
	.align	4
	.global	nbif_ebif_self_0
nbif_ebif_self_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_self_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_self_0,.-nbif_ebif_self_0
	.type	nbif_ebif_self_0,@function
#endif

#ifndef HAVE_nbif_setelement_3
#define HAVE_nbif_setelement_3
	.section ".text"
	.align	4
	.global	nbif_setelement_3
nbif_setelement_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	setelement_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_setelement_3,.-nbif_setelement_3
	.type	nbif_setelement_3,@function
#endif

#ifndef HAVE_nbif_ebif_setelement_3
#define HAVE_nbif_ebif_setelement_3
	.section ".text"
	.align	4
	.global	nbif_ebif_setelement_3
nbif_ebif_setelement_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_setelement_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_setelement_3,.-nbif_ebif_setelement_3
	.type	nbif_ebif_setelement_3,@function
#endif

#ifndef HAVE_nbif_size_1
#define HAVE_nbif_size_1
	.section ".text"
	.align	4
	.global	nbif_size_1
nbif_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_size_1,.-nbif_size_1
	.type	nbif_size_1,@function
#endif

#ifndef HAVE_nbif_ebif_size_1
#define HAVE_nbif_ebif_size_1
	.section ".text"
	.align	4
	.global	nbif_ebif_size_1
nbif_ebif_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_size_1,.-nbif_ebif_size_1
	.type	nbif_ebif_size_1,@function
#endif

#ifndef HAVE_nbif_spawn_3
#define HAVE_nbif_spawn_3
	.section ".text"
	.align	4
	.global	nbif_spawn_3
nbif_spawn_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	spawn_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_spawn_3,.-nbif_spawn_3
	.type	nbif_spawn_3,@function
#endif

#ifndef HAVE_nbif_ebif_spawn_3
#define HAVE_nbif_ebif_spawn_3
	.section ".text"
	.align	4
	.global	nbif_ebif_spawn_3
nbif_ebif_spawn_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_spawn_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_spawn_3,.-nbif_ebif_spawn_3
	.type	nbif_ebif_spawn_3,@function
#endif

#ifndef HAVE_nbif_spawn_link_3
#define HAVE_nbif_spawn_link_3
	.section ".text"
	.align	4
	.global	nbif_spawn_link_3
nbif_spawn_link_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	spawn_link_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_spawn_link_3,.-nbif_spawn_link_3
	.type	nbif_spawn_link_3,@function
#endif

#ifndef HAVE_nbif_ebif_spawn_link_3
#define HAVE_nbif_ebif_spawn_link_3
	.section ".text"
	.align	4
	.global	nbif_ebif_spawn_link_3
nbif_ebif_spawn_link_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_spawn_link_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_spawn_link_3,.-nbif_ebif_spawn_link_3
	.type	nbif_ebif_spawn_link_3,@function
#endif

#ifndef HAVE_nbif_split_binary_2
#define HAVE_nbif_split_binary_2
	.section ".text"
	.align	4
	.global	nbif_split_binary_2
nbif_split_binary_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	split_binary_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_split_binary_2,.-nbif_split_binary_2
	.type	nbif_split_binary_2,@function
#endif

#ifndef HAVE_nbif_ebif_split_binary_2
#define HAVE_nbif_ebif_split_binary_2
	.section ".text"
	.align	4
	.global	nbif_ebif_split_binary_2
nbif_ebif_split_binary_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_split_binary_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_split_binary_2,.-nbif_ebif_split_binary_2
	.type	nbif_ebif_split_binary_2,@function
#endif

#ifndef HAVE_nbif_statistics_1
#define HAVE_nbif_statistics_1
	.section ".text"
	.align	4
	.global	nbif_statistics_1
nbif_statistics_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	statistics_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_statistics_1,.-nbif_statistics_1
	.type	nbif_statistics_1,@function
#endif

#ifndef HAVE_nbif_ebif_statistics_1
#define HAVE_nbif_ebif_statistics_1
	.section ".text"
	.align	4
	.global	nbif_ebif_statistics_1
nbif_ebif_statistics_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_statistics_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_statistics_1,.-nbif_ebif_statistics_1
	.type	nbif_ebif_statistics_1,@function
#endif

#ifndef HAVE_nbif_term_to_binary_1
#define HAVE_nbif_term_to_binary_1
	.section ".text"
	.align	4
	.global	nbif_term_to_binary_1
nbif_term_to_binary_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	term_to_binary_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_term_to_binary_1,.-nbif_term_to_binary_1
	.type	nbif_term_to_binary_1,@function
#endif

#ifndef HAVE_nbif_ebif_term_to_binary_1
#define HAVE_nbif_ebif_term_to_binary_1
	.section ".text"
	.align	4
	.global	nbif_ebif_term_to_binary_1
nbif_ebif_term_to_binary_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_term_to_binary_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_term_to_binary_1,.-nbif_ebif_term_to_binary_1
	.type	nbif_ebif_term_to_binary_1,@function
#endif

#ifndef HAVE_nbif_term_to_binary_2
#define HAVE_nbif_term_to_binary_2
	.section ".text"
	.align	4
	.global	nbif_term_to_binary_2
nbif_term_to_binary_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	term_to_binary_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_term_to_binary_2,.-nbif_term_to_binary_2
	.type	nbif_term_to_binary_2,@function
#endif

#ifndef HAVE_nbif_ebif_term_to_binary_2
#define HAVE_nbif_ebif_term_to_binary_2
	.section ".text"
	.align	4
	.global	nbif_ebif_term_to_binary_2
nbif_ebif_term_to_binary_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_term_to_binary_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_term_to_binary_2,.-nbif_ebif_term_to_binary_2
	.type	nbif_ebif_term_to_binary_2,@function
#endif

#ifndef HAVE_nbif_throw_1
#define HAVE_nbif_throw_1
	.section ".text"
	.align	4
	.global	nbif_throw_1
nbif_throw_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	throw_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_throw_1,.-nbif_throw_1
	.type	nbif_throw_1,@function
#endif

#ifndef HAVE_nbif_ebif_throw_1
#define HAVE_nbif_ebif_throw_1
	.section ".text"
	.align	4
	.global	nbif_ebif_throw_1
nbif_ebif_throw_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_throw_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_throw_1,.-nbif_ebif_throw_1
	.type	nbif_ebif_throw_1,@function
#endif

#ifndef HAVE_nbif_time_0
#define HAVE_nbif_time_0
	.section ".text"
	.align	4
	.global	nbif_time_0
nbif_time_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	time_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_time_0,.-nbif_time_0
	.type	nbif_time_0,@function
#endif

#ifndef HAVE_nbif_ebif_time_0
#define HAVE_nbif_ebif_time_0
	.section ".text"
	.align	4
	.global	nbif_ebif_time_0
nbif_ebif_time_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_time_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_time_0,.-nbif_ebif_time_0
	.type	nbif_ebif_time_0,@function
#endif

#ifndef HAVE_nbif_tl_1
#define HAVE_nbif_tl_1
	.section ".text"
	.align	4
	.global	nbif_tl_1
nbif_tl_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	tl_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_tl_1,.-nbif_tl_1
	.type	nbif_tl_1,@function
#endif

#ifndef HAVE_nbif_ebif_tl_1
#define HAVE_nbif_ebif_tl_1
	.section ".text"
	.align	4
	.global	nbif_ebif_tl_1
nbif_ebif_tl_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_tl_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_tl_1,.-nbif_ebif_tl_1
	.type	nbif_ebif_tl_1,@function
#endif

#ifndef HAVE_nbif_trunc_1
#define HAVE_nbif_trunc_1
	.section ".text"
	.align	4
	.global	nbif_trunc_1
nbif_trunc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	trunc_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_trunc_1,.-nbif_trunc_1
	.type	nbif_trunc_1,@function
#endif

#ifndef HAVE_nbif_ebif_trunc_1
#define HAVE_nbif_ebif_trunc_1
	.section ".text"
	.align	4
	.global	nbif_ebif_trunc_1
nbif_ebif_trunc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_trunc_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_trunc_1,.-nbif_ebif_trunc_1
	.type	nbif_ebif_trunc_1,@function
#endif

#ifndef HAVE_nbif_tuple_to_list_1
#define HAVE_nbif_tuple_to_list_1
	.section ".text"
	.align	4
	.global	nbif_tuple_to_list_1
nbif_tuple_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	tuple_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_tuple_to_list_1,.-nbif_tuple_to_list_1
	.type	nbif_tuple_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_tuple_to_list_1
#define HAVE_nbif_ebif_tuple_to_list_1
	.section ".text"
	.align	4
	.global	nbif_ebif_tuple_to_list_1
nbif_ebif_tuple_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_tuple_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_tuple_to_list_1,.-nbif_ebif_tuple_to_list_1
	.type	nbif_ebif_tuple_to_list_1,@function
#endif

#ifndef HAVE_nbif_universaltime_0
#define HAVE_nbif_universaltime_0
	.section ".text"
	.align	4
	.global	nbif_universaltime_0
nbif_universaltime_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	universaltime_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_universaltime_0,.-nbif_universaltime_0
	.type	nbif_universaltime_0,@function
#endif

#ifndef HAVE_nbif_ebif_universaltime_0
#define HAVE_nbif_ebif_universaltime_0
	.section ".text"
	.align	4
	.global	nbif_ebif_universaltime_0
nbif_ebif_universaltime_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_universaltime_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_universaltime_0,.-nbif_ebif_universaltime_0
	.type	nbif_ebif_universaltime_0,@function
#endif

#ifndef HAVE_nbif_universaltime_to_localtime_1
#define HAVE_nbif_universaltime_to_localtime_1
	.section ".text"
	.align	4
	.global	nbif_universaltime_to_localtime_1
nbif_universaltime_to_localtime_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	universaltime_to_localtime_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_universaltime_to_localtime_1,.-nbif_universaltime_to_localtime_1
	.type	nbif_universaltime_to_localtime_1,@function
#endif

#ifndef HAVE_nbif_ebif_universaltime_to_localtime_1
#define HAVE_nbif_ebif_universaltime_to_localtime_1
	.section ".text"
	.align	4
	.global	nbif_ebif_universaltime_to_localtime_1
nbif_ebif_universaltime_to_localtime_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_universaltime_to_localtime_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_universaltime_to_localtime_1,.-nbif_ebif_universaltime_to_localtime_1
	.type	nbif_ebif_universaltime_to_localtime_1,@function
#endif

#ifndef HAVE_nbif_unlink_1
#define HAVE_nbif_unlink_1
	.section ".text"
	.align	4
	.global	nbif_unlink_1
nbif_unlink_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	unlink_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_unlink_1,.-nbif_unlink_1
	.type	nbif_unlink_1,@function
#endif

#ifndef HAVE_nbif_ebif_unlink_1
#define HAVE_nbif_ebif_unlink_1
	.section ".text"
	.align	4
	.global	nbif_ebif_unlink_1
nbif_ebif_unlink_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_unlink_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_unlink_1,.-nbif_ebif_unlink_1
	.type	nbif_ebif_unlink_1,@function
#endif

#ifndef HAVE_nbif_unregister_1
#define HAVE_nbif_unregister_1
	.section ".text"
	.align	4
	.global	nbif_unregister_1
nbif_unregister_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	unregister_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_unregister_1,.-nbif_unregister_1
	.type	nbif_unregister_1,@function
#endif

#ifndef HAVE_nbif_ebif_unregister_1
#define HAVE_nbif_ebif_unregister_1
	.section ".text"
	.align	4
	.global	nbif_ebif_unregister_1
nbif_ebif_unregister_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_unregister_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_unregister_1,.-nbif_ebif_unregister_1
	.type	nbif_ebif_unregister_1,@function
#endif

#ifndef HAVE_nbif_whereis_1
#define HAVE_nbif_whereis_1
	.section ".text"
	.align	4
	.global	nbif_whereis_1
nbif_whereis_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	whereis_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_whereis_1,.-nbif_whereis_1
	.type	nbif_whereis_1,@function
#endif

#ifndef HAVE_nbif_ebif_whereis_1
#define HAVE_nbif_ebif_whereis_1
	.section ".text"
	.align	4
	.global	nbif_ebif_whereis_1
nbif_ebif_whereis_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_whereis_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_whereis_1,.-nbif_ebif_whereis_1
	.type	nbif_ebif_whereis_1,@function
#endif

#ifndef HAVE_nbif_spawn_opt_1
#define HAVE_nbif_spawn_opt_1
	.section ".text"
	.align	4
	.global	nbif_spawn_opt_1
nbif_spawn_opt_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	spawn_opt_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_spawn_opt_1,.-nbif_spawn_opt_1
	.type	nbif_spawn_opt_1,@function
#endif

#ifndef HAVE_nbif_ebif_spawn_opt_1
#define HAVE_nbif_ebif_spawn_opt_1
	.section ".text"
	.align	4
	.global	nbif_ebif_spawn_opt_1
nbif_ebif_spawn_opt_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_spawn_opt_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_spawn_opt_1,.-nbif_ebif_spawn_opt_1
	.type	nbif_ebif_spawn_opt_1,@function
#endif

#ifndef HAVE_nbif_setnode_2
#define HAVE_nbif_setnode_2
	.section ".text"
	.align	4
	.global	nbif_setnode_2
nbif_setnode_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	setnode_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_setnode_2,.-nbif_setnode_2
	.type	nbif_setnode_2,@function
#endif

#ifndef HAVE_nbif_setnode_3
#define HAVE_nbif_setnode_3
	.section ".text"
	.align	4
	.global	nbif_setnode_3
nbif_setnode_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	setnode_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_setnode_3,.-nbif_setnode_3
	.type	nbif_setnode_3,@function
#endif

#ifndef HAVE_nbif_dist_exit_3
#define HAVE_nbif_dist_exit_3
	.section ".text"
	.align	4
	.global	nbif_dist_exit_3
nbif_dist_exit_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	dist_exit_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_dist_exit_3,.-nbif_dist_exit_3
	.type	nbif_dist_exit_3,@function
#endif

#ifndef HAVE_nbif_dist_unlink_2
#define HAVE_nbif_dist_unlink_2
	.section ".text"
	.align	4
	.global	nbif_dist_unlink_2
nbif_dist_unlink_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	dist_unlink_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_dist_unlink_2,.-nbif_dist_unlink_2
	.type	nbif_dist_unlink_2,@function
#endif

#ifndef HAVE_nbif_dist_link_2
#define HAVE_nbif_dist_link_2
	.section ".text"
	.align	4
	.global	nbif_dist_link_2
nbif_dist_link_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	dist_link_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_dist_link_2,.-nbif_dist_link_2
	.type	nbif_dist_link_2,@function
#endif

#ifndef HAVE_nbif_port_call_2
#define HAVE_nbif_port_call_2
	.section ".text"
	.align	4
	.global	nbif_port_call_2
nbif_port_call_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_call_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_port_call_2,.-nbif_port_call_2
	.type	nbif_port_call_2,@function
#endif

#ifndef HAVE_nbif_ebif_port_call_2
#define HAVE_nbif_ebif_port_call_2
	.section ".text"
	.align	4
	.global	nbif_ebif_port_call_2
nbif_ebif_port_call_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_port_call_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_call_2,.-nbif_ebif_port_call_2
	.type	nbif_ebif_port_call_2,@function
#endif

#ifndef HAVE_nbif_port_call_3
#define HAVE_nbif_port_call_3
	.section ".text"
	.align	4
	.global	nbif_port_call_3
nbif_port_call_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_call_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_port_call_3,.-nbif_port_call_3
	.type	nbif_port_call_3,@function
#endif

#ifndef HAVE_nbif_ebif_port_call_3
#define HAVE_nbif_ebif_port_call_3
	.section ".text"
	.align	4
	.global	nbif_ebif_port_call_3
nbif_ebif_port_call_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_port_call_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_call_3,.-nbif_ebif_port_call_3
	.type	nbif_ebif_port_call_3,@function
#endif

#ifndef HAVE_nbif_port_command_2
#define HAVE_nbif_port_command_2
	.section ".text"
	.align	4
	.global	nbif_port_command_2
nbif_port_command_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_command_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_port_command_2,.-nbif_port_command_2
	.type	nbif_port_command_2,@function
#endif

#ifndef HAVE_nbif_ebif_port_command_2
#define HAVE_nbif_ebif_port_command_2
	.section ".text"
	.align	4
	.global	nbif_ebif_port_command_2
nbif_ebif_port_command_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_port_command_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_command_2,.-nbif_ebif_port_command_2
	.type	nbif_ebif_port_command_2,@function
#endif

#ifndef HAVE_nbif_port_control_3
#define HAVE_nbif_port_control_3
	.section ".text"
	.align	4
	.global	nbif_port_control_3
nbif_port_control_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_control_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_port_control_3,.-nbif_port_control_3
	.type	nbif_port_control_3,@function
#endif

#ifndef HAVE_nbif_ebif_port_control_3
#define HAVE_nbif_ebif_port_control_3
	.section ".text"
	.align	4
	.global	nbif_ebif_port_control_3
nbif_ebif_port_control_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_port_control_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_control_3,.-nbif_ebif_port_control_3
	.type	nbif_ebif_port_control_3,@function
#endif

#ifndef HAVE_nbif_port_close_1
#define HAVE_nbif_port_close_1
	.section ".text"
	.align	4
	.global	nbif_port_close_1
nbif_port_close_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	port_close_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_port_close_1,.-nbif_port_close_1
	.type	nbif_port_close_1,@function
#endif

#ifndef HAVE_nbif_ebif_port_close_1
#define HAVE_nbif_ebif_port_close_1
	.section ".text"
	.align	4
	.global	nbif_ebif_port_close_1
nbif_ebif_port_close_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_port_close_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_close_1,.-nbif_ebif_port_close_1
	.type	nbif_ebif_port_close_1,@function
#endif

#ifndef HAVE_nbif_port_connect_2
#define HAVE_nbif_port_connect_2
	.section ".text"
	.align	4
	.global	nbif_port_connect_2
nbif_port_connect_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_connect_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_port_connect_2,.-nbif_port_connect_2
	.type	nbif_port_connect_2,@function
#endif

#ifndef HAVE_nbif_ebif_port_connect_2
#define HAVE_nbif_ebif_port_connect_2
	.section ".text"
	.align	4
	.global	nbif_ebif_port_connect_2
nbif_ebif_port_connect_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_port_connect_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_connect_2,.-nbif_ebif_port_connect_2
	.type	nbif_ebif_port_connect_2,@function
#endif

#ifndef HAVE_nbif_port_set_data_2
#define HAVE_nbif_port_set_data_2
	.section ".text"
	.align	4
	.global	nbif_port_set_data_2
nbif_port_set_data_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	port_set_data_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_port_set_data_2,.-nbif_port_set_data_2
	.type	nbif_port_set_data_2,@function
#endif

#ifndef HAVE_nbif_ebif_port_set_data_2
#define HAVE_nbif_ebif_port_set_data_2
	.section ".text"
	.align	4
	.global	nbif_ebif_port_set_data_2
nbif_ebif_port_set_data_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_port_set_data_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_set_data_2,.-nbif_ebif_port_set_data_2
	.type	nbif_ebif_port_set_data_2,@function
#endif

#ifndef HAVE_nbif_port_get_data_1
#define HAVE_nbif_port_get_data_1
	.section ".text"
	.align	4
	.global	nbif_port_get_data_1
nbif_port_get_data_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	port_get_data_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_port_get_data_1,.-nbif_port_get_data_1
	.type	nbif_port_get_data_1,@function
#endif

#ifndef HAVE_nbif_ebif_port_get_data_1
#define HAVE_nbif_ebif_port_get_data_1
	.section ".text"
	.align	4
	.global	nbif_ebif_port_get_data_1
nbif_ebif_port_get_data_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_port_get_data_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_get_data_1,.-nbif_ebif_port_get_data_1
	.type	nbif_ebif_port_get_data_1,@function
#endif

#ifndef HAVE_nbif_trace_pattern_2
#define HAVE_nbif_trace_pattern_2
	.section ".text"
	.align	4
	.global	nbif_trace_pattern_2
nbif_trace_pattern_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	trace_pattern_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_trace_pattern_2,.-nbif_trace_pattern_2
	.type	nbif_trace_pattern_2,@function
#endif

#ifndef HAVE_nbif_ebif_trace_pattern_2
#define HAVE_nbif_ebif_trace_pattern_2
	.section ".text"
	.align	4
	.global	nbif_ebif_trace_pattern_2
nbif_ebif_trace_pattern_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_trace_pattern_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_trace_pattern_2,.-nbif_ebif_trace_pattern_2
	.type	nbif_ebif_trace_pattern_2,@function
#endif

#ifndef HAVE_nbif_trace_pattern_3
#define HAVE_nbif_trace_pattern_3
	.section ".text"
	.align	4
	.global	nbif_trace_pattern_3
nbif_trace_pattern_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	trace_pattern_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_trace_pattern_3,.-nbif_trace_pattern_3
	.type	nbif_trace_pattern_3,@function
#endif

#ifndef HAVE_nbif_ebif_trace_pattern_3
#define HAVE_nbif_ebif_trace_pattern_3
	.section ".text"
	.align	4
	.global	nbif_ebif_trace_pattern_3
nbif_ebif_trace_pattern_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_trace_pattern_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_trace_pattern_3,.-nbif_ebif_trace_pattern_3
	.type	nbif_ebif_trace_pattern_3,@function
#endif

#ifndef HAVE_nbif_trace_3
#define HAVE_nbif_trace_3
	.section ".text"
	.align	4
	.global	nbif_trace_3
nbif_trace_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	trace_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_trace_3,.-nbif_trace_3
	.type	nbif_trace_3,@function
#endif

#ifndef HAVE_nbif_ebif_trace_3
#define HAVE_nbif_ebif_trace_3
	.section ".text"
	.align	4
	.global	nbif_ebif_trace_3
nbif_ebif_trace_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_trace_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_trace_3,.-nbif_ebif_trace_3
	.type	nbif_ebif_trace_3,@function
#endif

#ifndef HAVE_nbif_trace_info_2
#define HAVE_nbif_trace_info_2
	.section ".text"
	.align	4
	.global	nbif_trace_info_2
nbif_trace_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	trace_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_trace_info_2,.-nbif_trace_info_2
	.type	nbif_trace_info_2,@function
#endif

#ifndef HAVE_nbif_ebif_trace_info_2
#define HAVE_nbif_ebif_trace_info_2
	.section ".text"
	.align	4
	.global	nbif_ebif_trace_info_2
nbif_ebif_trace_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_trace_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_trace_info_2,.-nbif_ebif_trace_info_2
	.type	nbif_ebif_trace_info_2,@function
#endif

#ifndef HAVE_nbif_seq_trace_2
#define HAVE_nbif_seq_trace_2
	.section ".text"
	.align	4
	.global	nbif_seq_trace_2
nbif_seq_trace_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	seq_trace_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_seq_trace_2,.-nbif_seq_trace_2
	.type	nbif_seq_trace_2,@function
#endif

#ifndef HAVE_nbif_ebif_seq_trace_2
#define HAVE_nbif_ebif_seq_trace_2
	.section ".text"
	.align	4
	.global	nbif_ebif_seq_trace_2
nbif_ebif_seq_trace_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_seq_trace_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_seq_trace_2,.-nbif_ebif_seq_trace_2
	.type	nbif_ebif_seq_trace_2,@function
#endif

#ifndef HAVE_nbif_seq_trace_info_1
#define HAVE_nbif_seq_trace_info_1
	.section ".text"
	.align	4
	.global	nbif_seq_trace_info_1
nbif_seq_trace_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	seq_trace_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_seq_trace_info_1,.-nbif_seq_trace_info_1
	.type	nbif_seq_trace_info_1,@function
#endif

#ifndef HAVE_nbif_ebif_seq_trace_info_1
#define HAVE_nbif_ebif_seq_trace_info_1
	.section ".text"
	.align	4
	.global	nbif_ebif_seq_trace_info_1
nbif_ebif_seq_trace_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_seq_trace_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_seq_trace_info_1,.-nbif_ebif_seq_trace_info_1
	.type	nbif_ebif_seq_trace_info_1,@function
#endif

#ifndef HAVE_nbif_seq_trace_print_1
#define HAVE_nbif_seq_trace_print_1
	.section ".text"
	.align	4
	.global	nbif_seq_trace_print_1
nbif_seq_trace_print_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	seq_trace_print_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_seq_trace_print_1,.-nbif_seq_trace_print_1
	.type	nbif_seq_trace_print_1,@function
#endif

#ifndef HAVE_nbif_ebif_seq_trace_print_1
#define HAVE_nbif_ebif_seq_trace_print_1
	.section ".text"
	.align	4
	.global	nbif_ebif_seq_trace_print_1
nbif_ebif_seq_trace_print_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_seq_trace_print_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_seq_trace_print_1,.-nbif_ebif_seq_trace_print_1
	.type	nbif_ebif_seq_trace_print_1,@function
#endif

#ifndef HAVE_nbif_seq_trace_print_2
#define HAVE_nbif_seq_trace_print_2
	.section ".text"
	.align	4
	.global	nbif_seq_trace_print_2
nbif_seq_trace_print_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	seq_trace_print_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_seq_trace_print_2,.-nbif_seq_trace_print_2
	.type	nbif_seq_trace_print_2,@function
#endif

#ifndef HAVE_nbif_ebif_seq_trace_print_2
#define HAVE_nbif_ebif_seq_trace_print_2
	.section ".text"
	.align	4
	.global	nbif_ebif_seq_trace_print_2
nbif_ebif_seq_trace_print_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_seq_trace_print_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_seq_trace_print_2,.-nbif_ebif_seq_trace_print_2
	.type	nbif_ebif_seq_trace_print_2,@function
#endif

#ifndef HAVE_nbif_suspend_process_1
#define HAVE_nbif_suspend_process_1
	.section ".text"
	.align	4
	.global	nbif_suspend_process_1
nbif_suspend_process_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	suspend_process_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_suspend_process_1,.-nbif_suspend_process_1
	.type	nbif_suspend_process_1,@function
#endif

#ifndef HAVE_nbif_ebif_suspend_process_1
#define HAVE_nbif_ebif_suspend_process_1
	.section ".text"
	.align	4
	.global	nbif_ebif_suspend_process_1
nbif_ebif_suspend_process_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_suspend_process_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_suspend_process_1,.-nbif_ebif_suspend_process_1
	.type	nbif_ebif_suspend_process_1,@function
#endif

#ifndef HAVE_nbif_resume_process_1
#define HAVE_nbif_resume_process_1
	.section ".text"
	.align	4
	.global	nbif_resume_process_1
nbif_resume_process_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	resume_process_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_resume_process_1,.-nbif_resume_process_1
	.type	nbif_resume_process_1,@function
#endif

#ifndef HAVE_nbif_ebif_resume_process_1
#define HAVE_nbif_ebif_resume_process_1
	.section ".text"
	.align	4
	.global	nbif_ebif_resume_process_1
nbif_ebif_resume_process_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_resume_process_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_resume_process_1,.-nbif_ebif_resume_process_1
	.type	nbif_ebif_resume_process_1,@function
#endif

#ifndef HAVE_nbif_process_display_2
#define HAVE_nbif_process_display_2
	.section ".text"
	.align	4
	.global	nbif_process_display_2
nbif_process_display_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	process_display_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_process_display_2,.-nbif_process_display_2
	.type	nbif_process_display_2,@function
#endif

#ifndef HAVE_nbif_ebif_process_display_2
#define HAVE_nbif_ebif_process_display_2
	.section ".text"
	.align	4
	.global	nbif_ebif_process_display_2
nbif_ebif_process_display_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_process_display_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_process_display_2,.-nbif_ebif_process_display_2
	.type	nbif_ebif_process_display_2,@function
#endif

#ifndef HAVE_nbif_bump_reductions_1
#define HAVE_nbif_bump_reductions_1
	.section ".text"
	.align	4
	.global	nbif_bump_reductions_1
nbif_bump_reductions_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	bump_reductions_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_bump_reductions_1,.-nbif_bump_reductions_1
	.type	nbif_bump_reductions_1,@function
#endif

#ifndef HAVE_nbif_ebif_bump_reductions_1
#define HAVE_nbif_ebif_bump_reductions_1
	.section ".text"
	.align	4
	.global	nbif_ebif_bump_reductions_1
nbif_ebif_bump_reductions_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_bump_reductions_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_bump_reductions_1,.-nbif_ebif_bump_reductions_1
	.type	nbif_ebif_bump_reductions_1,@function
#endif

#ifndef HAVE_nbif_math_cos_1
#define HAVE_nbif_math_cos_1
	.section ".text"
	.align	4
	.global	nbif_math_cos_1
nbif_math_cos_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_cos_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_cos_1,.-nbif_math_cos_1
	.type	nbif_math_cos_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_cos_1
#define HAVE_nbif_ebif_math_cos_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_cos_1
nbif_ebif_math_cos_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_cos_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_cos_1,.-nbif_ebif_math_cos_1
	.type	nbif_ebif_math_cos_1,@function
#endif

#ifndef HAVE_nbif_math_cosh_1
#define HAVE_nbif_math_cosh_1
	.section ".text"
	.align	4
	.global	nbif_math_cosh_1
nbif_math_cosh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_cosh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_cosh_1,.-nbif_math_cosh_1
	.type	nbif_math_cosh_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_cosh_1
#define HAVE_nbif_ebif_math_cosh_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_cosh_1
nbif_ebif_math_cosh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_cosh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_cosh_1,.-nbif_ebif_math_cosh_1
	.type	nbif_ebif_math_cosh_1,@function
#endif

#ifndef HAVE_nbif_math_sin_1
#define HAVE_nbif_math_sin_1
	.section ".text"
	.align	4
	.global	nbif_math_sin_1
nbif_math_sin_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_sin_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_sin_1,.-nbif_math_sin_1
	.type	nbif_math_sin_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_sin_1
#define HAVE_nbif_ebif_math_sin_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_sin_1
nbif_ebif_math_sin_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_sin_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_sin_1,.-nbif_ebif_math_sin_1
	.type	nbif_ebif_math_sin_1,@function
#endif

#ifndef HAVE_nbif_math_sinh_1
#define HAVE_nbif_math_sinh_1
	.section ".text"
	.align	4
	.global	nbif_math_sinh_1
nbif_math_sinh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_sinh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_sinh_1,.-nbif_math_sinh_1
	.type	nbif_math_sinh_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_sinh_1
#define HAVE_nbif_ebif_math_sinh_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_sinh_1
nbif_ebif_math_sinh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_sinh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_sinh_1,.-nbif_ebif_math_sinh_1
	.type	nbif_ebif_math_sinh_1,@function
#endif

#ifndef HAVE_nbif_math_tan_1
#define HAVE_nbif_math_tan_1
	.section ".text"
	.align	4
	.global	nbif_math_tan_1
nbif_math_tan_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_tan_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_tan_1,.-nbif_math_tan_1
	.type	nbif_math_tan_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_tan_1
#define HAVE_nbif_ebif_math_tan_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_tan_1
nbif_ebif_math_tan_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_tan_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_tan_1,.-nbif_ebif_math_tan_1
	.type	nbif_ebif_math_tan_1,@function
#endif

#ifndef HAVE_nbif_math_tanh_1
#define HAVE_nbif_math_tanh_1
	.section ".text"
	.align	4
	.global	nbif_math_tanh_1
nbif_math_tanh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_tanh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_tanh_1,.-nbif_math_tanh_1
	.type	nbif_math_tanh_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_tanh_1
#define HAVE_nbif_ebif_math_tanh_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_tanh_1
nbif_ebif_math_tanh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_tanh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_tanh_1,.-nbif_ebif_math_tanh_1
	.type	nbif_ebif_math_tanh_1,@function
#endif

#ifndef HAVE_nbif_math_acos_1
#define HAVE_nbif_math_acos_1
	.section ".text"
	.align	4
	.global	nbif_math_acos_1
nbif_math_acos_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_acos_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_acos_1,.-nbif_math_acos_1
	.type	nbif_math_acos_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_acos_1
#define HAVE_nbif_ebif_math_acos_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_acos_1
nbif_ebif_math_acos_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_acos_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_acos_1,.-nbif_ebif_math_acos_1
	.type	nbif_ebif_math_acos_1,@function
#endif

#ifndef HAVE_nbif_math_acosh_1
#define HAVE_nbif_math_acosh_1
	.section ".text"
	.align	4
	.global	nbif_math_acosh_1
nbif_math_acosh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_acosh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_acosh_1,.-nbif_math_acosh_1
	.type	nbif_math_acosh_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_acosh_1
#define HAVE_nbif_ebif_math_acosh_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_acosh_1
nbif_ebif_math_acosh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_acosh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_acosh_1,.-nbif_ebif_math_acosh_1
	.type	nbif_ebif_math_acosh_1,@function
#endif

#ifndef HAVE_nbif_math_asin_1
#define HAVE_nbif_math_asin_1
	.section ".text"
	.align	4
	.global	nbif_math_asin_1
nbif_math_asin_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_asin_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_asin_1,.-nbif_math_asin_1
	.type	nbif_math_asin_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_asin_1
#define HAVE_nbif_ebif_math_asin_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_asin_1
nbif_ebif_math_asin_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_asin_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_asin_1,.-nbif_ebif_math_asin_1
	.type	nbif_ebif_math_asin_1,@function
#endif

#ifndef HAVE_nbif_math_asinh_1
#define HAVE_nbif_math_asinh_1
	.section ".text"
	.align	4
	.global	nbif_math_asinh_1
nbif_math_asinh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_asinh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_asinh_1,.-nbif_math_asinh_1
	.type	nbif_math_asinh_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_asinh_1
#define HAVE_nbif_ebif_math_asinh_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_asinh_1
nbif_ebif_math_asinh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_asinh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_asinh_1,.-nbif_ebif_math_asinh_1
	.type	nbif_ebif_math_asinh_1,@function
#endif

#ifndef HAVE_nbif_math_atan_1
#define HAVE_nbif_math_atan_1
	.section ".text"
	.align	4
	.global	nbif_math_atan_1
nbif_math_atan_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_atan_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_atan_1,.-nbif_math_atan_1
	.type	nbif_math_atan_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_atan_1
#define HAVE_nbif_ebif_math_atan_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_atan_1
nbif_ebif_math_atan_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_atan_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_atan_1,.-nbif_ebif_math_atan_1
	.type	nbif_ebif_math_atan_1,@function
#endif

#ifndef HAVE_nbif_math_atanh_1
#define HAVE_nbif_math_atanh_1
	.section ".text"
	.align	4
	.global	nbif_math_atanh_1
nbif_math_atanh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_atanh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_atanh_1,.-nbif_math_atanh_1
	.type	nbif_math_atanh_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_atanh_1
#define HAVE_nbif_ebif_math_atanh_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_atanh_1
nbif_ebif_math_atanh_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_atanh_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_atanh_1,.-nbif_ebif_math_atanh_1
	.type	nbif_ebif_math_atanh_1,@function
#endif

#ifndef HAVE_nbif_math_erf_1
#define HAVE_nbif_math_erf_1
	.section ".text"
	.align	4
	.global	nbif_math_erf_1
nbif_math_erf_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_erf_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_erf_1,.-nbif_math_erf_1
	.type	nbif_math_erf_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_erf_1
#define HAVE_nbif_ebif_math_erf_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_erf_1
nbif_ebif_math_erf_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_erf_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_erf_1,.-nbif_ebif_math_erf_1
	.type	nbif_ebif_math_erf_1,@function
#endif

#ifndef HAVE_nbif_math_erfc_1
#define HAVE_nbif_math_erfc_1
	.section ".text"
	.align	4
	.global	nbif_math_erfc_1
nbif_math_erfc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_erfc_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_erfc_1,.-nbif_math_erfc_1
	.type	nbif_math_erfc_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_erfc_1
#define HAVE_nbif_ebif_math_erfc_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_erfc_1
nbif_ebif_math_erfc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_erfc_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_erfc_1,.-nbif_ebif_math_erfc_1
	.type	nbif_ebif_math_erfc_1,@function
#endif

#ifndef HAVE_nbif_math_exp_1
#define HAVE_nbif_math_exp_1
	.section ".text"
	.align	4
	.global	nbif_math_exp_1
nbif_math_exp_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_exp_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_exp_1,.-nbif_math_exp_1
	.type	nbif_math_exp_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_exp_1
#define HAVE_nbif_ebif_math_exp_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_exp_1
nbif_ebif_math_exp_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_exp_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_exp_1,.-nbif_ebif_math_exp_1
	.type	nbif_ebif_math_exp_1,@function
#endif

#ifndef HAVE_nbif_math_log_1
#define HAVE_nbif_math_log_1
	.section ".text"
	.align	4
	.global	nbif_math_log_1
nbif_math_log_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_log_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_log_1,.-nbif_math_log_1
	.type	nbif_math_log_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_log_1
#define HAVE_nbif_ebif_math_log_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_log_1
nbif_ebif_math_log_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_log_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_log_1,.-nbif_ebif_math_log_1
	.type	nbif_ebif_math_log_1,@function
#endif

#ifndef HAVE_nbif_math_log10_1
#define HAVE_nbif_math_log10_1
	.section ".text"
	.align	4
	.global	nbif_math_log10_1
nbif_math_log10_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_log10_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_log10_1,.-nbif_math_log10_1
	.type	nbif_math_log10_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_log10_1
#define HAVE_nbif_ebif_math_log10_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_log10_1
nbif_ebif_math_log10_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_log10_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_log10_1,.-nbif_ebif_math_log10_1
	.type	nbif_ebif_math_log10_1,@function
#endif

#ifndef HAVE_nbif_math_sqrt_1
#define HAVE_nbif_math_sqrt_1
	.section ".text"
	.align	4
	.global	nbif_math_sqrt_1
nbif_math_sqrt_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	math_sqrt_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_math_sqrt_1,.-nbif_math_sqrt_1
	.type	nbif_math_sqrt_1,@function
#endif

#ifndef HAVE_nbif_ebif_math_sqrt_1
#define HAVE_nbif_ebif_math_sqrt_1
	.section ".text"
	.align	4
	.global	nbif_ebif_math_sqrt_1
nbif_ebif_math_sqrt_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_math_sqrt_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_sqrt_1,.-nbif_ebif_math_sqrt_1
	.type	nbif_ebif_math_sqrt_1,@function
#endif

#ifndef HAVE_nbif_math_atan2_2
#define HAVE_nbif_math_atan2_2
	.section ".text"
	.align	4
	.global	nbif_math_atan2_2
nbif_math_atan2_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	math_atan2_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_math_atan2_2,.-nbif_math_atan2_2
	.type	nbif_math_atan2_2,@function
#endif

#ifndef HAVE_nbif_ebif_math_atan2_2
#define HAVE_nbif_ebif_math_atan2_2
	.section ".text"
	.align	4
	.global	nbif_ebif_math_atan2_2
nbif_ebif_math_atan2_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_math_atan2_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_atan2_2,.-nbif_ebif_math_atan2_2
	.type	nbif_ebif_math_atan2_2,@function
#endif

#ifndef HAVE_nbif_math_pow_2
#define HAVE_nbif_math_pow_2
	.section ".text"
	.align	4
	.global	nbif_math_pow_2
nbif_math_pow_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	math_pow_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_math_pow_2,.-nbif_math_pow_2
	.type	nbif_math_pow_2,@function
#endif

#ifndef HAVE_nbif_ebif_math_pow_2
#define HAVE_nbif_ebif_math_pow_2
	.section ".text"
	.align	4
	.global	nbif_ebif_math_pow_2
nbif_ebif_math_pow_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_math_pow_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_math_pow_2,.-nbif_ebif_math_pow_2
	.type	nbif_ebif_math_pow_2,@function
#endif

#ifndef HAVE_nbif_start_timer_3
#define HAVE_nbif_start_timer_3
	.section ".text"
	.align	4
	.global	nbif_start_timer_3
nbif_start_timer_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	start_timer_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_start_timer_3,.-nbif_start_timer_3
	.type	nbif_start_timer_3,@function
#endif

#ifndef HAVE_nbif_ebif_start_timer_3
#define HAVE_nbif_ebif_start_timer_3
	.section ".text"
	.align	4
	.global	nbif_ebif_start_timer_3
nbif_ebif_start_timer_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_start_timer_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_start_timer_3,.-nbif_ebif_start_timer_3
	.type	nbif_ebif_start_timer_3,@function
#endif

#ifndef HAVE_nbif_send_after_3
#define HAVE_nbif_send_after_3
	.section ".text"
	.align	4
	.global	nbif_send_after_3
nbif_send_after_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	send_after_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_send_after_3,.-nbif_send_after_3
	.type	nbif_send_after_3,@function
#endif

#ifndef HAVE_nbif_ebif_send_after_3
#define HAVE_nbif_ebif_send_after_3
	.section ".text"
	.align	4
	.global	nbif_ebif_send_after_3
nbif_ebif_send_after_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_send_after_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_send_after_3,.-nbif_ebif_send_after_3
	.type	nbif_ebif_send_after_3,@function
#endif

#ifndef HAVE_nbif_cancel_timer_1
#define HAVE_nbif_cancel_timer_1
	.section ".text"
	.align	4
	.global	nbif_cancel_timer_1
nbif_cancel_timer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	cancel_timer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_cancel_timer_1,.-nbif_cancel_timer_1
	.type	nbif_cancel_timer_1,@function
#endif

#ifndef HAVE_nbif_ebif_cancel_timer_1
#define HAVE_nbif_ebif_cancel_timer_1
	.section ".text"
	.align	4
	.global	nbif_ebif_cancel_timer_1
nbif_ebif_cancel_timer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_cancel_timer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_cancel_timer_1,.-nbif_ebif_cancel_timer_1
	.type	nbif_ebif_cancel_timer_1,@function
#endif

#ifndef HAVE_nbif_read_timer_1
#define HAVE_nbif_read_timer_1
	.section ".text"
	.align	4
	.global	nbif_read_timer_1
nbif_read_timer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	read_timer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_read_timer_1,.-nbif_read_timer_1
	.type	nbif_read_timer_1,@function
#endif

#ifndef HAVE_nbif_ebif_read_timer_1
#define HAVE_nbif_ebif_read_timer_1
	.section ".text"
	.align	4
	.global	nbif_ebif_read_timer_1
nbif_ebif_read_timer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_read_timer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_read_timer_1,.-nbif_ebif_read_timer_1
	.type	nbif_ebif_read_timer_1,@function
#endif

#ifndef HAVE_nbif_make_tuple_2
#define HAVE_nbif_make_tuple_2
	.section ".text"
	.align	4
	.global	nbif_make_tuple_2
nbif_make_tuple_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	make_tuple_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_make_tuple_2,.-nbif_make_tuple_2
	.type	nbif_make_tuple_2,@function
#endif

#ifndef HAVE_nbif_ebif_make_tuple_2
#define HAVE_nbif_ebif_make_tuple_2
	.section ".text"
	.align	4
	.global	nbif_ebif_make_tuple_2
nbif_ebif_make_tuple_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_make_tuple_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_make_tuple_2,.-nbif_ebif_make_tuple_2
	.type	nbif_ebif_make_tuple_2,@function
#endif

#ifndef HAVE_nbif_append_element_2
#define HAVE_nbif_append_element_2
	.section ".text"
	.align	4
	.global	nbif_append_element_2
nbif_append_element_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	append_element_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_append_element_2,.-nbif_append_element_2
	.type	nbif_append_element_2,@function
#endif

#ifndef HAVE_nbif_ebif_append_element_2
#define HAVE_nbif_ebif_append_element_2
	.section ".text"
	.align	4
	.global	nbif_ebif_append_element_2
nbif_ebif_append_element_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_append_element_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_append_element_2,.-nbif_ebif_append_element_2
	.type	nbif_ebif_append_element_2,@function
#endif

#ifndef HAVE_nbif_system_flag_2
#define HAVE_nbif_system_flag_2
	.section ".text"
	.align	4
	.global	nbif_system_flag_2
nbif_system_flag_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	system_flag_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_system_flag_2,.-nbif_system_flag_2
	.type	nbif_system_flag_2,@function
#endif

#ifndef HAVE_nbif_ebif_system_flag_2
#define HAVE_nbif_ebif_system_flag_2
	.section ".text"
	.align	4
	.global	nbif_ebif_system_flag_2
nbif_ebif_system_flag_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_system_flag_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_system_flag_2,.-nbif_ebif_system_flag_2
	.type	nbif_ebif_system_flag_2,@function
#endif

#ifndef HAVE_nbif_system_info_1
#define HAVE_nbif_system_info_1
	.section ".text"
	.align	4
	.global	nbif_system_info_1
nbif_system_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	system_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_system_info_1,.-nbif_system_info_1
	.type	nbif_system_info_1,@function
#endif

#ifndef HAVE_nbif_ebif_system_info_1
#define HAVE_nbif_ebif_system_info_1
	.section ".text"
	.align	4
	.global	nbif_ebif_system_info_1
nbif_ebif_system_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_system_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_system_info_1,.-nbif_ebif_system_info_1
	.type	nbif_ebif_system_info_1,@function
#endif

#ifndef HAVE_nbif_system_monitor_0
#define HAVE_nbif_system_monitor_0
	.section ".text"
	.align	4
	.global	nbif_system_monitor_0
nbif_system_monitor_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	system_monitor_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_system_monitor_0,.-nbif_system_monitor_0
	.type	nbif_system_monitor_0,@function
#endif

#ifndef HAVE_nbif_ebif_system_monitor_0
#define HAVE_nbif_ebif_system_monitor_0
	.section ".text"
	.align	4
	.global	nbif_ebif_system_monitor_0
nbif_ebif_system_monitor_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_system_monitor_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_system_monitor_0,.-nbif_ebif_system_monitor_0
	.type	nbif_ebif_system_monitor_0,@function
#endif

#ifndef HAVE_nbif_system_monitor_1
#define HAVE_nbif_system_monitor_1
	.section ".text"
	.align	4
	.global	nbif_system_monitor_1
nbif_system_monitor_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	system_monitor_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_system_monitor_1,.-nbif_system_monitor_1
	.type	nbif_system_monitor_1,@function
#endif

#ifndef HAVE_nbif_ebif_system_monitor_1
#define HAVE_nbif_ebif_system_monitor_1
	.section ".text"
	.align	4
	.global	nbif_ebif_system_monitor_1
nbif_ebif_system_monitor_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_system_monitor_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_system_monitor_1,.-nbif_ebif_system_monitor_1
	.type	nbif_ebif_system_monitor_1,@function
#endif

#ifndef HAVE_nbif_system_monitor_2
#define HAVE_nbif_system_monitor_2
	.section ".text"
	.align	4
	.global	nbif_system_monitor_2
nbif_system_monitor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	system_monitor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_system_monitor_2,.-nbif_system_monitor_2
	.type	nbif_system_monitor_2,@function
#endif

#ifndef HAVE_nbif_ebif_system_monitor_2
#define HAVE_nbif_ebif_system_monitor_2
	.section ".text"
	.align	4
	.global	nbif_ebif_system_monitor_2
nbif_ebif_system_monitor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_system_monitor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_system_monitor_2,.-nbif_ebif_system_monitor_2
	.type	nbif_ebif_system_monitor_2,@function
#endif

#ifndef HAVE_nbif_ref_to_list_1
#define HAVE_nbif_ref_to_list_1
	.section ".text"
	.align	4
	.global	nbif_ref_to_list_1
nbif_ref_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ref_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ref_to_list_1,.-nbif_ref_to_list_1
	.type	nbif_ref_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_ref_to_string_1
#define HAVE_nbif_ebif_ref_to_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ref_to_string_1
nbif_ebif_ref_to_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ref_to_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ref_to_string_1,.-nbif_ebif_ref_to_string_1
	.type	nbif_ebif_ref_to_string_1,@function
#endif

#ifndef HAVE_nbif_port_to_list_1
#define HAVE_nbif_port_to_list_1
	.section ".text"
	.align	4
	.global	nbif_port_to_list_1
nbif_port_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	port_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_port_to_list_1,.-nbif_port_to_list_1
	.type	nbif_port_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_port_to_string_1
#define HAVE_nbif_ebif_port_to_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_port_to_string_1
nbif_ebif_port_to_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_port_to_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_port_to_string_1,.-nbif_ebif_port_to_string_1
	.type	nbif_ebif_port_to_string_1,@function
#endif

#ifndef HAVE_nbif_fun_to_list_1
#define HAVE_nbif_fun_to_list_1
	.section ".text"
	.align	4
	.global	nbif_fun_to_list_1
nbif_fun_to_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	fun_to_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_fun_to_list_1,.-nbif_fun_to_list_1
	.type	nbif_fun_to_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_fun_to_string_1
#define HAVE_nbif_ebif_fun_to_string_1
	.section ".text"
	.align	4
	.global	nbif_ebif_fun_to_string_1
nbif_ebif_fun_to_string_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_fun_to_string_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_fun_to_string_1,.-nbif_ebif_fun_to_string_1
	.type	nbif_ebif_fun_to_string_1,@function
#endif

#ifndef HAVE_nbif_monitor_2
#define HAVE_nbif_monitor_2
	.section ".text"
	.align	4
	.global	nbif_monitor_2
nbif_monitor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	monitor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_monitor_2,.-nbif_monitor_2
	.type	nbif_monitor_2,@function
#endif

#ifndef HAVE_nbif_ebif_monitor_2
#define HAVE_nbif_ebif_monitor_2
	.section ".text"
	.align	4
	.global	nbif_ebif_monitor_2
nbif_ebif_monitor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_monitor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_monitor_2,.-nbif_ebif_monitor_2
	.type	nbif_ebif_monitor_2,@function
#endif

#ifndef HAVE_nbif_demonitor_1
#define HAVE_nbif_demonitor_1
	.section ".text"
	.align	4
	.global	nbif_demonitor_1
nbif_demonitor_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	demonitor_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_demonitor_1,.-nbif_demonitor_1
	.type	nbif_demonitor_1,@function
#endif

#ifndef HAVE_nbif_ebif_demonitor_1
#define HAVE_nbif_ebif_demonitor_1
	.section ".text"
	.align	4
	.global	nbif_ebif_demonitor_1
nbif_ebif_demonitor_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_demonitor_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_demonitor_1,.-nbif_ebif_demonitor_1
	.type	nbif_ebif_demonitor_1,@function
#endif

#ifndef HAVE_nbif_is_process_alive_1
#define HAVE_nbif_is_process_alive_1
	.section ".text"
	.align	4
	.global	nbif_is_process_alive_1
nbif_is_process_alive_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_process_alive_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_process_alive_1,.-nbif_is_process_alive_1
	.type	nbif_is_process_alive_1,@function
#endif

#ifndef HAVE_nbif_ebif_proc_is_alive_1
#define HAVE_nbif_ebif_proc_is_alive_1
	.section ".text"
	.align	4
	.global	nbif_ebif_proc_is_alive_1
nbif_ebif_proc_is_alive_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_proc_is_alive_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_proc_is_alive_1,.-nbif_ebif_proc_is_alive_1
	.type	nbif_ebif_proc_is_alive_1,@function
#endif

#ifndef HAVE_nbif_error_1
#define HAVE_nbif_error_1
	.section ".text"
	.align	4
	.global	nbif_error_1
nbif_error_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	error_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_error_1,.-nbif_error_1
	.type	nbif_error_1,@function
#endif

#ifndef HAVE_nbif_fault_1
#define HAVE_nbif_fault_1
	.section ".text"
	.align	4
	.global	nbif_fault_1
nbif_fault_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	fault_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_fault_1,.-nbif_fault_1
	.type	nbif_fault_1,@function
#endif

#ifndef HAVE_nbif_ebif_error_1
#define HAVE_nbif_ebif_error_1
	.section ".text"
	.align	4
	.global	nbif_ebif_error_1
nbif_ebif_error_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_error_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_error_1,.-nbif_ebif_error_1
	.type	nbif_ebif_error_1,@function
#endif

#ifndef HAVE_nbif_error_2
#define HAVE_nbif_error_2
	.section ".text"
	.align	4
	.global	nbif_error_2
nbif_error_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	error_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_error_2,.-nbif_error_2
	.type	nbif_error_2,@function
#endif

#ifndef HAVE_nbif_fault_2
#define HAVE_nbif_fault_2
	.section ".text"
	.align	4
	.global	nbif_fault_2
nbif_fault_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	fault_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_fault_2,.-nbif_fault_2
	.type	nbif_fault_2,@function
#endif

#ifndef HAVE_nbif_ebif_error_2
#define HAVE_nbif_ebif_error_2
	.section ".text"
	.align	4
	.global	nbif_ebif_error_2
nbif_ebif_error_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_error_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_error_2,.-nbif_ebif_error_2
	.type	nbif_ebif_error_2,@function
#endif

#ifndef HAVE_nbif_raise_3
#define HAVE_nbif_raise_3
	.section ".text"
	.align	4
	.global	nbif_raise_3
nbif_raise_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	raise_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_raise_3,.-nbif_raise_3
	.type	nbif_raise_3,@function
#endif

#ifndef HAVE_nbif_ebif_raise_3
#define HAVE_nbif_ebif_raise_3
	.section ".text"
	.align	4
	.global	nbif_ebif_raise_3
nbif_ebif_raise_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_raise_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_raise_3,.-nbif_ebif_raise_3
	.type	nbif_ebif_raise_3,@function
#endif

#ifndef HAVE_nbif_get_stacktrace_0
#define HAVE_nbif_get_stacktrace_0
	.section ".text"
	.align	4
	.global	nbif_get_stacktrace_0
nbif_get_stacktrace_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	get_stacktrace_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_get_stacktrace_0,.-nbif_get_stacktrace_0
	.type	nbif_get_stacktrace_0,@function
#endif

#ifndef HAVE_nbif_ebif_get_stacktrace_0
#define HAVE_nbif_ebif_get_stacktrace_0
	.section ".text"
	.align	4
	.global	nbif_ebif_get_stacktrace_0
nbif_ebif_get_stacktrace_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_get_stacktrace_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_get_stacktrace_0,.-nbif_ebif_get_stacktrace_0
	.type	nbif_ebif_get_stacktrace_0,@function
#endif

#ifndef HAVE_nbif_is_builtin_3
#define HAVE_nbif_is_builtin_3
	.section ".text"
	.align	4
	.global	nbif_is_builtin_3
nbif_is_builtin_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	is_builtin_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_is_builtin_3,.-nbif_is_builtin_3
	.type	nbif_is_builtin_3,@function
#endif

#ifndef HAVE_nbif_ebif_is_builtin_3
#define HAVE_nbif_ebif_is_builtin_3
	.section ".text"
	.align	4
	.global	nbif_ebif_is_builtin_3
nbif_ebif_is_builtin_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_is_builtin_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_builtin_3,.-nbif_ebif_is_builtin_3
	.type	nbif_ebif_is_builtin_3,@function
#endif

#ifndef HAVE_nbif_and_2
#define HAVE_nbif_and_2
	.section ".text"
	.align	4
	.global	nbif_and_2
nbif_and_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	and_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_and_2,.-nbif_and_2
	.type	nbif_and_2,@function
#endif

#ifndef HAVE_nbif_ebif_and_2
#define HAVE_nbif_ebif_and_2
	.section ".text"
	.align	4
	.global	nbif_ebif_and_2
nbif_ebif_and_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_and_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_and_2,.-nbif_ebif_and_2
	.type	nbif_ebif_and_2,@function
#endif

#ifndef HAVE_nbif_or_2
#define HAVE_nbif_or_2
	.section ".text"
	.align	4
	.global	nbif_or_2
nbif_or_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	or_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_or_2,.-nbif_or_2
	.type	nbif_or_2,@function
#endif

#ifndef HAVE_nbif_ebif_or_2
#define HAVE_nbif_ebif_or_2
	.section ".text"
	.align	4
	.global	nbif_ebif_or_2
nbif_ebif_or_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_or_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_or_2,.-nbif_ebif_or_2
	.type	nbif_ebif_or_2,@function
#endif

#ifndef HAVE_nbif_xor_2
#define HAVE_nbif_xor_2
	.section ".text"
	.align	4
	.global	nbif_xor_2
nbif_xor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	xor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_xor_2,.-nbif_xor_2
	.type	nbif_xor_2,@function
#endif

#ifndef HAVE_nbif_ebif_xor_2
#define HAVE_nbif_ebif_xor_2
	.section ".text"
	.align	4
	.global	nbif_ebif_xor_2
nbif_ebif_xor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_xor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_xor_2,.-nbif_ebif_xor_2
	.type	nbif_ebif_xor_2,@function
#endif

#ifndef HAVE_nbif_not_1
#define HAVE_nbif_not_1
	.section ".text"
	.align	4
	.global	nbif_not_1
nbif_not_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	not_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_not_1,.-nbif_not_1
	.type	nbif_not_1,@function
#endif

#ifndef HAVE_nbif_ebif_not_1
#define HAVE_nbif_ebif_not_1
	.section ".text"
	.align	4
	.global	nbif_ebif_not_1
nbif_ebif_not_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_not_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_not_1,.-nbif_ebif_not_1
	.type	nbif_ebif_not_1,@function
#endif

#ifndef HAVE_nbif_sgt_2
#define HAVE_nbif_sgt_2
	.section ".text"
	.align	4
	.global	nbif_sgt_2
nbif_sgt_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	sgt_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_sgt_2,.-nbif_sgt_2
	.type	nbif_sgt_2,@function
#endif

#ifndef HAVE_nbif_ebif_gt_2
#define HAVE_nbif_ebif_gt_2
	.section ".text"
	.align	4
	.global	nbif_ebif_gt_2
nbif_ebif_gt_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_gt_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_gt_2,.-nbif_ebif_gt_2
	.type	nbif_ebif_gt_2,@function
#endif

#ifndef HAVE_nbif_sge_2
#define HAVE_nbif_sge_2
	.section ".text"
	.align	4
	.global	nbif_sge_2
nbif_sge_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	sge_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_sge_2,.-nbif_sge_2
	.type	nbif_sge_2,@function
#endif

#ifndef HAVE_nbif_ebif_ge_2
#define HAVE_nbif_ebif_ge_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ge_2
nbif_ebif_ge_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ge_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ge_2,.-nbif_ebif_ge_2
	.type	nbif_ebif_ge_2,@function
#endif

#ifndef HAVE_nbif_slt_2
#define HAVE_nbif_slt_2
	.section ".text"
	.align	4
	.global	nbif_slt_2
nbif_slt_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	slt_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_slt_2,.-nbif_slt_2
	.type	nbif_slt_2,@function
#endif

#ifndef HAVE_nbif_ebif_lt_2
#define HAVE_nbif_ebif_lt_2
	.section ".text"
	.align	4
	.global	nbif_ebif_lt_2
nbif_ebif_lt_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_lt_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_lt_2,.-nbif_ebif_lt_2
	.type	nbif_ebif_lt_2,@function
#endif

#ifndef HAVE_nbif_sle_2
#define HAVE_nbif_sle_2
	.section ".text"
	.align	4
	.global	nbif_sle_2
nbif_sle_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	sle_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_sle_2,.-nbif_sle_2
	.type	nbif_sle_2,@function
#endif

#ifndef HAVE_nbif_ebif_le_2
#define HAVE_nbif_ebif_le_2
	.section ".text"
	.align	4
	.global	nbif_ebif_le_2
nbif_ebif_le_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_le_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_le_2,.-nbif_ebif_le_2
	.type	nbif_ebif_le_2,@function
#endif

#ifndef HAVE_nbif_seq_2
#define HAVE_nbif_seq_2
	.section ".text"
	.align	4
	.global	nbif_seq_2
nbif_seq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	seq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_seq_2,.-nbif_seq_2
	.type	nbif_seq_2,@function
#endif

#ifndef HAVE_nbif_ebif_eq_2
#define HAVE_nbif_ebif_eq_2
	.section ".text"
	.align	4
	.global	nbif_ebif_eq_2
nbif_ebif_eq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_eq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_eq_2,.-nbif_ebif_eq_2
	.type	nbif_ebif_eq_2,@function
#endif

#ifndef HAVE_nbif_seqeq_2
#define HAVE_nbif_seqeq_2
	.section ".text"
	.align	4
	.global	nbif_seqeq_2
nbif_seqeq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	seqeq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_seqeq_2,.-nbif_seqeq_2
	.type	nbif_seqeq_2,@function
#endif

#ifndef HAVE_nbif_ebif_areq_2
#define HAVE_nbif_ebif_areq_2
	.section ".text"
	.align	4
	.global	nbif_ebif_areq_2
nbif_ebif_areq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_areq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_areq_2,.-nbif_ebif_areq_2
	.type	nbif_ebif_areq_2,@function
#endif

#ifndef HAVE_nbif_sneq_2
#define HAVE_nbif_sneq_2
	.section ".text"
	.align	4
	.global	nbif_sneq_2
nbif_sneq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	sneq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_sneq_2,.-nbif_sneq_2
	.type	nbif_sneq_2,@function
#endif

#ifndef HAVE_nbif_ebif_neq_2
#define HAVE_nbif_ebif_neq_2
	.section ".text"
	.align	4
	.global	nbif_ebif_neq_2
nbif_ebif_neq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_neq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_neq_2,.-nbif_ebif_neq_2
	.type	nbif_ebif_neq_2,@function
#endif

#ifndef HAVE_nbif_sneqeq_2
#define HAVE_nbif_sneqeq_2
	.section ".text"
	.align	4
	.global	nbif_sneqeq_2
nbif_sneqeq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	sneqeq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_sneqeq_2,.-nbif_sneqeq_2
	.type	nbif_sneqeq_2,@function
#endif

#ifndef HAVE_nbif_ebif_nareq_2
#define HAVE_nbif_ebif_nareq_2
	.section ".text"
	.align	4
	.global	nbif_ebif_nareq_2
nbif_ebif_nareq_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_nareq_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_nareq_2,.-nbif_ebif_nareq_2
	.type	nbif_ebif_nareq_2,@function
#endif

#ifndef HAVE_nbif_splus_2
#define HAVE_nbif_splus_2
	.section ".text"
	.align	4
	.global	nbif_splus_2
nbif_splus_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	splus_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_splus_2,.-nbif_splus_2
	.type	nbif_splus_2,@function
#endif

#ifndef HAVE_nbif_ebif_plus_2
#define HAVE_nbif_ebif_plus_2
	.section ".text"
	.align	4
	.global	nbif_ebif_plus_2
nbif_ebif_plus_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_plus_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_plus_2,.-nbif_ebif_plus_2
	.type	nbif_ebif_plus_2,@function
#endif

#ifndef HAVE_nbif_sminus_2
#define HAVE_nbif_sminus_2
	.section ".text"
	.align	4
	.global	nbif_sminus_2
nbif_sminus_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	sminus_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_sminus_2,.-nbif_sminus_2
	.type	nbif_sminus_2,@function
#endif

#ifndef HAVE_nbif_ebif_minus_2
#define HAVE_nbif_ebif_minus_2
	.section ".text"
	.align	4
	.global	nbif_ebif_minus_2
nbif_ebif_minus_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_minus_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_minus_2,.-nbif_ebif_minus_2
	.type	nbif_ebif_minus_2,@function
#endif

#ifndef HAVE_nbif_stimes_2
#define HAVE_nbif_stimes_2
	.section ".text"
	.align	4
	.global	nbif_stimes_2
nbif_stimes_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	stimes_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_stimes_2,.-nbif_stimes_2
	.type	nbif_stimes_2,@function
#endif

#ifndef HAVE_nbif_ebif_multiply_2
#define HAVE_nbif_ebif_multiply_2
	.section ".text"
	.align	4
	.global	nbif_ebif_multiply_2
nbif_ebif_multiply_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_multiply_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_multiply_2,.-nbif_ebif_multiply_2
	.type	nbif_ebif_multiply_2,@function
#endif

#ifndef HAVE_nbif_div_2
#define HAVE_nbif_div_2
	.section ".text"
	.align	4
	.global	nbif_div_2
nbif_div_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	div_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_div_2,.-nbif_div_2
	.type	nbif_div_2,@function
#endif

#ifndef HAVE_nbif_ebif_divide_2
#define HAVE_nbif_ebif_divide_2
	.section ".text"
	.align	4
	.global	nbif_ebif_divide_2
nbif_ebif_divide_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_divide_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_divide_2,.-nbif_ebif_divide_2
	.type	nbif_ebif_divide_2,@function
#endif

#ifndef HAVE_nbif_intdiv_2
#define HAVE_nbif_intdiv_2
	.section ".text"
	.align	4
	.global	nbif_intdiv_2
nbif_intdiv_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	intdiv_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_intdiv_2,.-nbif_intdiv_2
	.type	nbif_intdiv_2,@function
#endif

#ifndef HAVE_nbif_ebif_intdiv_2
#define HAVE_nbif_ebif_intdiv_2
	.section ".text"
	.align	4
	.global	nbif_ebif_intdiv_2
nbif_ebif_intdiv_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_intdiv_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_intdiv_2,.-nbif_ebif_intdiv_2
	.type	nbif_ebif_intdiv_2,@function
#endif

#ifndef HAVE_nbif_rem_2
#define HAVE_nbif_rem_2
	.section ".text"
	.align	4
	.global	nbif_rem_2
nbif_rem_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	rem_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_rem_2,.-nbif_rem_2
	.type	nbif_rem_2,@function
#endif

#ifndef HAVE_nbif_ebif_rem_2
#define HAVE_nbif_ebif_rem_2
	.section ".text"
	.align	4
	.global	nbif_ebif_rem_2
nbif_ebif_rem_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_rem_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_rem_2,.-nbif_ebif_rem_2
	.type	nbif_ebif_rem_2,@function
#endif

#ifndef HAVE_nbif_bor_2
#define HAVE_nbif_bor_2
	.section ".text"
	.align	4
	.global	nbif_bor_2
nbif_bor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bor_2,.-nbif_bor_2
	.type	nbif_bor_2,@function
#endif

#ifndef HAVE_nbif_ebif_bor_2
#define HAVE_nbif_ebif_bor_2
	.section ".text"
	.align	4
	.global	nbif_ebif_bor_2
nbif_ebif_bor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_bor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_bor_2,.-nbif_ebif_bor_2
	.type	nbif_ebif_bor_2,@function
#endif

#ifndef HAVE_nbif_band_2
#define HAVE_nbif_band_2
	.section ".text"
	.align	4
	.global	nbif_band_2
nbif_band_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	band_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_band_2,.-nbif_band_2
	.type	nbif_band_2,@function
#endif

#ifndef HAVE_nbif_ebif_band_2
#define HAVE_nbif_ebif_band_2
	.section ".text"
	.align	4
	.global	nbif_ebif_band_2
nbif_ebif_band_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_band_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_band_2,.-nbif_ebif_band_2
	.type	nbif_ebif_band_2,@function
#endif

#ifndef HAVE_nbif_bxor_2
#define HAVE_nbif_bxor_2
	.section ".text"
	.align	4
	.global	nbif_bxor_2
nbif_bxor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bxor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bxor_2,.-nbif_bxor_2
	.type	nbif_bxor_2,@function
#endif

#ifndef HAVE_nbif_ebif_bxor_2
#define HAVE_nbif_ebif_bxor_2
	.section ".text"
	.align	4
	.global	nbif_ebif_bxor_2
nbif_ebif_bxor_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_bxor_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_bxor_2,.-nbif_ebif_bxor_2
	.type	nbif_ebif_bxor_2,@function
#endif

#ifndef HAVE_nbif_bsl_2
#define HAVE_nbif_bsl_2
	.section ".text"
	.align	4
	.global	nbif_bsl_2
nbif_bsl_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bsl_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bsl_2,.-nbif_bsl_2
	.type	nbif_bsl_2,@function
#endif

#ifndef HAVE_nbif_ebif_bsl_2
#define HAVE_nbif_ebif_bsl_2
	.section ".text"
	.align	4
	.global	nbif_ebif_bsl_2
nbif_ebif_bsl_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_bsl_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_bsl_2,.-nbif_ebif_bsl_2
	.type	nbif_ebif_bsl_2,@function
#endif

#ifndef HAVE_nbif_bsr_2
#define HAVE_nbif_bsr_2
	.section ".text"
	.align	4
	.global	nbif_bsr_2
nbif_bsr_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	bsr_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_bsr_2,.-nbif_bsr_2
	.type	nbif_bsr_2,@function
#endif

#ifndef HAVE_nbif_ebif_bsr_2
#define HAVE_nbif_ebif_bsr_2
	.section ".text"
	.align	4
	.global	nbif_ebif_bsr_2
nbif_ebif_bsr_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_bsr_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_bsr_2,.-nbif_ebif_bsr_2
	.type	nbif_ebif_bsr_2,@function
#endif

#ifndef HAVE_nbif_bnot_1
#define HAVE_nbif_bnot_1
	.section ".text"
	.align	4
	.global	nbif_bnot_1
nbif_bnot_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	bnot_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_bnot_1,.-nbif_bnot_1
	.type	nbif_bnot_1,@function
#endif

#ifndef HAVE_nbif_ebif_bnot_1
#define HAVE_nbif_ebif_bnot_1
	.section ".text"
	.align	4
	.global	nbif_ebif_bnot_1
nbif_ebif_bnot_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_bnot_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_bnot_1,.-nbif_ebif_bnot_1
	.type	nbif_ebif_bnot_1,@function
#endif

#ifndef HAVE_nbif_sminus_1
#define HAVE_nbif_sminus_1
	.section ".text"
	.align	4
	.global	nbif_sminus_1
nbif_sminus_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	sminus_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_sminus_1,.-nbif_sminus_1
	.type	nbif_sminus_1,@function
#endif

#ifndef HAVE_nbif_ebif_minus_1
#define HAVE_nbif_ebif_minus_1
	.section ".text"
	.align	4
	.global	nbif_ebif_minus_1
nbif_ebif_minus_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_minus_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_minus_1,.-nbif_ebif_minus_1
	.type	nbif_ebif_minus_1,@function
#endif

#ifndef HAVE_nbif_splus_1
#define HAVE_nbif_splus_1
	.section ".text"
	.align	4
	.global	nbif_splus_1
nbif_splus_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	splus_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_splus_1,.-nbif_splus_1
	.type	nbif_splus_1,@function
#endif

#ifndef HAVE_nbif_ebif_plus_1
#define HAVE_nbif_ebif_plus_1
	.section ".text"
	.align	4
	.global	nbif_ebif_plus_1
nbif_ebif_plus_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_plus_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_plus_1,.-nbif_ebif_plus_1
	.type	nbif_ebif_plus_1,@function
#endif

#ifndef HAVE_nbif_ebif_bang_2
#define HAVE_nbif_ebif_bang_2
	.section ".text"
	.align	4
	.global	nbif_ebif_bang_2
nbif_ebif_bang_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_bang_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_bang_2,.-nbif_ebif_bang_2
	.type	nbif_ebif_bang_2,@function
#endif

#ifndef HAVE_nbif_ebif_send_2
#define HAVE_nbif_ebif_send_2
	.section ".text"
	.align	4
	.global	nbif_ebif_send_2
nbif_ebif_send_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_send_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_send_2,.-nbif_ebif_send_2
	.type	nbif_ebif_send_2,@function
#endif

#ifndef HAVE_nbif_send_2
#define HAVE_nbif_send_2
	.section ".text"
	.align	4
	.global	nbif_send_2
nbif_send_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	send_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_send_2,.-nbif_send_2
	.type	nbif_send_2,@function
#endif

#ifndef HAVE_nbif_ebif_send_3
#define HAVE_nbif_ebif_send_3
	.section ".text"
	.align	4
	.global	nbif_ebif_send_3
nbif_ebif_send_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_send_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_send_3,.-nbif_ebif_send_3
	.type	nbif_ebif_send_3,@function
#endif

#ifndef HAVE_nbif_send_3
#define HAVE_nbif_send_3
	.section ".text"
	.align	4
	.global	nbif_send_3
nbif_send_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	send_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_send_3,.-nbif_send_3
	.type	nbif_send_3,@function
#endif

#ifndef HAVE_nbif_ebif_plusplus_2
#define HAVE_nbif_ebif_plusplus_2
	.section ".text"
	.align	4
	.global	nbif_ebif_plusplus_2
nbif_ebif_plusplus_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_plusplus_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_plusplus_2,.-nbif_ebif_plusplus_2
	.type	nbif_ebif_plusplus_2,@function
#endif

#ifndef HAVE_nbif_ebif_append_2
#define HAVE_nbif_ebif_append_2
	.section ".text"
	.align	4
	.global	nbif_ebif_append_2
nbif_ebif_append_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_append_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_append_2,.-nbif_ebif_append_2
	.type	nbif_ebif_append_2,@function
#endif

#ifndef HAVE_nbif_append_2
#define HAVE_nbif_append_2
	.section ".text"
	.align	4
	.global	nbif_append_2
nbif_append_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	append_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_append_2,.-nbif_append_2
	.type	nbif_append_2,@function
#endif

#ifndef HAVE_nbif_ebif_minusminus_2
#define HAVE_nbif_ebif_minusminus_2
	.section ".text"
	.align	4
	.global	nbif_ebif_minusminus_2
nbif_ebif_minusminus_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_minusminus_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_minusminus_2,.-nbif_ebif_minusminus_2
	.type	nbif_ebif_minusminus_2,@function
#endif

#ifndef HAVE_nbif_ebif_list_subtract_2
#define HAVE_nbif_ebif_list_subtract_2
	.section ".text"
	.align	4
	.global	nbif_ebif_list_subtract_2
nbif_ebif_list_subtract_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_list_subtract_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_list_subtract_2,.-nbif_ebif_list_subtract_2
	.type	nbif_ebif_list_subtract_2,@function
#endif

#ifndef HAVE_nbif_subtract_2
#define HAVE_nbif_subtract_2
	.section ".text"
	.align	4
	.global	nbif_subtract_2
nbif_subtract_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	subtract_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_subtract_2,.-nbif_subtract_2
	.type	nbif_subtract_2,@function
#endif

#ifndef HAVE_nbif_is_atom_1
#define HAVE_nbif_is_atom_1
	.section ".text"
	.align	4
	.global	nbif_is_atom_1
nbif_is_atom_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_atom_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_atom_1,.-nbif_is_atom_1
	.type	nbif_is_atom_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_atom_1
#define HAVE_nbif_ebif_is_atom_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_atom_1
nbif_ebif_is_atom_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_atom_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_atom_1,.-nbif_ebif_is_atom_1
	.type	nbif_ebif_is_atom_1,@function
#endif

#ifndef HAVE_nbif_is_list_1
#define HAVE_nbif_is_list_1
	.section ".text"
	.align	4
	.global	nbif_is_list_1
nbif_is_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_list_1,.-nbif_is_list_1
	.type	nbif_is_list_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_list_1
#define HAVE_nbif_ebif_is_list_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_list_1
nbif_ebif_is_list_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_list_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_list_1,.-nbif_ebif_is_list_1
	.type	nbif_ebif_is_list_1,@function
#endif

#ifndef HAVE_nbif_is_tuple_1
#define HAVE_nbif_is_tuple_1
	.section ".text"
	.align	4
	.global	nbif_is_tuple_1
nbif_is_tuple_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_tuple_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_tuple_1,.-nbif_is_tuple_1
	.type	nbif_is_tuple_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_tuple_1
#define HAVE_nbif_ebif_is_tuple_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_tuple_1
nbif_ebif_is_tuple_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_tuple_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_tuple_1,.-nbif_ebif_is_tuple_1
	.type	nbif_ebif_is_tuple_1,@function
#endif

#ifndef HAVE_nbif_is_constant_1
#define HAVE_nbif_is_constant_1
	.section ".text"
	.align	4
	.global	nbif_is_constant_1
nbif_is_constant_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_constant_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_constant_1,.-nbif_is_constant_1
	.type	nbif_is_constant_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_constant_1
#define HAVE_nbif_ebif_is_constant_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_constant_1
nbif_ebif_is_constant_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_constant_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_constant_1,.-nbif_ebif_is_constant_1
	.type	nbif_ebif_is_constant_1,@function
#endif

#ifndef HAVE_nbif_is_float_1
#define HAVE_nbif_is_float_1
	.section ".text"
	.align	4
	.global	nbif_is_float_1
nbif_is_float_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_float_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_float_1,.-nbif_is_float_1
	.type	nbif_is_float_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_float_1
#define HAVE_nbif_ebif_is_float_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_float_1
nbif_ebif_is_float_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_float_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_float_1,.-nbif_ebif_is_float_1
	.type	nbif_ebif_is_float_1,@function
#endif

#ifndef HAVE_nbif_is_integer_1
#define HAVE_nbif_is_integer_1
	.section ".text"
	.align	4
	.global	nbif_is_integer_1
nbif_is_integer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_integer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_integer_1,.-nbif_is_integer_1
	.type	nbif_is_integer_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_integer_1
#define HAVE_nbif_ebif_is_integer_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_integer_1
nbif_ebif_is_integer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_integer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_integer_1,.-nbif_ebif_is_integer_1
	.type	nbif_ebif_is_integer_1,@function
#endif

#ifndef HAVE_nbif_is_number_1
#define HAVE_nbif_is_number_1
	.section ".text"
	.align	4
	.global	nbif_is_number_1
nbif_is_number_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_number_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_number_1,.-nbif_is_number_1
	.type	nbif_is_number_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_number_1
#define HAVE_nbif_ebif_is_number_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_number_1
nbif_ebif_is_number_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_number_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_number_1,.-nbif_ebif_is_number_1
	.type	nbif_ebif_is_number_1,@function
#endif

#ifndef HAVE_nbif_is_pid_1
#define HAVE_nbif_is_pid_1
	.section ".text"
	.align	4
	.global	nbif_is_pid_1
nbif_is_pid_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_pid_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_pid_1,.-nbif_is_pid_1
	.type	nbif_is_pid_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_pid_1
#define HAVE_nbif_ebif_is_pid_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_pid_1
nbif_ebif_is_pid_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_pid_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_pid_1,.-nbif_ebif_is_pid_1
	.type	nbif_ebif_is_pid_1,@function
#endif

#ifndef HAVE_nbif_is_port_1
#define HAVE_nbif_is_port_1
	.section ".text"
	.align	4
	.global	nbif_is_port_1
nbif_is_port_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_port_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_port_1,.-nbif_is_port_1
	.type	nbif_is_port_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_port_1
#define HAVE_nbif_ebif_is_port_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_port_1
nbif_ebif_is_port_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_port_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_port_1,.-nbif_ebif_is_port_1
	.type	nbif_ebif_is_port_1,@function
#endif

#ifndef HAVE_nbif_is_reference_1
#define HAVE_nbif_is_reference_1
	.section ".text"
	.align	4
	.global	nbif_is_reference_1
nbif_is_reference_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_reference_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_reference_1,.-nbif_is_reference_1
	.type	nbif_is_reference_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_reference_1
#define HAVE_nbif_ebif_is_reference_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_reference_1
nbif_ebif_is_reference_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_reference_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_reference_1,.-nbif_ebif_is_reference_1
	.type	nbif_ebif_is_reference_1,@function
#endif

#ifndef HAVE_nbif_is_binary_1
#define HAVE_nbif_is_binary_1
	.section ".text"
	.align	4
	.global	nbif_is_binary_1
nbif_is_binary_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_binary_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_binary_1,.-nbif_is_binary_1
	.type	nbif_is_binary_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_binary_1
#define HAVE_nbif_ebif_is_binary_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_binary_1
nbif_ebif_is_binary_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_binary_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_binary_1,.-nbif_ebif_is_binary_1
	.type	nbif_ebif_is_binary_1,@function
#endif

#ifndef HAVE_nbif_is_function_1
#define HAVE_nbif_is_function_1
	.section ".text"
	.align	4
	.global	nbif_is_function_1
nbif_is_function_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_function_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_function_1,.-nbif_is_function_1
	.type	nbif_is_function_1,@function
#endif

#ifndef HAVE_nbif_ebif_is_function_1
#define HAVE_nbif_ebif_is_function_1
	.section ".text"
	.align	4
	.global	nbif_ebif_is_function_1
nbif_ebif_is_function_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_is_function_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_function_1,.-nbif_ebif_is_function_1
	.type	nbif_ebif_is_function_1,@function
#endif

#ifndef HAVE_nbif_is_function_2
#define HAVE_nbif_is_function_2
	.section ".text"
	.align	4
	.global	nbif_is_function_2
nbif_is_function_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	is_function_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_is_function_2,.-nbif_is_function_2
	.type	nbif_is_function_2,@function
#endif

#ifndef HAVE_nbif_ebif_is_function_2
#define HAVE_nbif_ebif_is_function_2
	.section ".text"
	.align	4
	.global	nbif_ebif_is_function_2
nbif_ebif_is_function_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_is_function_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_function_2,.-nbif_ebif_is_function_2
	.type	nbif_ebif_is_function_2,@function
#endif

#ifndef HAVE_nbif_is_record_2
#define HAVE_nbif_is_record_2
	.section ".text"
	.align	4
	.global	nbif_is_record_2
nbif_is_record_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	is_record_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_is_record_2,.-nbif_is_record_2
	.type	nbif_is_record_2,@function
#endif

#ifndef HAVE_nbif_ebif_is_record_2
#define HAVE_nbif_ebif_is_record_2
	.section ".text"
	.align	4
	.global	nbif_ebif_is_record_2
nbif_ebif_is_record_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_is_record_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_record_2,.-nbif_ebif_is_record_2
	.type	nbif_ebif_is_record_2,@function
#endif

#ifndef HAVE_nbif_is_record_3
#define HAVE_nbif_is_record_3
	.section ".text"
	.align	4
	.global	nbif_is_record_3
nbif_is_record_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	is_record_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_is_record_3,.-nbif_is_record_3
	.type	nbif_is_record_3,@function
#endif

#ifndef HAVE_nbif_ebif_is_record_3
#define HAVE_nbif_ebif_is_record_3
	.section ".text"
	.align	4
	.global	nbif_ebif_is_record_3
nbif_ebif_is_record_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_is_record_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_is_record_3,.-nbif_ebif_is_record_3
	.type	nbif_ebif_is_record_3,@function
#endif

#ifndef HAVE_nbif_match_spec_test_3
#define HAVE_nbif_match_spec_test_3
	.section ".text"
	.align	4
	.global	nbif_match_spec_test_3
nbif_match_spec_test_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	match_spec_test_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_match_spec_test_3,.-nbif_match_spec_test_3
	.type	nbif_match_spec_test_3,@function
#endif

#ifndef HAVE_nbif_dollar_put_2
#define HAVE_nbif_dollar_put_2
	.section ".text"
	.align	4
	.global	nbif_dollar_put_2
nbif_dollar_put_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	dollar_put_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_dollar_put_2,.-nbif_dollar_put_2
	.type	nbif_dollar_put_2,@function
#endif

#ifndef HAVE_nbif_dollar_get_0
#define HAVE_nbif_dollar_get_0
	.section ".text"
	.align	4
	.global	nbif_dollar_get_0
nbif_dollar_get_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	dollar_get_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_dollar_get_0,.-nbif_dollar_get_0
	.type	nbif_dollar_get_0,@function
#endif

#ifndef HAVE_nbif_dollar_get_1
#define HAVE_nbif_dollar_get_1
	.section ".text"
	.align	4
	.global	nbif_dollar_get_1
nbif_dollar_get_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	dollar_get_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_dollar_get_1,.-nbif_dollar_get_1
	.type	nbif_dollar_get_1,@function
#endif

#ifndef HAVE_nbif_dollar_get_keys_1
#define HAVE_nbif_dollar_get_keys_1
	.section ".text"
	.align	4
	.global	nbif_dollar_get_keys_1
nbif_dollar_get_keys_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	dollar_get_keys_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_dollar_get_keys_1,.-nbif_dollar_get_keys_1
	.type	nbif_dollar_get_keys_1,@function
#endif

#ifndef HAVE_nbif_dollar_erase_0
#define HAVE_nbif_dollar_erase_0
	.section ".text"
	.align	4
	.global	nbif_dollar_erase_0
nbif_dollar_erase_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	dollar_erase_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_dollar_erase_0,.-nbif_dollar_erase_0
	.type	nbif_dollar_erase_0,@function
#endif

#ifndef HAVE_nbif_dollar_erase_1
#define HAVE_nbif_dollar_erase_1
	.section ".text"
	.align	4
	.global	nbif_dollar_erase_1
nbif_dollar_erase_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	dollar_erase_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_dollar_erase_1,.-nbif_dollar_erase_1
	.type	nbif_dollar_erase_1,@function
#endif

#ifndef HAVE_nbif_ets_all_0
#define HAVE_nbif_ets_all_0
	.section ".text"
	.align	4
	.global	nbif_ets_all_0
nbif_ets_all_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ets_all_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ets_all_0,.-nbif_ets_all_0
	.type	nbif_ets_all_0,@function
#endif

#ifndef HAVE_nbif_ebif_ets_all_0
#define HAVE_nbif_ebif_ets_all_0
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_all_0
nbif_ebif_ets_all_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_ets_all_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_all_0,.-nbif_ebif_ets_all_0
	.type	nbif_ebif_ets_all_0,@function
#endif

#ifndef HAVE_nbif_ets_new_2
#define HAVE_nbif_ets_new_2
	.section ".text"
	.align	4
	.global	nbif_ets_new_2
nbif_ets_new_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_new_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_new_2,.-nbif_ets_new_2
	.type	nbif_ets_new_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_new_2
#define HAVE_nbif_ebif_ets_new_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_new_2
nbif_ebif_ets_new_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_new_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_new_2,.-nbif_ebif_ets_new_2
	.type	nbif_ebif_ets_new_2,@function
#endif

#ifndef HAVE_nbif_ets_db_delete_1
#define HAVE_nbif_ets_db_delete_1
	.section ".text"
	.align	4
	.global	nbif_ets_db_delete_1
nbif_ets_db_delete_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_db_delete_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_db_delete_1,.-nbif_ets_db_delete_1
	.type	nbif_ets_db_delete_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_db_delete_1
#define HAVE_nbif_ebif_ets_db_delete_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_db_delete_1
nbif_ebif_ets_db_delete_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_db_delete_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_db_delete_1,.-nbif_ebif_ets_db_delete_1
	.type	nbif_ebif_ets_db_delete_1,@function
#endif

#ifndef HAVE_nbif_ets_delete_2
#define HAVE_nbif_ets_delete_2
	.section ".text"
	.align	4
	.global	nbif_ets_delete_2
nbif_ets_delete_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_delete_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_delete_2,.-nbif_ets_delete_2
	.type	nbif_ets_delete_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_delete_2
#define HAVE_nbif_ebif_ets_delete_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_delete_2
nbif_ebif_ets_delete_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_delete_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_delete_2,.-nbif_ebif_ets_delete_2
	.type	nbif_ebif_ets_delete_2,@function
#endif

#ifndef HAVE_nbif_ets_delete_all_objects_1
#define HAVE_nbif_ets_delete_all_objects_1
	.section ".text"
	.align	4
	.global	nbif_ets_delete_all_objects_1
nbif_ets_delete_all_objects_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_delete_all_objects_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_delete_all_objects_1,.-nbif_ets_delete_all_objects_1
	.type	nbif_ets_delete_all_objects_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_delete_all_objects_1
#define HAVE_nbif_ebif_ets_delete_all_objects_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_delete_all_objects_1
nbif_ebif_ets_delete_all_objects_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_delete_all_objects_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_delete_all_objects_1,.-nbif_ebif_ets_delete_all_objects_1
	.type	nbif_ebif_ets_delete_all_objects_1,@function
#endif

#ifndef HAVE_nbif_ets_delete_object_2
#define HAVE_nbif_ets_delete_object_2
	.section ".text"
	.align	4
	.global	nbif_ets_delete_object_2
nbif_ets_delete_object_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_delete_object_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_delete_object_2,.-nbif_ets_delete_object_2
	.type	nbif_ets_delete_object_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_delete_object_2
#define HAVE_nbif_ebif_ets_delete_object_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_delete_object_2
nbif_ebif_ets_delete_object_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_delete_object_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_delete_object_2,.-nbif_ebif_ets_delete_object_2
	.type	nbif_ebif_ets_delete_object_2,@function
#endif

#ifndef HAVE_nbif_ets_first_1
#define HAVE_nbif_ets_first_1
	.section ".text"
	.align	4
	.global	nbif_ets_first_1
nbif_ets_first_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_first_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_first_1,.-nbif_ets_first_1
	.type	nbif_ets_first_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_first_1
#define HAVE_nbif_ebif_ets_first_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_first_1
nbif_ebif_ets_first_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_first_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_first_1,.-nbif_ebif_ets_first_1
	.type	nbif_ebif_ets_first_1,@function
#endif

#ifndef HAVE_nbif_ets_fixtable_2
#define HAVE_nbif_ets_fixtable_2
	.section ".text"
	.align	4
	.global	nbif_ets_fixtable_2
nbif_ets_fixtable_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_fixtable_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_fixtable_2,.-nbif_ets_fixtable_2
	.type	nbif_ets_fixtable_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_fixtable_2
#define HAVE_nbif_ebif_ets_fixtable_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_fixtable_2
nbif_ebif_ets_fixtable_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_fixtable_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_fixtable_2,.-nbif_ebif_ets_fixtable_2
	.type	nbif_ebif_ets_fixtable_2,@function
#endif

#ifndef HAVE_nbif_ets_is_compiled_ms_1
#define HAVE_nbif_ets_is_compiled_ms_1
	.section ".text"
	.align	4
	.global	nbif_ets_is_compiled_ms_1
nbif_ets_is_compiled_ms_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_is_compiled_ms_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_is_compiled_ms_1,.-nbif_ets_is_compiled_ms_1
	.type	nbif_ets_is_compiled_ms_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_is_compiled_ms_1
#define HAVE_nbif_ebif_ets_is_compiled_ms_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_is_compiled_ms_1
nbif_ebif_ets_is_compiled_ms_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_is_compiled_ms_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_is_compiled_ms_1,.-nbif_ebif_ets_is_compiled_ms_1
	.type	nbif_ebif_ets_is_compiled_ms_1,@function
#endif

#ifndef HAVE_nbif_ets_lookup_2
#define HAVE_nbif_ets_lookup_2
	.section ".text"
	.align	4
	.global	nbif_ets_lookup_2
nbif_ets_lookup_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_lookup_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_lookup_2,.-nbif_ets_lookup_2
	.type	nbif_ets_lookup_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_lookup_2
#define HAVE_nbif_ebif_ets_lookup_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_lookup_2
nbif_ebif_ets_lookup_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_lookup_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_lookup_2,.-nbif_ebif_ets_lookup_2
	.type	nbif_ebif_ets_lookup_2,@function
#endif

#ifndef HAVE_nbif_ets_lookup_element_3
#define HAVE_nbif_ets_lookup_element_3
	.section ".text"
	.align	4
	.global	nbif_ets_lookup_element_3
nbif_ets_lookup_element_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_lookup_element_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ets_lookup_element_3,.-nbif_ets_lookup_element_3
	.type	nbif_ets_lookup_element_3,@function
#endif

#ifndef HAVE_nbif_ebif_ets_lookup_element_3
#define HAVE_nbif_ebif_ets_lookup_element_3
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_lookup_element_3
nbif_ebif_ets_lookup_element_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_lookup_element_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_lookup_element_3,.-nbif_ebif_ets_lookup_element_3
	.type	nbif_ebif_ets_lookup_element_3,@function
#endif

#ifndef HAVE_nbif_ets_db_info_2
#define HAVE_nbif_ets_db_info_2
	.section ".text"
	.align	4
	.global	nbif_ets_db_info_2
nbif_ets_db_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_db_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_db_info_2,.-nbif_ets_db_info_2
	.type	nbif_ets_db_info_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_db_info_2
#define HAVE_nbif_ebif_ets_db_info_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_db_info_2
nbif_ebif_ets_db_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_db_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_db_info_2,.-nbif_ebif_ets_db_info_2
	.type	nbif_ebif_ets_db_info_2,@function
#endif

#ifndef HAVE_nbif_ets_last_1
#define HAVE_nbif_ets_last_1
	.section ".text"
	.align	4
	.global	nbif_ets_last_1
nbif_ets_last_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_last_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_last_1,.-nbif_ets_last_1
	.type	nbif_ets_last_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_last_1
#define HAVE_nbif_ebif_ets_last_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_last_1
nbif_ebif_ets_last_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_last_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_last_1,.-nbif_ebif_ets_last_1
	.type	nbif_ebif_ets_last_1,@function
#endif

#ifndef HAVE_nbif_ets_match_1
#define HAVE_nbif_ets_match_1
	.section ".text"
	.align	4
	.global	nbif_ets_match_1
nbif_ets_match_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_match_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_match_1,.-nbif_ets_match_1
	.type	nbif_ets_match_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_1
#define HAVE_nbif_ebif_ets_match_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_1
nbif_ebif_ets_match_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_1,.-nbif_ebif_ets_match_1
	.type	nbif_ebif_ets_match_1,@function
#endif

#ifndef HAVE_nbif_ets_match_2
#define HAVE_nbif_ets_match_2
	.section ".text"
	.align	4
	.global	nbif_ets_match_2
nbif_ets_match_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_match_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_match_2,.-nbif_ets_match_2
	.type	nbif_ets_match_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_2
#define HAVE_nbif_ebif_ets_match_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_2
nbif_ebif_ets_match_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_2,.-nbif_ebif_ets_match_2
	.type	nbif_ebif_ets_match_2,@function
#endif

#ifndef HAVE_nbif_ets_match_3
#define HAVE_nbif_ets_match_3
	.section ".text"
	.align	4
	.global	nbif_ets_match_3
nbif_ets_match_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_match_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ets_match_3,.-nbif_ets_match_3
	.type	nbif_ets_match_3,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_3
#define HAVE_nbif_ebif_ets_match_3
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_3
nbif_ebif_ets_match_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_3,.-nbif_ebif_ets_match_3
	.type	nbif_ebif_ets_match_3,@function
#endif

#ifndef HAVE_nbif_ets_match_object_1
#define HAVE_nbif_ets_match_object_1
	.section ".text"
	.align	4
	.global	nbif_ets_match_object_1
nbif_ets_match_object_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_match_object_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_match_object_1,.-nbif_ets_match_object_1
	.type	nbif_ets_match_object_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_object_1
#define HAVE_nbif_ebif_ets_match_object_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_object_1
nbif_ebif_ets_match_object_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_object_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_object_1,.-nbif_ebif_ets_match_object_1
	.type	nbif_ebif_ets_match_object_1,@function
#endif

#ifndef HAVE_nbif_ets_match_object_2
#define HAVE_nbif_ets_match_object_2
	.section ".text"
	.align	4
	.global	nbif_ets_match_object_2
nbif_ets_match_object_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_match_object_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_match_object_2,.-nbif_ets_match_object_2
	.type	nbif_ets_match_object_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_object_2
#define HAVE_nbif_ebif_ets_match_object_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_object_2
nbif_ebif_ets_match_object_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_object_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_object_2,.-nbif_ebif_ets_match_object_2
	.type	nbif_ebif_ets_match_object_2,@function
#endif

#ifndef HAVE_nbif_ets_match_object_3
#define HAVE_nbif_ets_match_object_3
	.section ".text"
	.align	4
	.global	nbif_ets_match_object_3
nbif_ets_match_object_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_match_object_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ets_match_object_3,.-nbif_ets_match_object_3
	.type	nbif_ets_match_object_3,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_object_3
#define HAVE_nbif_ebif_ets_match_object_3
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_object_3
nbif_ebif_ets_match_object_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_object_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_object_3,.-nbif_ebif_ets_match_object_3
	.type	nbif_ebif_ets_match_object_3,@function
#endif

#ifndef HAVE_nbif_ets_member_2
#define HAVE_nbif_ets_member_2
	.section ".text"
	.align	4
	.global	nbif_ets_member_2
nbif_ets_member_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_member_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_member_2,.-nbif_ets_member_2
	.type	nbif_ets_member_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_member_2
#define HAVE_nbif_ebif_ets_member_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_member_2
nbif_ebif_ets_member_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_member_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_member_2,.-nbif_ebif_ets_member_2
	.type	nbif_ebif_ets_member_2,@function
#endif

#ifndef HAVE_nbif_ets_next_2
#define HAVE_nbif_ets_next_2
	.section ".text"
	.align	4
	.global	nbif_ets_next_2
nbif_ets_next_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_next_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_next_2,.-nbif_ets_next_2
	.type	nbif_ets_next_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_next_2
#define HAVE_nbif_ebif_ets_next_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_next_2
nbif_ebif_ets_next_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_next_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_next_2,.-nbif_ebif_ets_next_2
	.type	nbif_ebif_ets_next_2,@function
#endif

#ifndef HAVE_nbif_ets_prev_2
#define HAVE_nbif_ets_prev_2
	.section ".text"
	.align	4
	.global	nbif_ets_prev_2
nbif_ets_prev_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_prev_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_prev_2,.-nbif_ets_prev_2
	.type	nbif_ets_prev_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_prev_2
#define HAVE_nbif_ebif_ets_prev_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_prev_2
nbif_ebif_ets_prev_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_prev_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_prev_2,.-nbif_ebif_ets_prev_2
	.type	nbif_ebif_ets_prev_2,@function
#endif

#ifndef HAVE_nbif_ets_insert_2
#define HAVE_nbif_ets_insert_2
	.section ".text"
	.align	4
	.global	nbif_ets_insert_2
nbif_ets_insert_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_insert_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_insert_2,.-nbif_ets_insert_2
	.type	nbif_ets_insert_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_insert_2
#define HAVE_nbif_ebif_ets_insert_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_insert_2
nbif_ebif_ets_insert_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_insert_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_insert_2,.-nbif_ebif_ets_insert_2
	.type	nbif_ebif_ets_insert_2,@function
#endif

#ifndef HAVE_nbif_ets_insert_new_2
#define HAVE_nbif_ets_insert_new_2
	.section ".text"
	.align	4
	.global	nbif_ets_insert_new_2
nbif_ets_insert_new_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_insert_new_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_insert_new_2,.-nbif_ets_insert_new_2
	.type	nbif_ets_insert_new_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_insert_new_2
#define HAVE_nbif_ebif_ets_insert_new_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_insert_new_2
nbif_ebif_ets_insert_new_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_insert_new_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_insert_new_2,.-nbif_ebif_ets_insert_new_2
	.type	nbif_ebif_ets_insert_new_2,@function
#endif

#ifndef HAVE_nbif_ets_rename_2
#define HAVE_nbif_ets_rename_2
	.section ".text"
	.align	4
	.global	nbif_ets_rename_2
nbif_ets_rename_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_rename_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_rename_2,.-nbif_ets_rename_2
	.type	nbif_ets_rename_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_rename_2
#define HAVE_nbif_ebif_ets_rename_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_rename_2
nbif_ebif_ets_rename_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_rename_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_rename_2,.-nbif_ebif_ets_rename_2
	.type	nbif_ebif_ets_rename_2,@function
#endif

#ifndef HAVE_nbif_ets_safe_fixtable_2
#define HAVE_nbif_ets_safe_fixtable_2
	.section ".text"
	.align	4
	.global	nbif_ets_safe_fixtable_2
nbif_ets_safe_fixtable_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_safe_fixtable_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_safe_fixtable_2,.-nbif_ets_safe_fixtable_2
	.type	nbif_ets_safe_fixtable_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_safe_fixtable_2
#define HAVE_nbif_ebif_ets_safe_fixtable_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_safe_fixtable_2
nbif_ebif_ets_safe_fixtable_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_safe_fixtable_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_safe_fixtable_2,.-nbif_ebif_ets_safe_fixtable_2
	.type	nbif_ebif_ets_safe_fixtable_2,@function
#endif

#ifndef HAVE_nbif_ets_slot_2
#define HAVE_nbif_ets_slot_2
	.section ".text"
	.align	4
	.global	nbif_ets_slot_2
nbif_ets_slot_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_slot_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_slot_2,.-nbif_ets_slot_2
	.type	nbif_ets_slot_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_slot_2
#define HAVE_nbif_ebif_ets_slot_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_slot_2
nbif_ebif_ets_slot_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_slot_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_slot_2,.-nbif_ebif_ets_slot_2
	.type	nbif_ebif_ets_slot_2,@function
#endif

#ifndef HAVE_nbif_ets_update_counter_3
#define HAVE_nbif_ets_update_counter_3
	.section ".text"
	.align	4
	.global	nbif_ets_update_counter_3
nbif_ets_update_counter_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_update_counter_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ets_update_counter_3,.-nbif_ets_update_counter_3
	.type	nbif_ets_update_counter_3,@function
#endif

#ifndef HAVE_nbif_ebif_ets_update_counter_3
#define HAVE_nbif_ebif_ets_update_counter_3
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_update_counter_3
nbif_ebif_ets_update_counter_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_update_counter_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_update_counter_3,.-nbif_ebif_ets_update_counter_3
	.type	nbif_ebif_ets_update_counter_3,@function
#endif

#ifndef HAVE_nbif_ets_select_1
#define HAVE_nbif_ets_select_1
	.section ".text"
	.align	4
	.global	nbif_ets_select_1
nbif_ets_select_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_select_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_select_1,.-nbif_ets_select_1
	.type	nbif_ets_select_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_1
#define HAVE_nbif_ebif_ets_select_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_1
nbif_ebif_ets_select_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_1,.-nbif_ebif_ets_select_1
	.type	nbif_ebif_ets_select_1,@function
#endif

#ifndef HAVE_nbif_ets_select_2
#define HAVE_nbif_ets_select_2
	.section ".text"
	.align	4
	.global	nbif_ets_select_2
nbif_ets_select_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_select_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_select_2,.-nbif_ets_select_2
	.type	nbif_ets_select_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_2
#define HAVE_nbif_ebif_ets_select_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_2
nbif_ebif_ets_select_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_2,.-nbif_ebif_ets_select_2
	.type	nbif_ebif_ets_select_2,@function
#endif

#ifndef HAVE_nbif_ets_select_3
#define HAVE_nbif_ets_select_3
	.section ".text"
	.align	4
	.global	nbif_ets_select_3
nbif_ets_select_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_select_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ets_select_3,.-nbif_ets_select_3
	.type	nbif_ets_select_3,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_3
#define HAVE_nbif_ebif_ets_select_3
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_3
nbif_ebif_ets_select_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_3,.-nbif_ebif_ets_select_3
	.type	nbif_ebif_ets_select_3,@function
#endif

#ifndef HAVE_nbif_ets_select_count_2
#define HAVE_nbif_ets_select_count_2
	.section ".text"
	.align	4
	.global	nbif_ets_select_count_2
nbif_ets_select_count_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_select_count_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_select_count_2,.-nbif_ets_select_count_2
	.type	nbif_ets_select_count_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_count_2
#define HAVE_nbif_ebif_ets_select_count_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_count_2
nbif_ebif_ets_select_count_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_count_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_count_2,.-nbif_ebif_ets_select_count_2
	.type	nbif_ebif_ets_select_count_2,@function
#endif

#ifndef HAVE_nbif_ets_select_reverse_1
#define HAVE_nbif_ets_select_reverse_1
	.section ".text"
	.align	4
	.global	nbif_ets_select_reverse_1
nbif_ets_select_reverse_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_select_reverse_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_select_reverse_1,.-nbif_ets_select_reverse_1
	.type	nbif_ets_select_reverse_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_reverse_1
#define HAVE_nbif_ebif_ets_select_reverse_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_reverse_1
nbif_ebif_ets_select_reverse_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_reverse_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_reverse_1,.-nbif_ebif_ets_select_reverse_1
	.type	nbif_ebif_ets_select_reverse_1,@function
#endif

#ifndef HAVE_nbif_ets_select_reverse_2
#define HAVE_nbif_ets_select_reverse_2
	.section ".text"
	.align	4
	.global	nbif_ets_select_reverse_2
nbif_ets_select_reverse_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_select_reverse_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_select_reverse_2,.-nbif_ets_select_reverse_2
	.type	nbif_ets_select_reverse_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_reverse_2
#define HAVE_nbif_ebif_ets_select_reverse_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_reverse_2
nbif_ebif_ets_select_reverse_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_reverse_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_reverse_2,.-nbif_ebif_ets_select_reverse_2
	.type	nbif_ebif_ets_select_reverse_2,@function
#endif

#ifndef HAVE_nbif_ets_select_reverse_3
#define HAVE_nbif_ets_select_reverse_3
	.section ".text"
	.align	4
	.global	nbif_ets_select_reverse_3
nbif_ets_select_reverse_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_select_reverse_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ets_select_reverse_3,.-nbif_ets_select_reverse_3
	.type	nbif_ets_select_reverse_3,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_reverse_3
#define HAVE_nbif_ebif_ets_select_reverse_3
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_reverse_3
nbif_ebif_ets_select_reverse_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_reverse_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_reverse_3,.-nbif_ebif_ets_select_reverse_3
	.type	nbif_ebif_ets_select_reverse_3,@function
#endif

#ifndef HAVE_nbif_ets_select_delete_2
#define HAVE_nbif_ets_select_delete_2
	.section ".text"
	.align	4
	.global	nbif_ets_select_delete_2
nbif_ets_select_delete_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_select_delete_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ets_select_delete_2,.-nbif_ets_select_delete_2
	.type	nbif_ets_select_delete_2,@function
#endif

#ifndef HAVE_nbif_ebif_ets_select_delete_2
#define HAVE_nbif_ebif_ets_select_delete_2
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_select_delete_2
nbif_ebif_ets_select_delete_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_select_delete_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_select_delete_2,.-nbif_ebif_ets_select_delete_2
	.type	nbif_ebif_ets_select_delete_2,@function
#endif

#ifndef HAVE_nbif_ets_match_spec_compile_1
#define HAVE_nbif_ets_match_spec_compile_1
	.section ".text"
	.align	4
	.global	nbif_ets_match_spec_compile_1
nbif_ets_match_spec_compile_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ets_match_spec_compile_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ets_match_spec_compile_1,.-nbif_ets_match_spec_compile_1
	.type	nbif_ets_match_spec_compile_1,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_spec_compile_1
#define HAVE_nbif_ebif_ets_match_spec_compile_1
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_spec_compile_1
nbif_ebif_ets_match_spec_compile_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_spec_compile_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_spec_compile_1,.-nbif_ebif_ets_match_spec_compile_1
	.type	nbif_ebif_ets_match_spec_compile_1,@function
#endif

#ifndef HAVE_nbif_ets_match_spec_run_r_3
#define HAVE_nbif_ets_match_spec_run_r_3
	.section ".text"
	.align	4
	.global	nbif_ets_match_spec_run_r_3
nbif_ets_match_spec_run_r_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ets_match_spec_run_r_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ets_match_spec_run_r_3,.-nbif_ets_match_spec_run_r_3
	.type	nbif_ets_match_spec_run_r_3,@function
#endif

#ifndef HAVE_nbif_ebif_ets_match_spec_run_r_3
#define HAVE_nbif_ebif_ets_match_spec_run_r_3
	.section ".text"
	.align	4
	.global	nbif_ebif_ets_match_spec_run_r_3
nbif_ebif_ets_match_spec_run_r_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_ets_match_spec_run_r_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_ets_match_spec_run_r_3,.-nbif_ebif_ets_match_spec_run_r_3
	.type	nbif_ebif_ets_match_spec_run_r_3,@function
#endif

#ifndef HAVE_nbif_os_putenv_2
#define HAVE_nbif_os_putenv_2
	.section ".text"
	.align	4
	.global	nbif_os_putenv_2
nbif_os_putenv_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	os_putenv_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_os_putenv_2,.-nbif_os_putenv_2
	.type	nbif_os_putenv_2,@function
#endif

#ifndef HAVE_nbif_ebif_os_setenv_2
#define HAVE_nbif_ebif_os_setenv_2
	.section ".text"
	.align	4
	.global	nbif_ebif_os_setenv_2
nbif_ebif_os_setenv_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_os_setenv_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_os_setenv_2,.-nbif_ebif_os_setenv_2
	.type	nbif_ebif_os_setenv_2,@function
#endif

#ifndef HAVE_nbif_os_getenv_0
#define HAVE_nbif_os_getenv_0
	.section ".text"
	.align	4
	.global	nbif_os_getenv_0
nbif_os_getenv_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	os_getenv_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_os_getenv_0,.-nbif_os_getenv_0
	.type	nbif_os_getenv_0,@function
#endif

#ifndef HAVE_nbif_ebif_os_getenv_0
#define HAVE_nbif_ebif_os_getenv_0
	.section ".text"
	.align	4
	.global	nbif_ebif_os_getenv_0
nbif_ebif_os_getenv_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_os_getenv_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_os_getenv_0,.-nbif_ebif_os_getenv_0
	.type	nbif_ebif_os_getenv_0,@function
#endif

#ifndef HAVE_nbif_os_getenv_1
#define HAVE_nbif_os_getenv_1
	.section ".text"
	.align	4
	.global	nbif_os_getenv_1
nbif_os_getenv_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	os_getenv_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_os_getenv_1,.-nbif_os_getenv_1
	.type	nbif_os_getenv_1,@function
#endif

#ifndef HAVE_nbif_ebif_os_getenv_1
#define HAVE_nbif_ebif_os_getenv_1
	.section ".text"
	.align	4
	.global	nbif_ebif_os_getenv_1
nbif_ebif_os_getenv_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_os_getenv_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_os_getenv_1,.-nbif_ebif_os_getenv_1
	.type	nbif_ebif_os_getenv_1,@function
#endif

#ifndef HAVE_nbif_os_getpid_0
#define HAVE_nbif_os_getpid_0
	.section ".text"
	.align	4
	.global	nbif_os_getpid_0
nbif_os_getpid_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	os_getpid_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_os_getpid_0,.-nbif_os_getpid_0
	.type	nbif_os_getpid_0,@function
#endif

#ifndef HAVE_nbif_ebif_os_pid_0
#define HAVE_nbif_ebif_os_pid_0
	.section ".text"
	.align	4
	.global	nbif_ebif_os_pid_0
nbif_ebif_os_pid_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	ebif_os_pid_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_ebif_os_pid_0,.-nbif_ebif_os_pid_0
	.type	nbif_ebif_os_pid_0,@function
#endif

#ifndef HAVE_nbif_lists_member_2
#define HAVE_nbif_lists_member_2
	.section ".text"
	.align	4
	.global	nbif_lists_member_2
nbif_lists_member_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	lists_member_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_lists_member_2,.-nbif_lists_member_2
	.type	nbif_lists_member_2,@function
#endif

#ifndef HAVE_nbif_ebif_list_is_element_2
#define HAVE_nbif_ebif_list_is_element_2
	.section ".text"
	.align	4
	.global	nbif_ebif_list_is_element_2
nbif_ebif_list_is_element_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_list_is_element_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_list_is_element_2,.-nbif_ebif_list_is_element_2
	.type	nbif_ebif_list_is_element_2,@function
#endif

#ifndef HAVE_nbif_lists_reverse_2
#define HAVE_nbif_lists_reverse_2
	.section ".text"
	.align	4
	.global	nbif_lists_reverse_2
nbif_lists_reverse_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	lists_reverse_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_lists_reverse_2,.-nbif_lists_reverse_2
	.type	nbif_lists_reverse_2,@function
#endif

#ifndef HAVE_nbif_ebif_list_reverse_2
#define HAVE_nbif_ebif_list_reverse_2
	.section ".text"
	.align	4
	.global	nbif_ebif_list_reverse_2
nbif_ebif_list_reverse_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_list_reverse_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_list_reverse_2,.-nbif_ebif_list_reverse_2
	.type	nbif_ebif_list_reverse_2,@function
#endif

#ifndef HAVE_nbif_lists_keymember_3
#define HAVE_nbif_lists_keymember_3
	.section ".text"
	.align	4
	.global	nbif_lists_keymember_3
nbif_lists_keymember_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	lists_keymember_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_lists_keymember_3,.-nbif_lists_keymember_3
	.type	nbif_lists_keymember_3,@function
#endif

#ifndef HAVE_nbif_ebif_keylist_is_element_3
#define HAVE_nbif_ebif_keylist_is_element_3
	.section ".text"
	.align	4
	.global	nbif_ebif_keylist_is_element_3
nbif_ebif_keylist_is_element_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_keylist_is_element_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_keylist_is_element_3,.-nbif_ebif_keylist_is_element_3
	.type	nbif_ebif_keylist_is_element_3,@function
#endif

#ifndef HAVE_nbif_lists_keysearch_3
#define HAVE_nbif_lists_keysearch_3
	.section ".text"
	.align	4
	.global	nbif_lists_keysearch_3
nbif_lists_keysearch_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	lists_keysearch_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_lists_keysearch_3,.-nbif_lists_keysearch_3
	.type	nbif_lists_keysearch_3,@function
#endif

#ifndef HAVE_nbif_ebif_keylist_search_3
#define HAVE_nbif_ebif_keylist_search_3
	.section ".text"
	.align	4
	.global	nbif_ebif_keylist_search_3
nbif_ebif_keylist_search_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_keylist_search_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_keylist_search_3,.-nbif_ebif_keylist_search_3
	.type	nbif_ebif_keylist_search_3,@function
#endif

#ifndef HAVE_nbif_erts_debug_disassemble_1
#define HAVE_nbif_erts_debug_disassemble_1
	.section ".text"
	.align	4
	.global	nbif_erts_debug_disassemble_1
nbif_erts_debug_disassemble_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	erts_debug_disassemble_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_erts_debug_disassemble_1,.-nbif_erts_debug_disassemble_1
	.type	nbif_erts_debug_disassemble_1,@function
#endif

#ifndef HAVE_nbif_ebif_erts_debug_disassemble_1
#define HAVE_nbif_ebif_erts_debug_disassemble_1
	.section ".text"
	.align	4
	.global	nbif_ebif_erts_debug_disassemble_1
nbif_ebif_erts_debug_disassemble_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_erts_debug_disassemble_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_erts_debug_disassemble_1,.-nbif_ebif_erts_debug_disassemble_1
	.type	nbif_ebif_erts_debug_disassemble_1,@function
#endif

#ifndef HAVE_nbif_erts_debug_breakpoint_2
#define HAVE_nbif_erts_debug_breakpoint_2
	.section ".text"
	.align	4
	.global	nbif_erts_debug_breakpoint_2
nbif_erts_debug_breakpoint_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_debug_breakpoint_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_erts_debug_breakpoint_2,.-nbif_erts_debug_breakpoint_2
	.type	nbif_erts_debug_breakpoint_2,@function
#endif

#ifndef HAVE_nbif_ebif_erts_debug_breakpoint_2
#define HAVE_nbif_ebif_erts_debug_breakpoint_2
	.section ".text"
	.align	4
	.global	nbif_ebif_erts_debug_breakpoint_2
nbif_ebif_erts_debug_breakpoint_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_erts_debug_breakpoint_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_erts_debug_breakpoint_2,.-nbif_ebif_erts_debug_breakpoint_2
	.type	nbif_ebif_erts_debug_breakpoint_2,@function
#endif

#ifndef HAVE_nbif_erts_debug_same_2
#define HAVE_nbif_erts_debug_same_2
	.section ".text"
	.align	4
	.global	nbif_erts_debug_same_2
nbif_erts_debug_same_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	erts_debug_same_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_erts_debug_same_2,.-nbif_erts_debug_same_2
	.type	nbif_erts_debug_same_2,@function
#endif

#ifndef HAVE_nbif_ebif_erts_debug_same_2
#define HAVE_nbif_ebif_erts_debug_same_2
	.section ".text"
	.align	4
	.global	nbif_ebif_erts_debug_same_2
nbif_ebif_erts_debug_same_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_erts_debug_same_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_erts_debug_same_2,.-nbif_ebif_erts_debug_same_2
	.type	nbif_ebif_erts_debug_same_2,@function
#endif

#ifndef HAVE_nbif_erts_debug_flat_size_1
#define HAVE_nbif_erts_debug_flat_size_1
	.section ".text"
	.align	4
	.global	nbif_erts_debug_flat_size_1
nbif_erts_debug_flat_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	erts_debug_flat_size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_erts_debug_flat_size_1,.-nbif_erts_debug_flat_size_1
	.type	nbif_erts_debug_flat_size_1,@function
#endif

#ifndef HAVE_nbif_ebif_erts_debug_flat_size_1
#define HAVE_nbif_ebif_erts_debug_flat_size_1
	.section ".text"
	.align	4
	.global	nbif_ebif_erts_debug_flat_size_1
nbif_ebif_erts_debug_flat_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_erts_debug_flat_size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_erts_debug_flat_size_1,.-nbif_ebif_erts_debug_flat_size_1
	.type	nbif_ebif_erts_debug_flat_size_1,@function
#endif

#ifndef HAVE_nbif_erts_debug_dump_monitors_1
#define HAVE_nbif_erts_debug_dump_monitors_1
	.section ".text"
	.align	4
	.global	nbif_erts_debug_dump_monitors_1
nbif_erts_debug_dump_monitors_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	erts_debug_dump_monitors_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_erts_debug_dump_monitors_1,.-nbif_erts_debug_dump_monitors_1
	.type	nbif_erts_debug_dump_monitors_1,@function
#endif

#ifndef HAVE_nbif_erts_debug_dump_links_1
#define HAVE_nbif_erts_debug_dump_links_1
	.section ".text"
	.align	4
	.global	nbif_erts_debug_dump_links_1
nbif_erts_debug_dump_links_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	erts_debug_dump_links_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_erts_debug_dump_links_1,.-nbif_erts_debug_dump_links_1
	.type	nbif_erts_debug_dump_links_1,@function
#endif

#ifndef HAVE_nbif_code_get_chunk_2
#define HAVE_nbif_code_get_chunk_2
	.section ".text"
	.align	4
	.global	nbif_code_get_chunk_2
nbif_code_get_chunk_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	code_get_chunk_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_code_get_chunk_2,.-nbif_code_get_chunk_2
	.type	nbif_code_get_chunk_2,@function
#endif

#ifndef HAVE_nbif_ebif_code_get_chunk_2
#define HAVE_nbif_ebif_code_get_chunk_2
	.section ".text"
	.align	4
	.global	nbif_ebif_code_get_chunk_2
nbif_ebif_code_get_chunk_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_code_get_chunk_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_ebif_code_get_chunk_2,.-nbif_ebif_code_get_chunk_2
	.type	nbif_ebif_code_get_chunk_2,@function
#endif

#ifndef HAVE_nbif_code_module_md5_1
#define HAVE_nbif_code_module_md5_1
	.section ".text"
	.align	4
	.global	nbif_code_module_md5_1
nbif_code_module_md5_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	code_module_md5_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_code_module_md5_1,.-nbif_code_module_md5_1
	.type	nbif_code_module_md5_1,@function
#endif

#ifndef HAVE_nbif_ebif_code_module_md5_1
#define HAVE_nbif_ebif_code_module_md5_1
	.section ".text"
	.align	4
	.global	nbif_ebif_code_module_md5_1
nbif_ebif_code_module_md5_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_code_module_md5_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_code_module_md5_1,.-nbif_ebif_code_module_md5_1
	.type	nbif_ebif_code_module_md5_1,@function
#endif

#ifndef HAVE_nbif_code_make_stub_module_3
#define HAVE_nbif_code_make_stub_module_3
	.section ".text"
	.align	4
	.global	nbif_code_make_stub_module_3
nbif_code_make_stub_module_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	code_make_stub_module_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_code_make_stub_module_3,.-nbif_code_make_stub_module_3
	.type	nbif_code_make_stub_module_3,@function
#endif

#ifndef HAVE_nbif_ebif_code_make_stub_module_3
#define HAVE_nbif_ebif_code_make_stub_module_3
	.section ".text"
	.align	4
	.global	nbif_ebif_code_make_stub_module_3
nbif_ebif_code_make_stub_module_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	ebif_code_make_stub_module_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_ebif_code_make_stub_module_3,.-nbif_ebif_code_make_stub_module_3
	.type	nbif_ebif_code_make_stub_module_3,@function
#endif

#ifndef HAVE_nbif_code_is_module_native_1
#define HAVE_nbif_code_is_module_native_1
	.section ".text"
	.align	4
	.global	nbif_code_is_module_native_1
nbif_code_is_module_native_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	code_is_module_native_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_code_is_module_native_1,.-nbif_code_is_module_native_1
	.type	nbif_code_is_module_native_1,@function
#endif

#ifndef HAVE_nbif_ebif_code_is_native_1
#define HAVE_nbif_ebif_code_is_native_1
	.section ".text"
	.align	4
	.global	nbif_ebif_code_is_native_1
nbif_ebif_code_is_native_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	ebif_code_is_native_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_ebif_code_is_native_1,.-nbif_ebif_code_is_native_1
	.type	nbif_ebif_code_is_native_1,@function
#endif

#ifndef HAVE_nbif_blocking_read_file_1
#define HAVE_nbif_blocking_read_file_1
	.section ".text"
	.align	4
	.global	nbif_blocking_read_file_1
nbif_blocking_read_file_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	blocking_read_file_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_blocking_read_file_1,.-nbif_blocking_read_file_1
	.type	nbif_blocking_read_file_1,@function
#endif

#ifndef HAVE_nbif_hibernate_3
#define HAVE_nbif_hibernate_3
	.section ".text"
	.align	4
	.global	nbif_hibernate_3
nbif_hibernate_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hibernate_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hibernate_3,.-nbif_hibernate_3
	.type	nbif_hibernate_3,@function
#endif

#ifndef HAVE_nbif_error_logger_warning_map_0
#define HAVE_nbif_error_logger_warning_map_0
	.section ".text"
	.align	4
	.global	nbif_error_logger_warning_map_0
nbif_error_logger_warning_map_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	error_logger_warning_map_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_error_logger_warning_map_0,.-nbif_error_logger_warning_map_0
	.type	nbif_error_logger_warning_map_0,@function
#endif

#ifndef HAVE_nbif_get_module_info_1
#define HAVE_nbif_get_module_info_1
	.section ".text"
	.align	4
	.global	nbif_get_module_info_1
nbif_get_module_info_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	get_module_info_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_get_module_info_1,.-nbif_get_module_info_1
	.type	nbif_get_module_info_1,@function
#endif

#ifndef HAVE_nbif_get_module_info_2
#define HAVE_nbif_get_module_info_2
	.section ".text"
	.align	4
	.global	nbif_get_module_info_2
nbif_get_module_info_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	get_module_info_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_get_module_info_2,.-nbif_get_module_info_2
	.type	nbif_get_module_info_2,@function
#endif

#ifndef HAVE_nbif_is_boolean_1
#define HAVE_nbif_is_boolean_1
	.section ".text"
	.align	4
	.global	nbif_is_boolean_1
nbif_is_boolean_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	is_boolean_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_is_boolean_1,.-nbif_is_boolean_1
	.type	nbif_is_boolean_1,@function
#endif

#ifndef HAVE_nbif_string_to_integer_1
#define HAVE_nbif_string_to_integer_1
	.section ".text"
	.align	4
	.global	nbif_string_to_integer_1
nbif_string_to_integer_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	string_to_integer_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_string_to_integer_1,.-nbif_string_to_integer_1
	.type	nbif_string_to_integer_1,@function
#endif

#ifndef HAVE_nbif_string_to_float_1
#define HAVE_nbif_string_to_float_1
	.section ".text"
	.align	4
	.global	nbif_string_to_float_1
nbif_string_to_float_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	string_to_float_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_string_to_float_1,.-nbif_string_to_float_1
	.type	nbif_string_to_float_1,@function
#endif

#ifndef HAVE_nbif_make_fun_3
#define HAVE_nbif_make_fun_3
	.section ".text"
	.align	4
	.global	nbif_make_fun_3
nbif_make_fun_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	make_fun_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_make_fun_3,.-nbif_make_fun_3
	.type	nbif_make_fun_3,@function
#endif

#ifndef HAVE_nbif_iolist_size_1
#define HAVE_nbif_iolist_size_1
	.section ".text"
	.align	4
	.global	nbif_iolist_size_1
nbif_iolist_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	iolist_size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_iolist_size_1,.-nbif_iolist_size_1
	.type	nbif_iolist_size_1,@function
#endif

#ifndef HAVE_nbif_iolist_to_binary_1
#define HAVE_nbif_iolist_to_binary_1
	.section ".text"
	.align	4
	.global	nbif_iolist_to_binary_1
nbif_iolist_to_binary_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	iolist_to_binary_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_iolist_to_binary_1,.-nbif_iolist_to_binary_1
	.type	nbif_iolist_to_binary_1,@function
#endif

#ifndef HAVE_nbif_list_to_existing_atom_1
#define HAVE_nbif_list_to_existing_atom_1
	.section ".text"
	.align	4
	.global	nbif_list_to_existing_atom_1
nbif_list_to_existing_atom_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	list_to_existing_atom_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_list_to_existing_atom_1,.-nbif_list_to_existing_atom_1
	.type	nbif_list_to_existing_atom_1,@function
#endif

#ifndef HAVE_nbif_hash_2
#define HAVE_nbif_hash_2
	.section ".text"
	.align	4
	.global	nbif_hash_2
nbif_hash_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hash_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hash_2,.-nbif_hash_2
	.type	nbif_hash_2,@function
#endif

#ifndef HAVE_nbif_old_binary_to_term_1
#define HAVE_nbif_old_binary_to_term_1
	.section ".text"
	.align	4
	.global	nbif_old_binary_to_term_1
nbif_old_binary_to_term_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	old_binary_to_term_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_old_binary_to_term_1,.-nbif_old_binary_to_term_1
	.type	nbif_old_binary_to_term_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_write_u8_2
#define HAVE_nbif_hipe_bifs_write_u8_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_write_u8_2
nbif_hipe_bifs_write_u8_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_write_u8_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_write_u8_2,.-nbif_hipe_bifs_write_u8_2
	.type	nbif_hipe_bifs_write_u8_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_write_u32_2
#define HAVE_nbif_hipe_bifs_write_u32_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_write_u32_2
nbif_hipe_bifs_write_u32_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_write_u32_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_write_u32_2,.-nbif_hipe_bifs_write_u32_2
	.type	nbif_hipe_bifs_write_u32_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_bytearray_2
#define HAVE_nbif_hipe_bifs_bytearray_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_bytearray_2
nbif_hipe_bifs_bytearray_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_bytearray_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_bytearray_2,.-nbif_hipe_bifs_bytearray_2
	.type	nbif_hipe_bifs_bytearray_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_bytearray_sub_2
#define HAVE_nbif_hipe_bifs_bytearray_sub_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_bytearray_sub_2
nbif_hipe_bifs_bytearray_sub_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_bytearray_sub_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_bytearray_sub_2,.-nbif_hipe_bifs_bytearray_sub_2
	.type	nbif_hipe_bifs_bytearray_sub_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_bytearray_update_3
#define HAVE_nbif_hipe_bifs_bytearray_update_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_bytearray_update_3
nbif_hipe_bifs_bytearray_update_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_bytearray_update_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_bytearray_update_3,.-nbif_hipe_bifs_bytearray_update_3
	.type	nbif_hipe_bifs_bytearray_update_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_array_2
#define HAVE_nbif_hipe_bifs_array_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_array_2
nbif_hipe_bifs_array_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_array_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_array_2,.-nbif_hipe_bifs_array_2
	.type	nbif_hipe_bifs_array_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_array_length_1
#define HAVE_nbif_hipe_bifs_array_length_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_array_length_1
nbif_hipe_bifs_array_length_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_array_length_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_array_length_1,.-nbif_hipe_bifs_array_length_1
	.type	nbif_hipe_bifs_array_length_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_array_sub_2
#define HAVE_nbif_hipe_bifs_array_sub_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_array_sub_2
nbif_hipe_bifs_array_sub_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_array_sub_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_array_sub_2,.-nbif_hipe_bifs_array_sub_2
	.type	nbif_hipe_bifs_array_sub_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_array_update_3
#define HAVE_nbif_hipe_bifs_array_update_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_array_update_3
nbif_hipe_bifs_array_update_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_array_update_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_array_update_3,.-nbif_hipe_bifs_array_update_3
	.type	nbif_hipe_bifs_array_update_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_ref_1
#define HAVE_nbif_hipe_bifs_ref_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_ref_1
nbif_hipe_bifs_ref_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_ref_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_ref_1,.-nbif_hipe_bifs_ref_1
	.type	nbif_hipe_bifs_ref_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_ref_get_1
#define HAVE_nbif_hipe_bifs_ref_get_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_ref_get_1
nbif_hipe_bifs_ref_get_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_ref_get_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_ref_get_1,.-nbif_hipe_bifs_ref_get_1
	.type	nbif_hipe_bifs_ref_get_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_ref_set_2
#define HAVE_nbif_hipe_bifs_ref_set_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_ref_set_2
nbif_hipe_bifs_ref_set_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_ref_set_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_ref_set_2,.-nbif_hipe_bifs_ref_set_2
	.type	nbif_hipe_bifs_ref_set_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_enter_code_2
#define HAVE_nbif_hipe_bifs_enter_code_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_enter_code_2
nbif_hipe_bifs_enter_code_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_enter_code_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_enter_code_2,.-nbif_hipe_bifs_enter_code_2
	.type	nbif_hipe_bifs_enter_code_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_alloc_data_2
#define HAVE_nbif_hipe_bifs_alloc_data_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_alloc_data_2
nbif_hipe_bifs_alloc_data_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_alloc_data_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_alloc_data_2,.-nbif_hipe_bifs_alloc_data_2
	.type	nbif_hipe_bifs_alloc_data_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_constants_size_0
#define HAVE_nbif_hipe_bifs_constants_size_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_constants_size_0
nbif_hipe_bifs_constants_size_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_constants_size_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_constants_size_0,.-nbif_hipe_bifs_constants_size_0
	.type	nbif_hipe_bifs_constants_size_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_merge_term_1
#define HAVE_nbif_hipe_bifs_merge_term_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_merge_term_1
nbif_hipe_bifs_merge_term_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_merge_term_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_merge_term_1,.-nbif_hipe_bifs_merge_term_1
	.type	nbif_hipe_bifs_merge_term_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_fun_to_address_1
#define HAVE_nbif_hipe_bifs_fun_to_address_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_fun_to_address_1
nbif_hipe_bifs_fun_to_address_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_fun_to_address_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_fun_to_address_1,.-nbif_hipe_bifs_fun_to_address_1
	.type	nbif_hipe_bifs_fun_to_address_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_set_native_address_3
#define HAVE_nbif_hipe_bifs_set_native_address_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_set_native_address_3
nbif_hipe_bifs_set_native_address_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_set_native_address_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_set_native_address_3,.-nbif_hipe_bifs_set_native_address_3
	.type	nbif_hipe_bifs_set_native_address_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_set_funinfo_native_address_3
#define HAVE_nbif_hipe_bifs_set_funinfo_native_address_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_set_funinfo_native_address_3
nbif_hipe_bifs_set_funinfo_native_address_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_set_funinfo_native_address_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_set_funinfo_native_address_3,.-nbif_hipe_bifs_set_funinfo_native_address_3
	.type	nbif_hipe_bifs_set_funinfo_native_address_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_invalidate_funinfo_native_addresses_1
#define HAVE_nbif_hipe_bifs_invalidate_funinfo_native_addresses_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_invalidate_funinfo_native_addresses_1
nbif_hipe_bifs_invalidate_funinfo_native_addresses_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_invalidate_funinfo_native_addresses_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_invalidate_funinfo_native_addresses_1,.-nbif_hipe_bifs_invalidate_funinfo_native_addresses_1
	.type	nbif_hipe_bifs_invalidate_funinfo_native_addresses_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_update_code_size_3
#define HAVE_nbif_hipe_bifs_update_code_size_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_update_code_size_3
nbif_hipe_bifs_update_code_size_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_update_code_size_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_update_code_size_3,.-nbif_hipe_bifs_update_code_size_3
	.type	nbif_hipe_bifs_update_code_size_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_code_size_1
#define HAVE_nbif_hipe_bifs_code_size_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_code_size_1
nbif_hipe_bifs_code_size_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_code_size_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_code_size_1,.-nbif_hipe_bifs_code_size_1
	.type	nbif_hipe_bifs_code_size_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_enter_sdesc_1
#define HAVE_nbif_hipe_bifs_enter_sdesc_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_enter_sdesc_1
nbif_hipe_bifs_enter_sdesc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_enter_sdesc_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_enter_sdesc_1,.-nbif_hipe_bifs_enter_sdesc_1
	.type	nbif_hipe_bifs_enter_sdesc_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_bif_address_3
#define HAVE_nbif_hipe_bifs_bif_address_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_bif_address_3
nbif_hipe_bifs_bif_address_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_bif_address_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_bif_address_3,.-nbif_hipe_bifs_bif_address_3
	.type	nbif_hipe_bifs_bif_address_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_primop_address_1
#define HAVE_nbif_hipe_bifs_primop_address_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_primop_address_1
nbif_hipe_bifs_primop_address_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_primop_address_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_primop_address_1,.-nbif_hipe_bifs_primop_address_1
	.type	nbif_hipe_bifs_primop_address_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_atom_to_word_1
#define HAVE_nbif_hipe_bifs_atom_to_word_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_atom_to_word_1
nbif_hipe_bifs_atom_to_word_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_atom_to_word_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_atom_to_word_1,.-nbif_hipe_bifs_atom_to_word_1
	.type	nbif_hipe_bifs_atom_to_word_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_term_to_word_1
#define HAVE_nbif_hipe_bifs_term_to_word_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_term_to_word_1
nbif_hipe_bifs_term_to_word_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_term_to_word_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_term_to_word_1,.-nbif_hipe_bifs_term_to_word_1
	.type	nbif_hipe_bifs_term_to_word_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_make_fe_3
#define HAVE_nbif_hipe_bifs_make_fe_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_make_fe_3
nbif_hipe_bifs_make_fe_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_make_fe_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_make_fe_3,.-nbif_hipe_bifs_make_fe_3
	.type	nbif_hipe_bifs_make_fe_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_find_na_or_make_stub_2
#define HAVE_nbif_hipe_bifs_find_na_or_make_stub_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_find_na_or_make_stub_2
nbif_hipe_bifs_find_na_or_make_stub_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_find_na_or_make_stub_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_find_na_or_make_stub_2,.-nbif_hipe_bifs_find_na_or_make_stub_2
	.type	nbif_hipe_bifs_find_na_or_make_stub_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_check_crc_1
#define HAVE_nbif_hipe_bifs_check_crc_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_check_crc_1
nbif_hipe_bifs_check_crc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_check_crc_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_check_crc_1,.-nbif_hipe_bifs_check_crc_1
	.type	nbif_hipe_bifs_check_crc_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_system_crc_1
#define HAVE_nbif_hipe_bifs_system_crc_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_system_crc_1
nbif_hipe_bifs_system_crc_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_system_crc_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_system_crc_1,.-nbif_hipe_bifs_system_crc_1
	.type	nbif_hipe_bifs_system_crc_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_get_rts_param_1
#define HAVE_nbif_hipe_bifs_get_rts_param_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_get_rts_param_1
nbif_hipe_bifs_get_rts_param_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_get_rts_param_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_get_rts_param_1,.-nbif_hipe_bifs_get_rts_param_1
	.type	nbif_hipe_bifs_get_rts_param_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_patch_insn_3
#define HAVE_nbif_hipe_bifs_patch_insn_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_patch_insn_3
nbif_hipe_bifs_patch_insn_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_patch_insn_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_patch_insn_3,.-nbif_hipe_bifs_patch_insn_3
	.type	nbif_hipe_bifs_patch_insn_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_patch_call_3
#define HAVE_nbif_hipe_bifs_patch_call_3
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_patch_call_3
nbif_hipe_bifs_patch_call_3:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG2
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_patch_call_3
	addl	$16, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_3_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_patch_call_3,.-nbif_hipe_bifs_patch_call_3
	.type	nbif_hipe_bifs_patch_call_3,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_add_ref_2
#define HAVE_nbif_hipe_bifs_add_ref_2
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_add_ref_2
nbif_hipe_bifs_add_ref_2:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG1
	pushl	ARG0
	pushl	P
	call	hipe_bifs_add_ref_2
	addl	$12, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_2_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_add_ref_2,.-nbif_hipe_bifs_add_ref_2
	.type	nbif_hipe_bifs_add_ref_2,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_mark_referred_from_1
#define HAVE_nbif_hipe_bifs_mark_referred_from_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_mark_referred_from_1
nbif_hipe_bifs_mark_referred_from_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_mark_referred_from_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_mark_referred_from_1,.-nbif_hipe_bifs_mark_referred_from_1
	.type	nbif_hipe_bifs_mark_referred_from_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_remove_refs_from_1
#define HAVE_nbif_hipe_bifs_remove_refs_from_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_remove_refs_from_1
nbif_hipe_bifs_remove_refs_from_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_remove_refs_from_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_remove_refs_from_1,.-nbif_hipe_bifs_remove_refs_from_1
	.type	nbif_hipe_bifs_remove_refs_from_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_redirect_referred_from_1
#define HAVE_nbif_hipe_bifs_redirect_referred_from_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_redirect_referred_from_1
nbif_hipe_bifs_redirect_referred_from_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_redirect_referred_from_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_redirect_referred_from_1,.-nbif_hipe_bifs_redirect_referred_from_1
	.type	nbif_hipe_bifs_redirect_referred_from_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_call_count_on_1
#define HAVE_nbif_hipe_bifs_call_count_on_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_call_count_on_1
nbif_hipe_bifs_call_count_on_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_call_count_on_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_call_count_on_1,.-nbif_hipe_bifs_call_count_on_1
	.type	nbif_hipe_bifs_call_count_on_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_call_count_off_1
#define HAVE_nbif_hipe_bifs_call_count_off_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_call_count_off_1
nbif_hipe_bifs_call_count_off_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_call_count_off_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_call_count_off_1,.-nbif_hipe_bifs_call_count_off_1
	.type	nbif_hipe_bifs_call_count_off_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_call_count_get_1
#define HAVE_nbif_hipe_bifs_call_count_get_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_call_count_get_1
nbif_hipe_bifs_call_count_get_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_call_count_get_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_call_count_get_1,.-nbif_hipe_bifs_call_count_get_1
	.type	nbif_hipe_bifs_call_count_get_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_call_count_clear_1
#define HAVE_nbif_hipe_bifs_call_count_clear_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_call_count_clear_1
nbif_hipe_bifs_call_count_clear_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_call_count_clear_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_call_count_clear_1,.-nbif_hipe_bifs_call_count_clear_1
	.type	nbif_hipe_bifs_call_count_clear_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_trap_count_get_0
#define HAVE_nbif_hipe_bifs_trap_count_get_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_trap_count_get_0
nbif_hipe_bifs_trap_count_get_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_trap_count_get_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_trap_count_get_0,.-nbif_hipe_bifs_trap_count_get_0
	.type	nbif_hipe_bifs_trap_count_get_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_trap_count_clear_0
#define HAVE_nbif_hipe_bifs_trap_count_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_trap_count_clear_0
nbif_hipe_bifs_trap_count_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_trap_count_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_trap_count_clear_0,.-nbif_hipe_bifs_trap_count_clear_0
	.type	nbif_hipe_bifs_trap_count_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_process_info_0
#define HAVE_nbif_hipe_bifs_process_info_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_process_info_0
nbif_hipe_bifs_process_info_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_process_info_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_process_info_0,.-nbif_hipe_bifs_process_info_0
	.type	nbif_hipe_bifs_process_info_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_process_info_clear_0
#define HAVE_nbif_hipe_bifs_process_info_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_process_info_clear_0
nbif_hipe_bifs_process_info_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_process_info_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_process_info_clear_0,.-nbif_hipe_bifs_process_info_clear_0
	.type	nbif_hipe_bifs_process_info_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_message_info_0
#define HAVE_nbif_hipe_bifs_message_info_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_message_info_0
nbif_hipe_bifs_message_info_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_message_info_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_message_info_0,.-nbif_hipe_bifs_message_info_0
	.type	nbif_hipe_bifs_message_info_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_message_info_clear_0
#define HAVE_nbif_hipe_bifs_message_info_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_message_info_clear_0
nbif_hipe_bifs_message_info_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_message_info_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_message_info_clear_0,.-nbif_hipe_bifs_message_info_clear_0
	.type	nbif_hipe_bifs_message_info_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_message_sizes_0
#define HAVE_nbif_hipe_bifs_message_sizes_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_message_sizes_0
nbif_hipe_bifs_message_sizes_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_message_sizes_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_message_sizes_0,.-nbif_hipe_bifs_message_sizes_0
	.type	nbif_hipe_bifs_message_sizes_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_gc_info_0
#define HAVE_nbif_hipe_bifs_gc_info_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_gc_info_0
nbif_hipe_bifs_gc_info_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_gc_info_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_gc_info_0,.-nbif_hipe_bifs_gc_info_0
	.type	nbif_hipe_bifs_gc_info_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_shared_gc_info_0
#define HAVE_nbif_hipe_bifs_shared_gc_info_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_shared_gc_info_0
nbif_hipe_bifs_shared_gc_info_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_shared_gc_info_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_shared_gc_info_0,.-nbif_hipe_bifs_shared_gc_info_0
	.type	nbif_hipe_bifs_shared_gc_info_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_incremental_gc_info_0
#define HAVE_nbif_hipe_bifs_incremental_gc_info_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_incremental_gc_info_0
nbif_hipe_bifs_incremental_gc_info_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_incremental_gc_info_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_incremental_gc_info_0,.-nbif_hipe_bifs_incremental_gc_info_0
	.type	nbif_hipe_bifs_incremental_gc_info_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_gc_info_clear_0
#define HAVE_nbif_hipe_bifs_gc_info_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_gc_info_clear_0
nbif_hipe_bifs_gc_info_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_gc_info_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_gc_info_clear_0,.-nbif_hipe_bifs_gc_info_clear_0
	.type	nbif_hipe_bifs_gc_info_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_pause_times_0
#define HAVE_nbif_hipe_bifs_pause_times_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_pause_times_0
nbif_hipe_bifs_pause_times_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_pause_times_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_pause_times_0,.-nbif_hipe_bifs_pause_times_0
	.type	nbif_hipe_bifs_pause_times_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_system_timer_0
#define HAVE_nbif_hipe_bifs_system_timer_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_system_timer_0
nbif_hipe_bifs_system_timer_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_system_timer_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_system_timer_0,.-nbif_hipe_bifs_system_timer_0
	.type	nbif_hipe_bifs_system_timer_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_system_timer_clear_0
#define HAVE_nbif_hipe_bifs_system_timer_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_system_timer_clear_0
nbif_hipe_bifs_system_timer_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_system_timer_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_system_timer_clear_0,.-nbif_hipe_bifs_system_timer_clear_0
	.type	nbif_hipe_bifs_system_timer_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_send_timer_0
#define HAVE_nbif_hipe_bifs_send_timer_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_send_timer_0
nbif_hipe_bifs_send_timer_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_send_timer_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_send_timer_0,.-nbif_hipe_bifs_send_timer_0
	.type	nbif_hipe_bifs_send_timer_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_send_timer_clear_0
#define HAVE_nbif_hipe_bifs_send_timer_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_send_timer_clear_0
nbif_hipe_bifs_send_timer_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_send_timer_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_send_timer_clear_0,.-nbif_hipe_bifs_send_timer_clear_0
	.type	nbif_hipe_bifs_send_timer_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_gc_timer_0
#define HAVE_nbif_hipe_bifs_gc_timer_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_gc_timer_0
nbif_hipe_bifs_gc_timer_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_gc_timer_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_gc_timer_0,.-nbif_hipe_bifs_gc_timer_0
	.type	nbif_hipe_bifs_gc_timer_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_shared_gc_timer_0
#define HAVE_nbif_hipe_bifs_shared_gc_timer_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_shared_gc_timer_0
nbif_hipe_bifs_shared_gc_timer_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_shared_gc_timer_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_shared_gc_timer_0,.-nbif_hipe_bifs_shared_gc_timer_0
	.type	nbif_hipe_bifs_shared_gc_timer_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_gc_timer_clear_0
#define HAVE_nbif_hipe_bifs_gc_timer_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_gc_timer_clear_0
nbif_hipe_bifs_gc_timer_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_gc_timer_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_gc_timer_clear_0,.-nbif_hipe_bifs_gc_timer_clear_0
	.type	nbif_hipe_bifs_gc_timer_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_misc_timer_0
#define HAVE_nbif_hipe_bifs_misc_timer_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_misc_timer_0
nbif_hipe_bifs_misc_timer_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_misc_timer_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_misc_timer_0,.-nbif_hipe_bifs_misc_timer_0
	.type	nbif_hipe_bifs_misc_timer_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_misc_timer_clear_0
#define HAVE_nbif_hipe_bifs_misc_timer_clear_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_misc_timer_clear_0
nbif_hipe_bifs_misc_timer_clear_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_misc_timer_clear_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_misc_timer_clear_0,.-nbif_hipe_bifs_misc_timer_clear_0
	.type	nbif_hipe_bifs_misc_timer_clear_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_get_hrvtime_0
#define HAVE_nbif_hipe_bifs_get_hrvtime_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_get_hrvtime_0
nbif_hipe_bifs_get_hrvtime_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_get_hrvtime_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_get_hrvtime_0,.-nbif_hipe_bifs_get_hrvtime_0
	.type	nbif_hipe_bifs_get_hrvtime_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_stop_hrvtime_0
#define HAVE_nbif_hipe_bifs_stop_hrvtime_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_stop_hrvtime_0
nbif_hipe_bifs_stop_hrvtime_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_stop_hrvtime_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_stop_hrvtime_0,.-nbif_hipe_bifs_stop_hrvtime_0
	.type	nbif_hipe_bifs_stop_hrvtime_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_estack_1
#define HAVE_nbif_hipe_bifs_show_estack_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_estack_1
nbif_hipe_bifs_show_estack_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_show_estack_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_estack_1,.-nbif_hipe_bifs_show_estack_1
	.type	nbif_hipe_bifs_show_estack_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_heap_1
#define HAVE_nbif_hipe_bifs_show_heap_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_heap_1
nbif_hipe_bifs_show_heap_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_show_heap_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_heap_1,.-nbif_hipe_bifs_show_heap_1
	.type	nbif_hipe_bifs_show_heap_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_nstack_1
#define HAVE_nbif_hipe_bifs_show_nstack_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_nstack_1
nbif_hipe_bifs_show_nstack_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_show_nstack_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_nstack_1,.-nbif_hipe_bifs_show_nstack_1
	.type	nbif_hipe_bifs_show_nstack_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_nstack_used_size_0
#define HAVE_nbif_hipe_bifs_nstack_used_size_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_nstack_used_size_0
nbif_hipe_bifs_nstack_used_size_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_nstack_used_size_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_nstack_used_size_0,.-nbif_hipe_bifs_nstack_used_size_0
	.type	nbif_hipe_bifs_nstack_used_size_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_pcb_1
#define HAVE_nbif_hipe_bifs_show_pcb_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_pcb_1
nbif_hipe_bifs_show_pcb_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_show_pcb_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_pcb_1,.-nbif_hipe_bifs_show_pcb_1
	.type	nbif_hipe_bifs_show_pcb_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_term_1
#define HAVE_nbif_hipe_bifs_show_term_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_term_1
nbif_hipe_bifs_show_term_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_show_term_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_term_1,.-nbif_hipe_bifs_show_term_1
	.type	nbif_hipe_bifs_show_term_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_literals_0
#define HAVE_nbif_hipe_bifs_show_literals_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_literals_0
nbif_hipe_bifs_show_literals_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_show_literals_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_literals_0,.-nbif_hipe_bifs_show_literals_0
	.type	nbif_hipe_bifs_show_literals_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_in_native_0
#define HAVE_nbif_hipe_bifs_in_native_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_in_native_0
nbif_hipe_bifs_in_native_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_in_native_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_in_native_0,.-nbif_hipe_bifs_in_native_0
	.type	nbif_hipe_bifs_in_native_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_modeswitch_debug_on_0
#define HAVE_nbif_hipe_bifs_modeswitch_debug_on_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_modeswitch_debug_on_0
nbif_hipe_bifs_modeswitch_debug_on_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_modeswitch_debug_on_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_modeswitch_debug_on_0,.-nbif_hipe_bifs_modeswitch_debug_on_0
	.type	nbif_hipe_bifs_modeswitch_debug_on_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_modeswitch_debug_off_0
#define HAVE_nbif_hipe_bifs_modeswitch_debug_off_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_modeswitch_debug_off_0
nbif_hipe_bifs_modeswitch_debug_off_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_modeswitch_debug_off_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_modeswitch_debug_off_0,.-nbif_hipe_bifs_modeswitch_debug_off_0
	.type	nbif_hipe_bifs_modeswitch_debug_off_0,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_test_reschedule_1
#define HAVE_nbif_hipe_bifs_test_reschedule_1
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_test_reschedule_1
nbif_hipe_bifs_test_reschedule_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	hipe_bifs_test_reschedule_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_1_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_test_reschedule_1,.-nbif_hipe_bifs_test_reschedule_1
	.type	nbif_hipe_bifs_test_reschedule_1,@function
#endif

#ifndef HAVE_nbif_hipe_bifs_show_message_area_0
#define HAVE_nbif_hipe_bifs_show_message_area_0
	.section ".text"
	.align	4
	.global	nbif_hipe_bifs_show_message_area_0
nbif_hipe_bifs_show_message_area_0:
	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	P
	call	hipe_bifs_show_message_area_0
	addl	$4, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* throw exception if failure, otherwise return */
	TEST_GOT_EXN
	jz	nbif_0_simple_exception
	NSP_RET0
	.size	nbif_hipe_bifs_show_message_area_0,.-nbif_hipe_bifs_show_message_area_0
	.type	nbif_hipe_bifs_show_message_area_0,@function
#endif


/*
 * Guard BIFs.
 * GBIF_LIST(FunctionAtom,Arity,CFun)
 */

/* $Id$
 * GBIF_LIST(FunctionAtom,Arity,CFun)
 * manually maintained for now -- expand when necessary
 */

#ifndef HAVE_gbif_node_1
#define HAVE_gbif_node_1
	.section ".text"
	.align	4
	.global	gbif_node_1
gbif_node_1:
	/* copy native stack pointer */
	

	/* switch to C stack */
	SWITCH_ERLANG_TO_C

	/* make the call on the C stack */
	pushl	ARG0
	pushl	P
	call	node_1
	addl	$8, %esp

	/* switch to native stack */
	SWITCH_C_TO_ERLANG

	/* return */
	NSP_RET0
	.size	gbif_node_1,.-gbif_node_1
	.type	gbif_node_1,@function
#endif


